{
    This file is part of the Free Pascal run time library and compiler.
    Copyright (c) 1998-2002 by the Free Pascal development team

    Internal Function/Constant Evaluator numbers

    See the file COPYING.FPC, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}
unit compinnr;

{$i fpcdefs.inc}

{$packenum 4}

interface

const
  { this file needs to be kept in sync with rtl/inc/innr.in }
  in_cpu_first   = 10000;
  in_x86_mm_first    = 11000;

type
   tinlinenumber=(
     in_none              = -1,
{ Internal functions }
     in_lo_word           = 1,
     in_hi_word           = 2,
     in_lo_long           = 3,
     in_hi_long           = 4,
     in_ord_x             = 5,
     in_length_x          = 6,
     in_chr_byte          = 7,
     in_write_x           = 14,
     in_writeln_x         = 15,
     in_read_x            = 16,
     in_readln_x          = 17,
     in_concat_x          = 18,
     in_assigned_x        = 19,
     in_str_x_string      = 20,
     in_ofs_x             = 21,
     in_sizeof_x          = 22,
     in_typeof_x          = 23,
     in_val_x             = 24,
     in_reset_x           = 25,
     in_rewrite_x         = 26,
     in_low_x             = 27,
     in_high_x            = 28,
     in_seg_x             = 29,
     in_pred_x            = 30,
     in_succ_x            = 31,
     in_reset_typedfile   = 32,
     in_rewrite_typedfile = 33,
     in_settextbuf_file_x = 34,
     in_inc_x             = 35,
     in_dec_x             = 36,
     in_include_x_y       = 37,
     in_exclude_x_y       = 38,
     in_break             = 39,
     in_continue          = 40,
     in_assert_x_y        = 41,
     in_addr_x            = 42,
     in_typeinfo_x        = 43,
     in_setlength_x       = 44,
     in_finalize_x        = 45,
     in_new_x             = 46,
     in_dispose_x         = 47,
     in_exit              = 48,
     in_copy_x            = 49,
     in_initialize_x      = 50,
     in_leave             = 51, {macpas}
     in_cycle             = 52, {macpas}
     in_slice_x           = 53,
     in_unaligned_x       = 54,
     in_get_frame         = 56,
     in_get_caller_addr   = 57,
     in_get_caller_frame  = 58,
     in_pack_x_y_z        = 59,
     in_unpack_x_y_z      = 60,
     in_bitsizeof_x       = 61,
     in_writestr_x        = 62,
     in_readstr_x         = 63,
     in_abs_long          = 64,
     in_ror_x             = 65,
     in_ror_x_y           = 66,
     in_rol_x             = 67,
     in_rol_x_y           = 68,
     in_objc_selector_x   = 69,
     in_objc_protocol_x   = 70,
     in_objc_encode_x     = 71,
     in_sar_x_y           = 72,
     in_sar_x             = 73,
     in_bsf_x             = 74,
     in_bsr_x             = 75,
     in_default_x         = 76,
     in_box_x             = 77, { managed platforms: wrap in class instance }
     in_unbox_x_y         = 78, { manage platforms: extract from class instance }
     in_popcnt_x          = 79,
     in_aligned_x         = 80,
     in_setstring_x_y_z   = 81,
     in_insert_x_y_z      = 82,
     in_delete_x_y_z      = 83,
     in_reset_typedfile_name   = 84,
     in_rewrite_typedfile_name = 85,
     in_and_assign_x_y    = 86,
     in_or_assign_x_y     = 87,
     in_xor_assign_x_y    = 88,
     in_sar_assign_x_y    = 89,
     in_shl_assign_x_y    = 90,
     in_shr_assign_x_y    = 91,
     in_rol_assign_x_y    = 92,
     in_ror_assign_x_y    = 93,
     in_neg_assign_x      = 94,
     in_not_assign_x      = 95,
     in_gettypekind_x     = 96,
     in_faraddr_x         = 97,
     in_volatile_x        = 98,
     in_ismanagedtype_x   = 99,

{ Internal constant functions }
     in_const_sqr        = 100,
     in_const_abs        = 101,
     in_const_odd        = 102,
     in_const_ptr        = 103,
     in_const_swap_word  = 104,
     in_const_swap_long  = 105,
     in_lo_qword         = 106,
     in_hi_qword         = 107,
     in_const_swap_qword = 108,
     in_prefetch_var     = 109,
     in_const_eh_return_data_regno = 110,

{ FPU functions }
     in_trunc_real       = 120,
     in_round_real       = 121,
     in_frac_real        = 122,
     in_int_real         = 123,
     in_exp_real         = 124,
     in_cos_real         = 125,
     in_pi_real          = 126,
     in_abs_real         = 127,
     in_sqr_real         = 128,
     in_sqrt_real        = 129,
     in_arctan_real      = 130,
     in_ln_real          = 131,
     in_sin_real         = 132,
     in_fma_single       = 133,
     in_fma_double       = 134,
     in_fma_extended     = 135,
     in_fma_float128     = 136,

     { the min/max intrinsics must follow the x86 sse
       behaviour of min/max regarding handling
       NaN: in case of a NaN the result is always the second
       operand. This allows a simple translation of
       if a>b then result:=a else result:=b;
       statements into these intrinsics

       The min/max intrinsics are not supposed to
       be exposed to the user but only
       used internally by the compiler/optimizer }
     in_max_single       = 137,
     in_max_double       = 138,
     in_min_single       = 139,
     in_min_double       = 140,
     in_min_dword        = 141,
     in_min_longint      = 142,
     in_max_dword        = 143,
     in_max_longint      = 144,

{ MMX functions }
{ these contants are used by the mmx unit }

     { MMX }
     in_mmx_pcmpeqb      = 200,
     in_mmx_pcmpeqw      = 201,
     in_mmx_pcmpeqd      = 202,
     in_mmx_pcmpgtb      = 203,
     in_mmx_pcmpgtw      = 204,
     in_mmx_pcmpgtd      = 205,

     { 3DNow }

     { SSE }

{ More internal functions }
     in_isconstvalue_x    = 1000

     ,
  in_x86_inportb = in_cpu_first,
  in_x86_inportw = in_cpu_first+1,
  in_x86_inportl = in_cpu_first+2,
  in_x86_outportb = in_cpu_first+3,
  in_x86_outportw = in_cpu_first+4,
  in_x86_outportl = in_cpu_first+5,
  in_x86_cli      = in_cpu_first+6,
  in_x86_sti      = in_cpu_first+7,
  in_x86_get_cs   = in_cpu_first+8,
  in_x86_get_ss   = in_cpu_first+9,
  in_x86_get_ds   = in_cpu_first+10,
  in_x86_get_es   = in_cpu_first+11,
  in_x86_get_fs   = in_cpu_first+12,
  in_x86_get_gs   = in_cpu_first+13,

  { include mm inline routines }
  in_x86_movss = in_x86_mm_first+0,
  in_x86_movaps = in_x86_mm_first+1,
  in_x86_movups = in_x86_mm_first+2,
  in_x86_movss_to_mem = in_x86_mm_first+3,
  in_x86_movaps_to_mem = in_x86_mm_first+4,
  in_x86_movups_to_mem = in_x86_mm_first+5,
  in_x86_movss_to_val = in_x86_mm_first+6,
  in_x86_movss_from_val = in_x86_mm_first+7,
  in_x86_movlps = in_x86_mm_first+8,
  in_x86_movhps = in_x86_mm_first+9,
  in_x86_movlps_to_mem = in_x86_mm_first+10,
  in_x86_movhps_to_mem = in_x86_mm_first+11,
  in_x86_movlhps = in_x86_mm_first+12,
  in_x86_movhlps = in_x86_mm_first+13,
  in_x86_addss = in_x86_mm_first+14,
  in_x86_addss_from_mem = in_x86_mm_first+15,
  in_x86_subss = in_x86_mm_first+16,
  in_x86_subss_from_mem = in_x86_mm_first+17,
  in_x86_mulss = in_x86_mm_first+18,
  in_x86_mulss_from_mem = in_x86_mm_first+19,
  in_x86_divss = in_x86_mm_first+20,
  in_x86_divss_from_mem = in_x86_mm_first+21,
  in_x86_rcpss = in_x86_mm_first+22,
  in_x86_rcpss_from_mem = in_x86_mm_first+23,
  in_x86_sqrtss = in_x86_mm_first+24,
  in_x86_sqrtss_from_mem = in_x86_mm_first+25,
  in_x86_maxss = in_x86_mm_first+26,
  in_x86_maxss_from_mem = in_x86_mm_first+27,
  in_x86_minss = in_x86_mm_first+28,
  in_x86_minss_from_mem = in_x86_mm_first+29,
  in_x86_rsqrtss = in_x86_mm_first+30,
  in_x86_rsqrtss_from_mem = in_x86_mm_first+31,
  in_x86_addps = in_x86_mm_first+32,
  in_x86_addps_from_mem = in_x86_mm_first+33,
  in_x86_subps = in_x86_mm_first+34,
  in_x86_subps_from_mem = in_x86_mm_first+35,
  in_x86_mulps = in_x86_mm_first+36,
  in_x86_mulps_from_mem = in_x86_mm_first+37,
  in_x86_divps = in_x86_mm_first+38,
  in_x86_divps_from_mem = in_x86_mm_first+39,
  in_x86_rcpps = in_x86_mm_first+40,
  in_x86_rcpps_from_mem = in_x86_mm_first+41,
  in_x86_sqrtps = in_x86_mm_first+42,
  in_x86_sqrtps_from_mem = in_x86_mm_first+43,
  in_x86_maxps = in_x86_mm_first+44,
  in_x86_maxps_from_mem = in_x86_mm_first+45,
  in_x86_minps = in_x86_mm_first+46,
  in_x86_minps_from_mem = in_x86_mm_first+47,
  in_x86_rsqrtps = in_x86_mm_first+48,
  in_x86_rsqrtps_from_mem = in_x86_mm_first+49,
  in_x86_andps = in_x86_mm_first+50,
  in_x86_andps_from_mem = in_x86_mm_first+51,
  in_x86_orps = in_x86_mm_first+52,
  in_x86_orps_from_mem = in_x86_mm_first+53,
  in_x86_xorps = in_x86_mm_first+54,
  in_x86_xorps_from_mem = in_x86_mm_first+55,
  in_x86_andnps = in_x86_mm_first+56,
  in_x86_andnps_from_mem = in_x86_mm_first+57,
  in_x86_cmpss = in_x86_mm_first+58,
  in_x86_cmpss_from_mem = in_x86_mm_first+59,
  in_x86_cmpps = in_x86_mm_first+60,
  in_x86_cmpps_from_mem = in_x86_mm_first+61,
  in_x86_shufps = in_x86_mm_first+62,
  in_x86_shufps_from_mem = in_x86_mm_first+63,
  in_x86_unpckhps = in_x86_mm_first+64,
  in_x86_unpckhps_from_mem = in_x86_mm_first+65,
  in_x86_unpcklps = in_x86_mm_first+66,
  in_x86_unpcklps_from_mem = in_x86_mm_first+67,
  in_x86_cvtsi2ss = in_x86_mm_first+68,
  in_x86_cvtsi2ss_from_mem = in_x86_mm_first+69,
  in_x86_cvtss2si = in_x86_mm_first+70,
  in_x86_cvtss2si_from_mem = in_x86_mm_first+71,
  in_x86_cvttss2si = in_x86_mm_first+72,
  in_x86_cvttss2si_from_mem = in_x86_mm_first+73,
  in_x86_cvtpi2ps = in_x86_mm_first+74,
  in_x86_cvtpi2ps_from_mem = in_x86_mm_first+75,
  in_x86_cvtps2pi = in_x86_mm_first+76,
  in_x86_cvtps2pi_from_mem = in_x86_mm_first+77,
  in_x86_cvttps2pi = in_x86_mm_first+78,
  in_x86_cvttps2pi_from_mem = in_x86_mm_first+79,
  in_x86_pmulhuw_mmx = in_x86_mm_first+80,
  in_x86_pmulhuw_mmx_from_mem = in_x86_mm_first+81,
  in_x86_psadbw_mmx = in_x86_mm_first+82,
  in_x86_psadbw_mmx_from_mem = in_x86_mm_first+83,
  in_x86_pavgb_mmx = in_x86_mm_first+84,
  in_x86_pavgb_mmx_from_mem = in_x86_mm_first+85,
  in_x86_pavgw_mmx = in_x86_mm_first+86,
  in_x86_pavgw_mmx_from_mem = in_x86_mm_first+87,
  in_x86_pmaxub_mmx = in_x86_mm_first+88,
  in_x86_pmaxub_mmx_from_mem = in_x86_mm_first+89,
  in_x86_pminub_mmx = in_x86_mm_first+90,
  in_x86_pminub_mmx_from_mem = in_x86_mm_first+91,
  in_x86_pmaxsw_mmx = in_x86_mm_first+92,
  in_x86_pmaxsw_mmx_from_mem = in_x86_mm_first+93,
  in_x86_pminsw_mmx = in_x86_mm_first+94,
  in_x86_pminsw_mmx_from_mem = in_x86_mm_first+95,
  in_x86_pextrw_mmx = in_x86_mm_first+96,
  in_x86_pinsrw_mmx = in_x86_mm_first+97,
  in_x86_pmovmskb_mmx = in_x86_mm_first+98,
  in_x86_pshufw = in_x86_mm_first+99,
  in_x86_pshufw_from_mem = in_x86_mm_first+100,
  in_x86_movapd = in_x86_mm_first+101,
  in_x86_movapd_to_mem = in_x86_mm_first+102,
  in_x86_movntpd_to_mem = in_x86_mm_first+103,
  in_x86_movhpd = in_x86_mm_first+104,
  in_x86_movhpd_to_mem = in_x86_mm_first+105,
  in_x86_movlpd = in_x86_mm_first+106,
  in_x86_movlpd_to_mem = in_x86_mm_first+107,
  in_x86_movupd = in_x86_mm_first+108,
  in_x86_movupd_to_mem = in_x86_mm_first+109,
  in_x86_movmskpd = in_x86_mm_first+110,
  in_x86_movsd_from_mem = in_x86_mm_first+111,
  in_x86_movsd_to_mem = in_x86_mm_first+112,
  in_x86_movsd_to_val = in_x86_mm_first+113,
  in_x86_movsd_from_val = in_x86_mm_first+114,
  in_x86_addpd = in_x86_mm_first+115,
  in_x86_addpd_from_mem = in_x86_mm_first+116,
  in_x86_addsd = in_x86_mm_first+117,
  in_x86_addsd_from_mem = in_x86_mm_first+118,
  in_x86_divpd = in_x86_mm_first+119,
  in_x86_divpd_from_mem = in_x86_mm_first+120,
  in_x86_divsd = in_x86_mm_first+121,
  in_x86_divsd_from_mem = in_x86_mm_first+122,
  in_x86_maxpd = in_x86_mm_first+123,
  in_x86_maxpd_from_mem = in_x86_mm_first+124,
  in_x86_maxsd = in_x86_mm_first+125,
  in_x86_maxsd_from_mem = in_x86_mm_first+126,
  in_x86_minpd = in_x86_mm_first+127,
  in_x86_minpd_from_mem = in_x86_mm_first+128,
  in_x86_minsd = in_x86_mm_first+129,
  in_x86_minsd_from_mem = in_x86_mm_first+130,
  in_x86_mulpd = in_x86_mm_first+131,
  in_x86_mulpd_from_mem = in_x86_mm_first+132,
  in_x86_mulsd = in_x86_mm_first+133,
  in_x86_mulsd_from_mem = in_x86_mm_first+134,
  in_x86_sqrtpd = in_x86_mm_first+135,
  in_x86_sqrtpd_from_mem = in_x86_mm_first+136,
  in_x86_sqrtsd = in_x86_mm_first+137,
  in_x86_sqrtsd_from_mem = in_x86_mm_first+138,
  in_x86_subpd = in_x86_mm_first+139,
  in_x86_subpd_from_mem = in_x86_mm_first+140,
  in_x86_subsd = in_x86_mm_first+141,
  in_x86_subsd_from_mem = in_x86_mm_first+142,
  in_x86_andpd = in_x86_mm_first+143,
  in_x86_andpd_from_mem = in_x86_mm_first+144,
  in_x86_andnpd = in_x86_mm_first+145,
  in_x86_andnpd_from_mem = in_x86_mm_first+146,
  in_x86_orpd = in_x86_mm_first+147,
  in_x86_orpd_from_mem = in_x86_mm_first+148,
  in_x86_xorpd = in_x86_mm_first+149,
  in_x86_xorpd_from_mem = in_x86_mm_first+150,
  in_x86_cmppd = in_x86_mm_first+151,
  in_x86_cmppd_from_mem = in_x86_mm_first+152,
  in_x86_cmpsd = in_x86_mm_first+153,
  in_x86_cmpsd_from_mem = in_x86_mm_first+154,
  in_x86_comisd = in_x86_mm_first+155,
  in_x86_comisd_from_mem = in_x86_mm_first+156,
  in_x86_ucomisd = in_x86_mm_first+157,
  in_x86_ucomisd_from_mem = in_x86_mm_first+158,
  in_x86_shufpd = in_x86_mm_first+159,
  in_x86_shufpd_from_mem = in_x86_mm_first+160,
  in_x86_unpckhpd = in_x86_mm_first+161,
  in_x86_unpckhpd_from_mem = in_x86_mm_first+162,
  in_x86_unpcklpd = in_x86_mm_first+163,
  in_x86_unpcklpd_from_mem = in_x86_mm_first+164,
  in_x86_cvtdq2pd = in_x86_mm_first+165,
  in_x86_cvtdq2pd_from_mem = in_x86_mm_first+166,
  in_x86_cvtdq2ps = in_x86_mm_first+167,
  in_x86_cvtdq2ps_from_mem = in_x86_mm_first+168,
  in_x86_cvtpd2dq = in_x86_mm_first+169,
  in_x86_cvtpd2dq_from_mem = in_x86_mm_first+170,
  in_x86_cvtpd2pi = in_x86_mm_first+171,
  in_x86_cvtpd2pi_from_mem = in_x86_mm_first+172,
  in_x86_cvtpd2ps = in_x86_mm_first+173,
  in_x86_cvtpd2ps_from_mem = in_x86_mm_first+174,
  in_x86_cvtpi2pd = in_x86_mm_first+175,
  in_x86_cvtpi2pd_from_mem = in_x86_mm_first+176,
  in_x86_cvtps2dq = in_x86_mm_first+177,
  in_x86_cvtps2dq_from_mem = in_x86_mm_first+178,
  in_x86_cvtps2pd = in_x86_mm_first+179,
  in_x86_cvtps2pd_from_mem = in_x86_mm_first+180,
  in_x86_cvtsd2si = in_x86_mm_first+181,
  in_x86_cvtsd2si_from_mem = in_x86_mm_first+182,
  in_x86_cvtsd2ss = in_x86_mm_first+183,
  in_x86_cvtsd2ss_from_mem = in_x86_mm_first+184,
  in_x86_cvtsi2sd = in_x86_mm_first+185,
  in_x86_cvtsi2sd_from_mem = in_x86_mm_first+186,
  in_x86_cvtss2sd = in_x86_mm_first+187,
  in_x86_cvtss2sd_from_mem = in_x86_mm_first+188,
  in_x86_cvttpd2dq = in_x86_mm_first+189,
  in_x86_cvttpd2dq_from_mem = in_x86_mm_first+190,
  in_x86_cvttpd2pi = in_x86_mm_first+191,
  in_x86_cvttpd2pi_from_mem = in_x86_mm_first+192,
  in_x86_cvttps2dq = in_x86_mm_first+193,
  in_x86_cvttps2dq_from_mem = in_x86_mm_first+194,
  in_x86_cvttsd2si = in_x86_mm_first+195,
  in_x86_cvttsd2si_from_mem = in_x86_mm_first+196,
  in_x86_movd_from_reg = in_x86_mm_first+197,
  in_x86_movd_from_mem = in_x86_mm_first+198,
  in_x86_movd_to_reg = in_x86_mm_first+199,
  in_x86_movd_to_mem = in_x86_mm_first+200,
  in_x86_movq_from_mem = in_x86_mm_first+201,
  in_x86_movq_to_mem = in_x86_mm_first+202,
  in_x86_pmovmskb = in_x86_mm_first+203,
  in_x86_pextrw_sse2 = in_x86_mm_first+204,
  in_x86_pinsrw_sse2 = in_x86_mm_first+205,
  in_x86_pinsrw_from_mem = in_x86_mm_first+206,
  in_x86_packssdw = in_x86_mm_first+207,
  in_x86_packssdw_from_mem = in_x86_mm_first+208,
  in_x86_packsswb = in_x86_mm_first+209,
  in_x86_packsswb_from_mem = in_x86_mm_first+210,
  in_x86_packuswb = in_x86_mm_first+211,
  in_x86_packuswb_from_mem = in_x86_mm_first+212,
  in_x86_paddb = in_x86_mm_first+213,
  in_x86_paddb_from_mem = in_x86_mm_first+214,
  in_x86_paddw = in_x86_mm_first+215,
  in_x86_paddw_from_mem = in_x86_mm_first+216,
  in_x86_paddd = in_x86_mm_first+217,
  in_x86_paddd_from_mem = in_x86_mm_first+218,
  in_x86_paddq = in_x86_mm_first+219,
  in_x86_paddq_from_mem = in_x86_mm_first+220,
  in_x86_paddsb = in_x86_mm_first+221,
  in_x86_paddsb_from_mem = in_x86_mm_first+222,
  in_x86_paddsw = in_x86_mm_first+223,
  in_x86_paddsw_from_mem = in_x86_mm_first+224,
  in_x86_paddusb = in_x86_mm_first+225,
  in_x86_paddusb_from_mem = in_x86_mm_first+226,
  in_x86_paddusw = in_x86_mm_first+227,
  in_x86_paddusw_from_mem = in_x86_mm_first+228,
  in_x86_pand = in_x86_mm_first+229,
  in_x86_pand_from_mem = in_x86_mm_first+230,
  in_x86_pandn = in_x86_mm_first+231,
  in_x86_pandn_from_mem = in_x86_mm_first+232,
  in_x86_por = in_x86_mm_first+233,
  in_x86_por_from_mem = in_x86_mm_first+234,
  in_x86_pxor = in_x86_mm_first+235,
  in_x86_pxor_from_mem = in_x86_mm_first+236,
  in_x86_pcmpeqb = in_x86_mm_first+237,
  in_x86_pcmpeqb_from_mem = in_x86_mm_first+238,
  in_x86_pcmpeqw = in_x86_mm_first+239,
  in_x86_pcmpeqw_from_mem = in_x86_mm_first+240,
  in_x86_pcmpeqd = in_x86_mm_first+241,
  in_x86_pcmpeqd_from_mem = in_x86_mm_first+242,
  in_x86_pcmpgtb = in_x86_mm_first+243,
  in_x86_pcmpgtb_from_mem = in_x86_mm_first+244,
  in_x86_pcmpgtw = in_x86_mm_first+245,
  in_x86_pcmpgtw_from_mem = in_x86_mm_first+246,
  in_x86_pcmpgtd = in_x86_mm_first+247,
  in_x86_pcmpgtd_from_mem = in_x86_mm_first+248,
  in_x86_pmullw = in_x86_mm_first+249,
  in_x86_pmullw_from_mem = in_x86_mm_first+250,
  in_x86_pmulhw = in_x86_mm_first+251,
  in_x86_pmulhw_from_mem = in_x86_mm_first+252,
  in_x86_pmulhuw_sse2 = in_x86_mm_first+253,
  in_x86_pmulhuw_from_mem = in_x86_mm_first+254,
  in_x86_pmuludq = in_x86_mm_first+255,
  in_x86_pmuludq_from_mem = in_x86_mm_first+256,
  in_x86_psllw_sse2 = in_x86_mm_first+257,
  in_x86_psllw_from_mem = in_x86_mm_first+258,
  in_x86_psllw_sse2_imm = in_x86_mm_first+259,
  in_x86_pslld_sse2 = in_x86_mm_first+260,
  in_x86_pslld_from_mem = in_x86_mm_first+261,
  in_x86_pslld_sse2_imm = in_x86_mm_first+262,
  in_x86_psllq_sse2 = in_x86_mm_first+263,
  in_x86_psllq_from_mem = in_x86_mm_first+264,
  in_x86_psllq_sse2_imm = in_x86_mm_first+265,
  in_x86_psrad_sse2 = in_x86_mm_first+266,
  in_x86_psrad_from_mem = in_x86_mm_first+267,
  in_x86_psrad_sse2_imm = in_x86_mm_first+268,
  in_x86_psraw_sse2 = in_x86_mm_first+269,
  in_x86_psraw_from_mem = in_x86_mm_first+270,
  in_x86_psraw_sse2_imm = in_x86_mm_first+271,
  in_x86_psrlw_sse2 = in_x86_mm_first+272,
  in_x86_psrlw_from_mem = in_x86_mm_first+273,
  in_x86_psrlw_sse2_imm = in_x86_mm_first+274,
  in_x86_psrld_sse2 = in_x86_mm_first+275,
  in_x86_psrld_from_mem = in_x86_mm_first+276,
  in_x86_psrld_sse2_imm = in_x86_mm_first+277,
  in_x86_psrlq_sse2 = in_x86_mm_first+278,
  in_x86_psrlq_from_mem = in_x86_mm_first+279,
  in_x86_psrlq_sse2_imm = in_x86_mm_first+280,
  in_x86_psubb = in_x86_mm_first+281,
  in_x86_psubb_from_mem = in_x86_mm_first+282,
  in_x86_psubw = in_x86_mm_first+283,
  in_x86_psubw_from_mem = in_x86_mm_first+284,
  in_x86_psubd = in_x86_mm_first+285,
  in_x86_psubd_from_mem = in_x86_mm_first+286,
  in_x86_psubq = in_x86_mm_first+287,
  in_x86_psubq_from_mem = in_x86_mm_first+288,
  in_x86_psubsb = in_x86_mm_first+289,
  in_x86_psubsb_from_mem = in_x86_mm_first+290,
  in_x86_psubsw = in_x86_mm_first+291,
  in_x86_psubsw_from_mem = in_x86_mm_first+292,
  in_x86_pmaddwd = in_x86_mm_first+293,
  in_x86_pmaddwd_from_mem = in_x86_mm_first+294,
  in_x86_psubusb = in_x86_mm_first+295,
  in_x86_psubusb_from_mem = in_x86_mm_first+296,
  in_x86_psubusw = in_x86_mm_first+297,
  in_x86_psubusw_from_mem = in_x86_mm_first+298,
  in_x86_punpckhbw = in_x86_mm_first+299,
  in_x86_punpckhbw_from_mem = in_x86_mm_first+300,
  in_x86_punpckhwd = in_x86_mm_first+301,
  in_x86_punpckhwd_from_mem = in_x86_mm_first+302,
  in_x86_punpckhdq = in_x86_mm_first+303,
  in_x86_punpckhdq_from_mem = in_x86_mm_first+304,
  in_x86_punpcklbw = in_x86_mm_first+305,
  in_x86_punpcklbw_from_mem = in_x86_mm_first+306,
  in_x86_punpcklwd = in_x86_mm_first+307,
  in_x86_punpcklwd_from_mem = in_x86_mm_first+308,
  in_x86_punpckldq = in_x86_mm_first+309,
  in_x86_punpckldq_from_mem = in_x86_mm_first+310,
  in_x86_pavgb_sse2 = in_x86_mm_first+311,
  in_x86_pavgb_from_mem = in_x86_mm_first+312,
  in_x86_pavgw_sse2 = in_x86_mm_first+313,
  in_x86_pavgw_from_mem = in_x86_mm_first+314,
  in_x86_pminub_sse2 = in_x86_mm_first+315,
  in_x86_pminub_from_mem = in_x86_mm_first+316,
  in_x86_pminsw_sse2 = in_x86_mm_first+317,
  in_x86_pminsw_from_mem = in_x86_mm_first+318,
  in_x86_pmaxsw_sse2 = in_x86_mm_first+319,
  in_x86_pmaxsw_from_mem = in_x86_mm_first+320,
  in_x86_pmaxub_sse2 = in_x86_mm_first+321,
  in_x86_pmaxub_from_mem = in_x86_mm_first+322,
  in_x86_psadbw_sse2 = in_x86_mm_first+323,
  in_x86_psadbw_from_mem = in_x86_mm_first+324,
  in_x86_maskmovdqu = in_x86_mm_first+325,
  in_x86_movdq2q = in_x86_mm_first+326,
  in_x86_movdqa_from_mem = in_x86_mm_first+327,
  in_x86_movdqa = in_x86_mm_first+328,
  in_x86_movdqu_from_mem = in_x86_mm_first+329,
  in_x86_movdqu = in_x86_mm_first+330,
  in_x86_movq2dq = in_x86_mm_first+331,
  in_x86_movntdq = in_x86_mm_first+332,
  in_x86_pshufhw = in_x86_mm_first+333,
  in_x86_pshuflw = in_x86_mm_first+334,
  in_x86_pshufd = in_x86_mm_first+335,
  in_x86_pshufhw_from_mem = in_x86_mm_first+336,
  in_x86_pshuflw_from_mem = in_x86_mm_first+337,
  in_x86_pshufd_from_mem = in_x86_mm_first+338,
  in_x86_pslldq = in_x86_mm_first+339,
  in_x86_psrldq = in_x86_mm_first+340,
  in_x86_punpckhqdq = in_x86_mm_first+341,
  in_x86_punpckhqdq_from_mem = in_x86_mm_first+342,
  in_x86_punpcklqdq = in_x86_mm_first+343,
  in_x86_punpcklqdq_from_mem = in_x86_mm_first+344,
  in_x86_addsubps = in_x86_mm_first+345,
  in_x86_addsubps_from_mem = in_x86_mm_first+346,
  in_x86_addsubpd = in_x86_mm_first+347,
  in_x86_addsubpd_from_mem = in_x86_mm_first+348,
  in_x86_movddup = in_x86_mm_first+349,
  in_x86_movddup_from_mem = in_x86_mm_first+350,
  in_x86_movsldup = in_x86_mm_first+351,
  in_x86_movsldup_from_mem = in_x86_mm_first+352,
  in_x86_movshdup = in_x86_mm_first+353,
  in_x86_movshdup_from_mem = in_x86_mm_first+354,
  in_x86_haddps = in_x86_mm_first+355,
  in_x86_haddps_from_mem = in_x86_mm_first+356,
  in_x86_haddpd = in_x86_mm_first+357,
  in_x86_haddpd_from_mem = in_x86_mm_first+358,
  in_x86_hsubps = in_x86_mm_first+359,
  in_x86_hsubps_from_mem = in_x86_mm_first+360,
  in_x86_hsubpd = in_x86_mm_first+361,
  in_x86_hsubpd_from_mem = in_x86_mm_first+362,
  in_x86_lddqu = in_x86_mm_first+363,
  in_x86_psignb = in_x86_mm_first+364,
  in_x86_psignb_from_mem = in_x86_mm_first+365,
  in_x86_psignw = in_x86_mm_first+366,
  in_x86_psignw_from_mem = in_x86_mm_first+367,
  in_x86_psignd = in_x86_mm_first+368,
  in_x86_psignd_from_mem = in_x86_mm_first+369,
  in_x86_pshufb = in_x86_mm_first+370,
  in_x86_pshufb_from_mem = in_x86_mm_first+371,
  in_x86_pmulhrsw = in_x86_mm_first+372,
  in_x86_pmulhrsw_from_mem = in_x86_mm_first+373,
  in_x86_pmaddubsw = in_x86_mm_first+374,
  in_x86_pmaddubsw_from_mem = in_x86_mm_first+375,
  in_x86_phsubw = in_x86_mm_first+376,
  in_x86_phsubw_from_mem = in_x86_mm_first+377,
  in_x86_phsubsw = in_x86_mm_first+378,
  in_x86_phsubsw_from_mem = in_x86_mm_first+379,
  in_x86_phsubd = in_x86_mm_first+380,
  in_x86_phsubd_from_mem = in_x86_mm_first+381,
  in_x86_phaddsw = in_x86_mm_first+382,
  in_x86_phaddsw_from_mem = in_x86_mm_first+383,
  in_x86_phaddw = in_x86_mm_first+384,
  in_x86_phaddw_from_mem = in_x86_mm_first+385,
  in_x86_phaddd = in_x86_mm_first+386,
  in_x86_phaddd_from_mem = in_x86_mm_first+387,
  in_x86_palignr = in_x86_mm_first+388,
  in_x86_palignr_from_mem = in_x86_mm_first+389,
  in_x86_pabsb = in_x86_mm_first+390,
  in_x86_pabsb_from_mem = in_x86_mm_first+391,
  in_x86_pabsw = in_x86_mm_first+392,
  in_x86_pabsw_from_mem = in_x86_mm_first+393,
  in_x86_pabsd = in_x86_mm_first+394,
  in_x86_pabsd_from_mem = in_x86_mm_first+395,
  in_x86_dpps = in_x86_mm_first+396,
  in_x86_dpps_from_mem = in_x86_mm_first+397,
  in_x86_dppd = in_x86_mm_first+398,
  in_x86_dppd_from_mem = in_x86_mm_first+399,
  in_x86_blendps = in_x86_mm_first+400,
  in_x86_blendps_from_mem = in_x86_mm_first+401,
  in_x86_blendvps = in_x86_mm_first+402,
  in_x86_blendvps_from_mem = in_x86_mm_first+403,
  in_x86_blendpd = in_x86_mm_first+404,
  in_x86_blendpd_from_mem = in_x86_mm_first+405,
  in_x86_blendvpd = in_x86_mm_first+406,
  in_x86_blendvpd_from_mem = in_x86_mm_first+407,
  in_x86_roundps = in_x86_mm_first+408,
  in_x86_roundps_from_mem = in_x86_mm_first+409,
  in_x86_roundss = in_x86_mm_first+410,
  in_x86_roundss_from_mem = in_x86_mm_first+411,
  in_x86_roundpd = in_x86_mm_first+412,
  in_x86_roundpd_from_mem = in_x86_mm_first+413,
  in_x86_roundsd = in_x86_mm_first+414,
  in_x86_roundsd_from_mem = in_x86_mm_first+415,
  in_x86_insertps = in_x86_mm_first+416,
  in_x86_insertps_from_mem = in_x86_mm_first+417,
  in_x86_extractps = in_x86_mm_first+418,
  in_x86_extractps_from_mem = in_x86_mm_first+419,
  in_x86_mpsadbw = in_x86_mm_first+420,
  in_x86_mpsadbw_from_mem = in_x86_mm_first+421,
  in_x86_phminposuw = in_x86_mm_first+422,
  in_x86_phminposuw_from_mem = in_x86_mm_first+423,
  in_x86_pmulld = in_x86_mm_first+424,
  in_x86_pmulld_from_mem = in_x86_mm_first+425,
  in_x86_pmuldq = in_x86_mm_first+426,
  in_x86_pmuldq_from_mem = in_x86_mm_first+427,
  in_x86_pblendvb = in_x86_mm_first+428,
  in_x86_pblendvb_from_mem = in_x86_mm_first+429,
  in_x86_pblendw = in_x86_mm_first+430,
  in_x86_pblendw_from_mem = in_x86_mm_first+431,
  in_x86_pminsb = in_x86_mm_first+432,
  in_x86_pminsb_from_mem = in_x86_mm_first+433,
  in_x86_pminuw = in_x86_mm_first+434,
  in_x86_pminuw_from_mem = in_x86_mm_first+435,
  in_x86_pminsd = in_x86_mm_first+436,
  in_x86_pminsd_from_mem = in_x86_mm_first+437,
  in_x86_pminud = in_x86_mm_first+438,
  in_x86_pminud_from_mem = in_x86_mm_first+439,
  in_x86_pmaxsb = in_x86_mm_first+440,
  in_x86_pmaxsb_from_mem = in_x86_mm_first+441,
  in_x86_pmaxuw = in_x86_mm_first+442,
  in_x86_pmaxuw_from_mem = in_x86_mm_first+443,
  in_x86_pmaxsd = in_x86_mm_first+444,
  in_x86_pmaxsd_from_mem = in_x86_mm_first+445,
  in_x86_pmaxud = in_x86_mm_first+446,
  in_x86_pmaxud_from_mem = in_x86_mm_first+447,
  in_x86_pinsrb = in_x86_mm_first+448,
  in_x86_pinsrb_from_mem = in_x86_mm_first+449,
  in_x86_pinsrd = in_x86_mm_first+450,
  in_x86_pinsrd_from_mem = in_x86_mm_first+451,
  in_x86_pinsrq = in_x86_mm_first+452,
  in_x86_pinsrq_from_mem = in_x86_mm_first+453,
  in_x86_pextrb = in_x86_mm_first+454,
  in_x86_pextrb_to_mem = in_x86_mm_first+455,
  in_x86_pextrw_sse41_to_mem = in_x86_mm_first+456,
  in_x86_pextrd = in_x86_mm_first+457,
  in_x86_pextrd_to_mem = in_x86_mm_first+458,
  in_x86_pextrq = in_x86_mm_first+459,
  in_x86_pextrq_to_mem = in_x86_mm_first+460,
  in_x86_pmovsxbw = in_x86_mm_first+461,
  in_x86_pmovsxbw_from_mem = in_x86_mm_first+462,
  in_x86_pmovzxbw = in_x86_mm_first+463,
  in_x86_pmovzxbw_from_mem = in_x86_mm_first+464,
  in_x86_pmovsxbd = in_x86_mm_first+465,
  in_x86_pmovsxbd_from_mem = in_x86_mm_first+466,
  in_x86_pmovzxbd = in_x86_mm_first+467,
  in_x86_pmovzxbd_from_mem = in_x86_mm_first+468,
  in_x86_pmovsxbq = in_x86_mm_first+469,
  in_x86_pmovsxbq_from_mem = in_x86_mm_first+470,
  in_x86_pmovzxbq = in_x86_mm_first+471,
  in_x86_pmovzxbq_from_mem = in_x86_mm_first+472,
  in_x86_pmovsxwd = in_x86_mm_first+473,
  in_x86_pmovsxwd_from_mem = in_x86_mm_first+474,
  in_x86_pmovzxwd = in_x86_mm_first+475,
  in_x86_pmovzxwd_from_mem = in_x86_mm_first+476,
  in_x86_pmovsxwq = in_x86_mm_first+477,
  in_x86_pmovsxwq_from_mem = in_x86_mm_first+478,
  in_x86_pmovzxwq = in_x86_mm_first+479,
  in_x86_pmovzxwq_from_mem = in_x86_mm_first+480,
  in_x86_pmovsxdq = in_x86_mm_first+481,
  in_x86_pmovsxdq_from_mem = in_x86_mm_first+482,
  in_x86_pmovzxdq = in_x86_mm_first+483,
  in_x86_pmovzxdq_from_mem = in_x86_mm_first+484,
  in_x86_ptest = in_x86_mm_first+485,
  in_x86_ptest_from_mem = in_x86_mm_first+486,
  in_x86_pcmpeqq = in_x86_mm_first+487,
  in_x86_pcmpeqq_from_mem = in_x86_mm_first+488,
  in_x86_packusdw = in_x86_mm_first+489,
  in_x86_packusdw_from_mem = in_x86_mm_first+490,
  in_x86_movntdqa = in_x86_mm_first+491,
  in_x86_pcmpestri = in_x86_mm_first+492,
  in_x86_pcmpestri_from_mem = in_x86_mm_first+493,
  in_x86_pcmpestrm = in_x86_mm_first+494,
  in_x86_pcmpestrm_from_mem = in_x86_mm_first+495,
  in_x86_pcmpistri = in_x86_mm_first+496,
  in_x86_pcmpistri_from_mem = in_x86_mm_first+497,
  in_x86_pcmpistrm = in_x86_mm_first+498,
  in_x86_pcmpistrm_from_mem = in_x86_mm_first+499,
  in_x86_pcmpgtq = in_x86_mm_first+500,
  in_x86_pcmpgtq_from_mem = in_x86_mm_first+501
   );

implementation
end.
