{
    Copyright (c) 1998-2002 by Peter Vreman

    This unit handles the writing of script files

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 ****************************************************************************
}
unit cscript;

{$i fpcdefs.inc}

interface
{$H+}
uses
  sysutils,
  globtype,
  cclasses;

type
  TLinkRes = Class
    fn   : TCmdStr;
    data : TCmdStrList;
    executable : boolean;
    section: string[30];
    fRealResponseFile: Boolean;
    fForceUseForwardSlash: Boolean;
    constructor Create(const ScriptName : TCmdStr; RealResponseFile: Boolean);
    destructor Destroy;override;
    procedure AddStart(const s:TCmdStr);
    procedure Add(const s:TCmdStr);
    procedure AddFileName(const s:TCmdStr);
    procedure EndSection(const s:TCmdStr);
    procedure StartSection(const s:TCmdStr);
    Function  Empty:boolean;
    procedure WriteToDisk;virtual;
  end;

Function ScriptFixFileName(const s:TCmdStr):TCmdStr;

implementation

uses
  BaseUnix,
  cutils,cfileutl,
  globals,systems,verbose;

    Function ScriptFixFileName(const s:TCmdStr):TCmdStr;
     begin
       if cs_link_on_target in current_settings.globalswitches then
         ScriptFixFileName:=TargetFixFileName(s)
       else
         ScriptFixFileName:=FixFileName(s);
     end;

constructor TLinkRes.Create(const ScriptName: TCmdStr; RealResponseFile: Boolean);
begin
  fn:=FixFileName(scriptname);
  executable:=false;
  data:=TCmdStrList.Create;
  fRealResponseFile:=RealResponseFile;
  fForceUseForwardSlash:=false;
end;

destructor TLinkRes.Destroy;
begin
  data.Free;
end;

procedure TLinkRes.AddStart(const s:TCmdStr);
begin
  data.Insert(s);
end;

Function TLinkRes.Empty:boolean;
begin
  Empty:=Data.Empty;
end;

procedure TLinkRes.WriteToDisk;
var
  t : file;
  i : longint;
  s : TCmdStr;
  le: string[2];
begin
  Assign(t,fn);
    le:= #10;
  {$push}{$I-}
  Rewrite(t,1);
  if ioresult<>0 then
    exit;
  while not data.Empty do
    begin
      s:=data.GetFirst;
      Blockwrite(t,s[1],length(s),i);
      Blockwrite(t,le[1],length(le),i);
    end;
  Close(t);
  {$pop}
  i:=ioresult;
  if executable then
   fpchmod(fn,493);
end;

procedure TLinkRes.Add(const s:TCmdStr);
begin
  if s<>'' then
   data.Concat(s);
end;

procedure TLinkRes.AddFileName(const s:TCmdStr);
begin
  if section<>'' then
   begin
    data.Concat(section);
    section:='';
   end;
  if s<>'' then
   begin
     if fRealResponseFile and
        (s[1]='''') then
       data.Concat(UnixRequoteWithDoubleQuotes(s))
     else if not(s[1] in ['a'..'z','A'..'Z','/','\','.','"']) then
      begin
          data.Concat('./'+s);
      end
     else
      data.Concat(s);
   end;
end;

procedure TLinkRes.EndSection(const s:TCmdStr);
begin
  if section='' then
    data.Concat(s);
  section:='';
end;

procedure TLinkRes.StartSection(const s:TCmdStr);
begin
  section:=s;
end;

end.
