﻿{
    Copyright (c) 1998-2002 by Florian Klaempfl and Peter Vreman

    Contains the abstract assembler implementation for the i386

    * Portions of this code was inspired by the NASM sources
      The Netwide Assembler is Copyright (c) 1996 Simon Tatham and
      Julian Hall. All rights reserved.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 ****************************************************************************
}
unit aasmcpu;

{$i fpcdefs.inc}

interface

    uses
      globtype,verbose,
      cpubase,
      cgbase,cgutils,
      aasmbase,aasmtai,aasmsym,
      ogbase;

    const
      { "mov reg,reg" source operand number }
      O_MOV_SOURCE = 0;
      { "mov reg,reg" destination operand number }
      O_MOV_DEST = 1;

    { Operand types }
      OT_NONE      = $00000000;

      { Bits 0..7: sizes }
      OT_BITS8     = $00000001;
      OT_BITS16    = $00000002;
      OT_BITS32    = $00000004;
      OT_BITS64    = $00000008;  { x86_64 and FPU }
      OT_BITS128   = $20000000;  { 16 byte SSE }
      OT_BITS256   = $40000000;  { 32 byte AVX }
      OT_BITS512   = $80000000;  { 64 byte AVX512 }

      OT_VECTORMASK = $1000000000;  { OPTIONAL VECTORMASK AVX512}
      OT_VECTORZERO = $2000000000;  { OPTIONAL ZERO-FLAG  AVX512}
      OT_VECTORBCST = $4000000000;  { BROADCAST-MEM-FLAG  AVX512}
      OT_VECTORSAE  = $8000000000;  { OPTIONAL SAE-FLAG  AVX512}
      OT_VECTORER   = $10000000000; { OPTIONAL ER-FLAG-FLAG  AVX512}

      OT_VECTOR_EXT = OT_VECTORMASK or OT_VECTORZERO or OT_VECTORBCST or OT_VECTORSAE or OT_VECTORER;

      OT_BITSB32    = OT_BITS32 or OT_VECTORBCST;
      OT_BITSB64    = OT_BITS64 or OT_VECTORBCST;

      OT_BITS80    = $00000010;  { FPU only  }
      OT_FAR       = $00000020;  { this means 16:16 or 16:32, like in CALL/JMP }
      OT_NEAR      = $00000040;
      OT_SHORT     = $00000080;

      { TODO: FAR/NEAR/SHORT are sizes too, they should be included into size mask,
        but this requires adjusting the opcode table }
      OT_SIZE_MASK = $E000001F;  { all the size attributes  }
      OT_NON_SIZE  = longint(not(longint(OT_SIZE_MASK)));

      { Bits 8..11: modifiers }
      OT_SIGNED    = $00000100;  { the operand need to be signed -128-127 }
      OT_TO        = $00000200;  { reverse effect in FADD, FSUB &c  }
      OT_COLON     = $00000400;  { operand is followed by a colon  }
      OT_MODIFIER_MASK = $00000F00;

      { Bits 12..15: type of operand }
      OT_REGISTER  = $00001000;
      OT_IMMEDIATE = $00002000;
      OT_MEMORY    = $0000C000;  { always includes 'OT_REGMEM' bit as well }
      OT_REGMEM    = $00008000;  { for r/m, ie EA, operands  }
      OT_TYPE_MASK = OT_REGISTER or OT_IMMEDIATE or OT_MEMORY or OT_REGMEM;

      OT_REGNORM   = OT_REGISTER or OT_REGMEM;  { 'normal' reg, qualifies as EA  }

      { Bits 20..22, 24..26: register classes
        otf_* consts are not used alone, only to build other constants. }
      otf_reg_cdt  = $00100000;
      otf_reg_gpr  = $00200000;
      otf_reg_sreg = $00400000;
      otf_reg_k    = $00800000;
      otf_reg_fpu  = $01000000;
      otf_reg_mmx  = $02000000;
      otf_reg_xmm  = $04000000;
      otf_reg_ymm  = $08000000;
      otf_reg_zmm  = $10000000;


      otf_reg_extra_mask = $0F000000;
      { Bits 16..19: subclasses, meaning depends on classes field }
      otf_sub0     = $00010000;
      otf_sub1     = $00020000;
      otf_sub2     = $00040000;
      otf_sub3     = $00080000;
      OT_REG_SMASK = otf_sub0 or otf_sub1 or otf_sub2 or otf_sub3;

      //OT_REG_EXTRA_MASK = $0F000000;
      OT_REG_EXTRA_MASK = $1F000000;

      OT_REG_TYPMASK = otf_reg_cdt or otf_reg_gpr or otf_reg_sreg or otf_reg_k or otf_reg_extra_mask;
      { register class 0: CRx, DRx and TRx }
      OT_REG_CDT   = OT_REGISTER or otf_reg_cdt or OT_BITS64;
      OT_REG_CREG  = OT_REG_CDT or otf_sub0;  { CRn  }
      OT_REG_DREG  = OT_REG_CDT or otf_sub1;  { DRn  }
      OT_REG_TREG  = OT_REG_CDT or otf_sub2;  { TRn  }
      OT_REG_CR4   = OT_REG_CDT or otf_sub3;  { CR4 (Pentium only)  }

      { register class 1: general-purpose registers }
      OT_REG_GPR   = OT_REGNORM or otf_reg_gpr;
      OT_RM_GPR    = OT_REGMEM or otf_reg_gpr;
      OT_REG8      = OT_REG_GPR or OT_BITS8;  { 8-bit GPR }
      OT_REG16     = OT_REG_GPR or OT_BITS16;
      OT_REG32     = OT_REG_GPR or OT_BITS32;
      OT_REG64     = OT_REG_GPR or OT_BITS64;

      { GPR subclass 0: accumulator: AL, AX, EAX or RAX }
      OT_REG_ACCUM = OT_REG_GPR or otf_sub0;
      OT_REG_AL    = OT_REG_ACCUM or OT_BITS8;
      OT_REG_AX    = OT_REG_ACCUM or OT_BITS16;
      OT_REG_EAX   = OT_REG_ACCUM or OT_BITS32;
      OT_REG_RAX   = OT_REG_ACCUM or OT_BITS64;
      { GPR subclass 1: counter: CL, CX, ECX or RCX }
      OT_REG_COUNT = OT_REG_GPR or otf_sub1;
      OT_REG_CL    = OT_REG_COUNT or OT_BITS8;
      OT_REG_CX    = OT_REG_COUNT or OT_BITS16;
      OT_REG_ECX   = OT_REG_COUNT or OT_BITS32;
      OT_REG_RCX   = OT_REG_COUNT or OT_BITS64;
      { GPR subclass 2: data register: DL, DX, EDX or RDX }
      OT_REG_DX    = OT_REG_GPR or otf_sub2 or OT_BITS16;
      OT_REG_EDX   = OT_REG_GPR or otf_sub2 or OT_BITS32;

      { register class 2: Segment registers }
      OT_REG_SREG  = OT_REGISTER or otf_reg_sreg or OT_BITS16;
      OT_REG_CS    = OT_REG_SREG or otf_sub0;  { CS  }
      OT_REG_DESS  = OT_REG_SREG or otf_sub1;  { DS, ES, SS (non-CS 86 registers)  }
      OT_REG_FSGS  = OT_REG_SREG or otf_sub2;  { FS, GS (386 extended registers)  }

      { register class 3: FPU registers }
      OT_FPUREG    = OT_REGISTER or otf_reg_fpu;
      OT_FPU0      = OT_FPUREG or otf_sub0;    { FPU stack register zero  }

      { register class 4: MMX (both reg and r/m) }
      OT_MMXREG    = OT_REGNORM or otf_reg_mmx;
      OT_MMXRM     = OT_REGMEM or otf_reg_mmx;

      { register class 5: XMM (both reg and r/m) }
      OT_XMMREG    = OT_REGNORM or otf_reg_xmm;
      OT_XMMRM     = OT_REGMEM or otf_reg_xmm;
      OT_XMEM32    = OT_REGNORM or otf_reg_xmm or otf_reg_gpr or OT_BITS32;
      OT_XMEM32_M  = OT_XMEM32 or OT_VECTORMASK;
      OT_XMEM64    = OT_REGNORM or otf_reg_xmm or otf_reg_gpr or OT_BITS64;
      OT_XMEM64_M  = OT_XMEM64 or OT_VECTORMASK;

      OT_XMMREG_M   = OT_XMMREG or OT_VECTORMASK;
      OT_XMMREG_MZ  = OT_XMMREG or OT_VECTORMASK or OT_VECTORZERO;
      OT_XMMRM_MZ   = OT_XMMRM  or OT_VECTORMASK or OT_VECTORZERO;
      OT_XMMREG_SAE = OT_XMMREG or OT_VECTORSAE;
      OT_XMMRM_SAE  = OT_XMMRM  or OT_VECTORSAE;
      OT_XMMREG_ER  = OT_XMMREG or OT_VECTORER;
      OT_XMMRM_ER   = OT_XMMRM  or OT_VECTORER;



      { register class 5: YMM (both reg and r/m) }
      OT_YMMREG     = OT_REGNORM or otf_reg_ymm;
      OT_YMMRM      = OT_REGMEM or otf_reg_ymm;
      OT_YMEM32     = OT_REGNORM or otf_reg_ymm or otf_reg_gpr or OT_BITS32;
      OT_YMEM32_M   = OT_YMEM32 or OT_VECTORMASK;
      OT_YMEM64     = OT_REGNORM or otf_reg_ymm or otf_reg_gpr or OT_BITS64;
      OT_YMEM64_M   = OT_YMEM64 or OT_VECTORMASK;

      OT_YMMREG_M   = OT_YMMREG or OT_VECTORMASK;
      OT_YMMREG_MZ  = OT_YMMREG or OT_VECTORMASK or OT_VECTORZERO;
      OT_YMMRM_MZ   = OT_YMMRM  or OT_VECTORMASK or OT_VECTORZERO;
      OT_YMMREG_SAE = OT_YMMREG or OT_VECTORSAE;
      OT_YMMRM_SAE  = OT_YMMRM  or OT_VECTORSAE;
      OT_YMMREG_ER  = OT_YMMREG or OT_VECTORER;
      OT_YMMRM_ER   = OT_YMMRM  or OT_VECTORER;


      { register class 5: ZMM (both reg and r/m) }
      OT_ZMMREG     = OT_REGNORM or otf_reg_zmm;
      OT_ZMMRM      = OT_REGMEM or otf_reg_zmm;
      OT_ZMEM32     = OT_REGNORM or otf_reg_zmm or otf_reg_gpr or OT_BITS32;
      OT_ZMEM32_M   = OT_ZMEM32 or OT_VECTORMASK;
      OT_ZMEM64     = OT_REGNORM or otf_reg_zmm or otf_reg_gpr or OT_BITS64;
      OT_ZMEM64_M   = OT_ZMEM64 or OT_VECTORMASK;


      OT_ZMMREG_M   = OT_ZMMREG or OT_VECTORMASK;
      OT_ZMMREG_MZ  = OT_ZMMREG or OT_VECTORMASK or OT_VECTORZERO;
      OT_ZMMRM_MZ   = OT_ZMMRM  or OT_VECTORMASK or OT_VECTORZERO;
      OT_ZMMREG_SAE = OT_ZMMREG or OT_VECTORSAE;
      OT_ZMMRM_SAE  = OT_ZMMRM  or OT_VECTORSAE;
      OT_ZMMREG_ER  = OT_ZMMREG or OT_VECTORER;
      OT_ZMMRM_ER   = OT_ZMMRM  or OT_VECTORER;


      OT_KREG       = OT_REGNORM or otf_reg_k;
      OT_KREG_M     = OT_KREG or OT_VECTORMASK;

      { Vector-Memory operands }
      OT_VMEM_ANY  = OT_XMEM32 or OT_XMEM64 or OT_YMEM32 or OT_YMEM64 or OT_ZMEM32 or OT_ZMEM64;

      { Memory operands }
      OT_MEM8      = OT_MEMORY or OT_BITS8;
      OT_MEM16     = OT_MEMORY or OT_BITS16;
      OT_MEM16_M   = OT_MEM16  or OT_VECTORMASK;
      OT_MEM32     = OT_MEMORY or OT_BITS32;
      OT_MEM32_M   = OT_MEMORY or OT_BITS32 or OT_VECTORMASK;
      OT_BMEM32    = OT_MEMORY or OT_BITS32 or OT_VECTORBCST;
      OT_BMEM32_SAE= OT_MEMORY or OT_BITS32 or OT_VECTORBCST or OT_VECTORSAE;
      OT_MEM64     = OT_MEMORY or OT_BITS64;
      OT_MEM64_M   = OT_MEMORY or OT_BITS64 or OT_VECTORMASK;
      OT_BMEM64    = OT_MEMORY or OT_BITS64 or OT_VECTORBCST;
      OT_BMEM64_SAE= OT_MEMORY or OT_BITS64 or OT_VECTORBCST or OT_VECTORSAE;
      OT_MEM128    = OT_MEMORY or OT_BITS128;
      OT_MEM128_M  = OT_MEMORY or OT_BITS128 or OT_VECTORMASK;
      OT_MEM256    = OT_MEMORY or OT_BITS256;
      OT_MEM256_M  = OT_MEMORY or OT_BITS256 or OT_VECTORMASK;
      OT_MEM512    = OT_MEMORY or OT_BITS512;
      OT_MEM512_M  = OT_MEMORY or OT_BITS512 or OT_VECTORMASK;
      OT_MEM80     = OT_MEMORY or OT_BITS80;

      OT_MEM_OFFS  = OT_MEMORY or otf_sub0;  { special type of EA  }
                                             { simple [address] offset  }

      { Matches any type of r/m operand }
      OT_MEMORY_ANY = OT_MEMORY or OT_RM_GPR or OT_XMMRM or OT_MMXRM or OT_YMMRM or OT_ZMMRM or OT_REG_EXTRA_MASK;

      { Immediate operands }
      OT_IMM8      = OT_IMMEDIATE or OT_BITS8;
      OT_IMM16     = OT_IMMEDIATE or OT_BITS16;
      OT_IMM32     = OT_IMMEDIATE or OT_BITS32;
      OT_IMM64     = OT_IMMEDIATE or OT_BITS64;

      OT_ONENESS   = otf_sub0;  { special type of immediate operand  }
      OT_UNITY     = OT_IMMEDIATE or OT_ONENESS;  { for shift/rotate instructions  }

      OTVE_VECTOR_SAE          = 1 shl 8;
      OTVE_VECTOR_ER           = 1 shl 9;
      OTVE_VECTOR_ZERO         = 1 shl 10;
      OTVE_VECTOR_WRITEMASK    = 1 shl 11;
      OTVE_VECTOR_BCST         = 1 shl 12;
      OTVE_VECTOR_BCST2        = 0;
      OTVE_VECTOR_BCST4        = 1 shl 4;
      OTVE_VECTOR_BCST8        = 1 shl 5;
      OTVE_VECTOR_BCST16       = 3 shl 4;
      OTVE_VECTOR_RNSAE        = OTVE_VECTOR_ER or 0;
      OTVE_VECTOR_RDSAE        = OTVE_VECTOR_ER or 1 shl 6;
      OTVE_VECTOR_RUSAE        = OTVE_VECTOR_ER or 1 shl 7;
      OTVE_VECTOR_RZSAE        = OTVE_VECTOR_ER or 3 shl 6;


      OTVE_VECTOR_BCST_MASK    = OTVE_VECTOR_BCST2 or OTVE_VECTOR_BCST4 or OTVE_VECTOR_BCST8 or OTVE_VECTOR_BCST16;
      OTVE_VECTOR_ER_MASK      = OTVE_VECTOR_RNSAE or OTVE_VECTOR_RDSAE or OTVE_VECTOR_RUSAE or OTVE_VECTOR_RZSAE;

      OTVE_VECTOR_MASK = OTVE_VECTOR_SAE or OTVE_VECTOR_ER or OTVE_VECTOR_ZERO or OTVE_VECTOR_WRITEMASK or OTVE_VECTOR_BCST;

      { Size of the instruction table converted by nasmconv.pas }
      instabentries = 4235;
      maxinfolen    = 10;

    type
      { What an instruction can change. Needed for optimizer and spilling code.

        Note: The order of this enumeration is should not be changed! }
      TInsChange = (Ch_None,
        {Read from a register}
        Ch_REAX, Ch_RECX, Ch_REDX, Ch_REBX, Ch_RESP, Ch_REBP, Ch_RESI, Ch_REDI,
        {write from a register}
        Ch_WEAX, Ch_WECX, Ch_WEDX, Ch_WEBX, Ch_WESP, Ch_WEBP, Ch_WESI, Ch_WEDI,
        {read and write from/to a register}
        Ch_RWEAX, Ch_RWECX, Ch_RWEDX, Ch_RWEBX, Ch_RWESP, Ch_RWEBP, Ch_RWESI, Ch_RWEDI,
        {modify the contents of a register with the purpose of using
         this changed content afterwards (add/sub/..., but e.g. not rep
         or movsd)}
        Ch_MEAX, Ch_MECX, Ch_MEDX, Ch_MEBX, Ch_MESP, Ch_MEBP, Ch_MESI, Ch_MEDI,
        {read individual flag bits from the flags register}
        Ch_RCarryFlag,Ch_RParityFlag,Ch_RAuxiliaryFlag,Ch_RZeroFlag,Ch_RSignFlag,Ch_ROverflowFlag,
        {write individual flag bits to the flags register}
        Ch_WCarryFlag,Ch_WParityFlag,Ch_WAuxiliaryFlag,Ch_WZeroFlag,Ch_WSignFlag,Ch_WOverflowFlag,
        {set individual flag bits to 0 in the flags register}
        Ch_W0CarryFlag,Ch_W0ParityFlag,Ch_W0AuxiliaryFlag,Ch_W0ZeroFlag,Ch_W0SignFlag,Ch_W0OverflowFlag,
        {set individual flag bits to 1 in the flags register}
        Ch_W1CarryFlag,Ch_W1ParityFlag,Ch_W1AuxiliaryFlag,Ch_W1ZeroFlag,Ch_W1SignFlag,Ch_W1OverflowFlag,
        {write an undefined value to individual flag bits in the flags register}
        Ch_WUCarryFlag,Ch_WUParityFlag,Ch_WUAuxiliaryFlag,Ch_WUZeroFlag,Ch_WUSignFlag,Ch_WUOverflowFlag,
        {read and write flag bits}
        Ch_RWCarryFlag,Ch_RWParityFlag,Ch_RWAuxiliaryFlag,Ch_RWZeroFlag,Ch_RWSignFlag,Ch_RWOverflowFlag,
        {more specialized flag bits (not considered part of NR_DEFAULTFLAGS by the compiler)}
        Ch_RDirFlag,Ch_W0DirFlag,Ch_W1DirFlag,Ch_W0IntFlag,Ch_W1IntFlag,
        {instruction reads flag bits, according to its condition (used by Jcc/SETcc/CMOVcc)}
        Ch_RFLAGScc,
        {read/write/read+write the entire flags/eflags/rflags register}
        Ch_RFlags, Ch_WFlags, Ch_RWFlags,
        Ch_FPU,
        Ch_Rop1, Ch_Wop1, Ch_RWop1, Ch_Mop1,
        Ch_Rop2, Ch_Wop2, Ch_RWop2, Ch_Mop2,
        Ch_Rop3, Ch_WOp3, Ch_RWOp3, Ch_Mop3,
        Ch_Rop4, Ch_WOp4, Ch_RWOp4, Ch_Mop4,
        { instruction doesn't read it's input register, in case both parameters
          are the same register (e.g. xor eax,eax; sub eax,eax; sbb eax,eax (reads flags only), etc.) }
        Ch_NoReadIfEqualRegs,
        Ch_RMemEDI,Ch_WMemEDI,
        Ch_All,
        { x86_64 registers }
        Ch_RRAX, Ch_RRCX, Ch_RRDX, Ch_RRBX, Ch_RRSP, Ch_RRBP, Ch_RRSI, Ch_RRDI,
        Ch_WRAX, Ch_WRCX, Ch_WRDX, Ch_WRBX, Ch_WRSP, Ch_WRBP, Ch_WRSI, Ch_WRDI,
        Ch_RWRAX, Ch_RWRCX, Ch_RWRDX, Ch_RWRBX, Ch_RWRSP, Ch_RWRBP, Ch_RWRSI, Ch_RWRDI,
        Ch_MRAX, Ch_MRCX, Ch_MRDX, Ch_MRBX, Ch_MRSP, Ch_MRBP, Ch_MRSI, Ch_MRDI,

        { xmm register }
        Ch_RXMM0,
        Ch_WXMM0,
        Ch_RWXMM0,
        Ch_MXMM0
      );

      TInsProp = packed record
        Ch : set of TInsChange;
      end;

      TMemRefSizeInfo = (msiUnknown, msiUnsupported, msiNoSize, msiNoMemRef,
                         msiMultiple, msiMultipleMinSize8, msiMultipleMinSize16, msiMultipleMinSize32,
                         msiMultipleMinSize64, msiMultipleMinSize128, msiMultipleminSize256, msiMultipleMinSize512,
                         msiMemRegSize, msiMemRegx16y32, msiMemRegx16y32z64, msiMemRegx32y64, msiMemRegx32y64z128, msiMemRegx64y128, msiMemRegx64y128z256,
                         msiMemRegx64y256, msiMemRegx64y256z512,
                         msiMem8, msiMem16, msiMem32, msiBMem32, msiMem64, msiBMem64, msiMem128, msiMem256, msiMem512,
                         msiXMem32, msiXMem64, msiYMem32, msiYMem64, msiZMem32, msiZMem64,
                         msiVMemMultiple, msiVMemRegSize,
                         msiMemRegConst128,msiMemRegConst256,msiMemRegConst512);

      TMemRefSizeInfoBCST = (msbUnknown, msbBCST32, msbBCST64, msbMultiple);
      TMemRefSizeInfoBCSTType = (btUnknown, bt1to2, bt1to4, bt1to8, bt1to16);

      TEVEXTupleState = (etsUnknown, etsIsTuple, etsNotTuple);
      TConstSizeInfo  = (csiUnknown, csiMultiple, csiNoSize, csiMem8, csiMem16, csiMem32, csiMem64);

      TInsTabMemRefSizeInfoRec = record
        MemRefSize               : TMemRefSizeInfo;
        MemRefSizeBCST           : TMemRefSizeInfoBCST;
        BCSTXMMMultiplicator     : byte;
        ExistsSSEAVX             : boolean;
        ConstSize                : TConstSizeInfo;
        BCSTTypes                : Set of TMemRefSizeInfoBCSTType;
        RegXMMSizeMask           : int64;
        RegYMMSizeMask           : int64;
        RegZMMSizeMask           : int64;
      end;



    const
      MemRefMultiples: set of TMemRefSizeInfo = [msiMultiple, msiMultipleMinSize8,
                                                 msiMultipleMinSize16, msiMultipleMinSize32,
                                                 msiMultipleMinSize64, msiMultipleMinSize128,
                                                 msiMultipleMinSize256, msiMultipleMinSize512,
                                                 msiVMemMultiple];

      MemRefSizeInfoVMems: Set of TMemRefSizeInfo = [msiXMem32, msiXMem64, msiYMem32, msiYMem64,
                                                     msiZMem32, msiZMem64,
                                                     msiVMemMultiple, msiVMemRegSize];



      InsProp : array[tasmop] of TInsProp =
(
(Ch: []),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RWCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_W0OverflowFlag, Ch_W0CarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Wop2, Ch_Rop1, Ch_WZeroFlag, Ch_WUCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_Wop2, Ch_Rop1, Ch_WZeroFlag, Ch_WUCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_MOp1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_ROp1, Ch_All]),
(Ch: [Ch_MEAX]),
(Ch: [Ch_MEAX, Ch_WEDX]),
(Ch: [Ch_W0CarryFlag]),
(Ch: [Ch_W0DirFlag]),
(Ch: [Ch_W0IntFlag]),
(Ch: []),
(Ch: [Ch_RWCarryFlag]),
(Ch: [Ch_ROp1, Ch_ROp2, Ch_NoReadIfEqualRegs, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_RWESI, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_RWESI, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_RWESI, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_All, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_All, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_All, Ch_WZeroFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_MEAX, Ch_WEDX]),
(Ch: [Ch_MEAX]),
(Ch: [Ch_Mop1, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag]),
(Ch: [Ch_RWEAX, Ch_WEDX, Ch_WUCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUZeroFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_FPU]),
(Ch: [Ch_RWESP]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_Rop1, Ch_FPU]),
(Ch: [Ch_Wop1, Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU, Ch_RCarryFlag]),
(Ch: [Ch_FPU, Ch_RCarryFlag, Ch_RZeroFlag]),
(Ch: [Ch_FPU, Ch_RZeroFlag]),
(Ch: [Ch_FPU, Ch_RCarryFlag]),
(Ch: [Ch_FPU, Ch_RCarryFlag, Ch_RZeroFlag]),
(Ch: [Ch_FPU, Ch_RZeroFlag]),
(Ch: [Ch_FPU, Ch_RParityFlag]),
(Ch: [Ch_FPU, Ch_RParityFlag]),
(Ch: [Ch_FPU]),
(Ch: [Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_FPU, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_All]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_FPU, Ch_ROp1]),
(Ch: [Ch_Rop1, Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1, Ch_FPU]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_ROp1, Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: []),
(Ch: [Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_FPU, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: [Ch_FPU]),
(Ch: []),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_RWEAX, Ch_WEDX, Ch_WUCarryFlag, Ch_WUOverflowFlag, Ch_WUSignFlag, Ch_WUZeroFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_RWEAX, Ch_WEDX, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WUSignFlag, Ch_WUZeroFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop1, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag]),
(Ch: [Ch_WMemEDI, Ch_RWEDI, Ch_REDX, Ch_RDirFlag]),
(Ch: [Ch_WMemEDI, Ch_RWEDI, Ch_REDX, Ch_RDirFlag]),
(Ch: [Ch_WMemEDI, Ch_RWEDI, Ch_REDX, Ch_RDirFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: []),
(Ch: []),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_RECX]),
(Ch: [Ch_RECX]),
(Ch: [Ch_ROp1]),
(Ch: [Ch_WEAX, Ch_RFlags]),
(Ch: [Ch_Wop2, Ch_WZeroFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_RWESP, Ch_RWEBP]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: []),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: []),
(Ch: [Ch_ROp1]),
(Ch: []),
(Ch: []),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: []),
(Ch: [Ch_WEAX, Ch_RWESI, Ch_RDirFlag]),
(Ch: [Ch_WEAX, Ch_RWESI, Ch_RDirFlag]),
(Ch: [Ch_WEAX, Ch_RWESI, Ch_RDirFlag]),
(Ch: [Ch_RWECX]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_Wop2, Ch_WZeroFlag]),
(Ch: [Ch_Wop2, Ch_ROP1]),
(Ch: []),
(Ch: []),
(Ch: [Ch_Wop2, Ch_Rop1, Ch_NoReadIfEqualRegs]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_RWESI, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_RWESI, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_RWRSI, Ch_WMemEDI, Ch_RWRDI, Ch_RDirFlag]),
(Ch: [Ch_RWESI, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_RWEAX, Ch_WEDX, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WUSignFlag, Ch_WUZeroFlag, Ch_WUAuxiliaryFlag, Ch_WUParityFlag]),
(Ch: []),
(Ch: [Ch_Mop1, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag]),
(Ch: []),
(Ch: [Ch_Mop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_W0OverflowFlag, Ch_W0CarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Rop1, Ch_Rop2]),
(Ch: [Ch_All, Ch_RDirFlag]),
(Ch: [Ch_All, Ch_RDirFlag]),
(Ch: [Ch_All, Ch_RDirFlag]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop1, Ch_RWESP]),
(Ch: [Ch_RWESP, Ch_WFlags]),
(Ch: [Ch_RWESP, Ch_WFLAGS]),
(Ch: [Ch_RWESP, Ch_WFlags]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_RWESP]),
(Ch: [Ch_RWESP, Ch_RFlags]),
(Ch: [Ch_RWESP, Ch_RFLAGS]),
(Ch: [Ch_RWESP, Ch_RFlags]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RWCarryFlag, Ch_WOverflowFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RWCarryFlag, Ch_WOverflowFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_WEAX, Ch_WEDX]),
(Ch: [Ch_WEAX, Ch_WEDX]),
(Ch: [Ch_WEAX, Ch_WEDX]),
(Ch: [Ch_RWECX]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_RWECX, Ch_RZeroFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_REAX, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag, Ch_WCarryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_NoReadIfEqualRegs, Ch_RWCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WParityFlag]),
(Ch: [Ch_REAX, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_REAX, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_RRAX, Ch_RMemEDI, Ch_RWRDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_REAX, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: []),
(Ch: []),
(Ch: []),
(Ch: []),
(Ch: []),
(Ch: []),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_MOp3, Ch_Rop2, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_MOp3, Ch_Rop2, Ch_WCarryFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_W1CarryFlag]),
(Ch: [Ch_W1DirFlag]),
(Ch: [Ch_W1IntFlag]),
(Ch: [Ch_REAX, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_REAX, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_REAX, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_NoReadIfEqualRegs, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_W0OverflowFlag, Ch_W0CarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_WZeroFlag]),
(Ch: [Ch_WZeroFlag]),
(Ch: []),
(Ch: []),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_RWop1, Ch_RWop2, Ch_NoReadIfEqualRegs]),
(Ch: [Ch_WEAX, Ch_REBX]),
(Ch: [Ch_WEAX, Ch_REBX]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_NoReadIfEqualRegs, Ch_W0OverflowFlag, Ch_W0CarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_ROp1, Ch_MOp2, Ch_RFLAGScc]),
(Ch: [Ch_RFLAGScc]),
(Ch: [Ch_RFLAGScc, Ch_WOp1]),
(Ch: [Ch_RWESI, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_RWESI, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_REAX, Ch_RMemEDI, Ch_RWEDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_WEAX, Ch_RWESI, Ch_RDirFlag]),
(Ch: [Ch_REAX, Ch_WMemEDI, Ch_RWEDI, Ch_RDirFlag]),
(Ch: [Ch_WMemEDI, Ch_RWEDI, Ch_REDX, Ch_RDirFlag]),
(Ch: [Ch_All, Ch_RDirFlag]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WFlags]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_ROp1, Ch_WOp2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WMemEDI]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WMemEDI]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Rop1, Ch_Mop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WFlags]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_ROp1, Ch_WOp2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_MRAX, Ch_WRDX]),
(Ch: [Ch_MRAX]),
(Ch: [Ch_RWRAX, Ch_RWRDX, Ch_WZeroFlag, Ch_RWOp1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RXMM0]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RXMM0]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RXMM0]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Rop2, Ch_Rop1, Ch_WZeroFlag, Ch_WCarryFlag]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Mop1, Ch_Rop2]),
(Ch: [Ch_Rop3, Ch_Rop2, Ch_REAX, Ch_REDX, Ch_WECX]),
(Ch: [Ch_Rop3, Ch_Rop2, Ch_REAX, Ch_REDX, Ch_WXMM0]),
(Ch: [Ch_Rop3, Ch_Rop2, Ch_WECX]),
(Ch: [Ch_Rop3, Ch_Rop2, Ch_WXMM0]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_WEAX, Ch_WEDX]),
(Ch: [Ch_RRAX, Ch_WMemEDI, Ch_RWRDI, Ch_RDirFlag]),
(Ch: [Ch_WRAX, Ch_RWRSI, Ch_RDirFlag]),
(Ch: [Ch_RWRSI, Ch_RMemEDI, Ch_RWRDI, Ch_RDirFlag, Ch_WOverflowFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WAuxiliaryFlag, Ch_WCarryFlag, Ch_WParityFlag]),
(Ch: [Ch_Rop1, Ch_Wop2]),
(Ch: [Ch_Mop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WFlags]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WFlags]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WMemEDI]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Rop3, Ch_Wop4]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Rop3, Ch_Wop4]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop4, Ch_Rop3, Ch_Rop2]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_WZeroFlag, Ch_WParityFlag, Ch_WCarryFlag, Ch_W0OverflowFlag, Ch_W0SignFlag, Ch_W0AuxiliaryFlag]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Wop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Wop2, Ch_W0OverflowFlag, Ch_WCarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WUParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Rop1, Ch_Wop2, Ch_W0OverflowFlag, Ch_WCarryFlag, Ch_WSignFlag, Ch_W0ZeroFlag, Ch_WUParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Rop1, Ch_Wop2, Ch_W0OverflowFlag, Ch_WCarryFlag, Ch_WSignFlag, Ch_WZeroFlag, Ch_WUParityFlag, Ch_WUAuxiliaryFlag]),
(Ch: [Ch_Wop2, Ch_WFlags, Ch_Rop1]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_REDX, Ch_Rop1, Ch_Wop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Mop2, Ch_RWCarryFlag]),
(Ch: [Ch_Rop1, Ch_Mop2, Ch_RWOverflowFlag]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: [Ch_Mop3, Ch_Rop2, Ch_Rop1]),
(Ch: []),
(Ch: []),
(Ch: []),
(Ch: [Ch_All]),
(Ch: []),
(Ch: [Ch_WFlags]),
(Ch: [Ch_Wop1, Ch_WFlags]),
(Ch: [Ch_Wop1, Ch_WFlags]),
(Ch: [Ch_WEAX, Ch_WEDX, Ch_RECX]),
(Ch: [Ch_REAX, Ch_REDX, Ch_RECX]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_Wop2, Ch_Rop1]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Wop3]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Rop3, Ch_Wop4]),
(Ch: [Ch_Rop1, Ch_Rop2, Ch_Rop3, Ch_Wop4]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_All]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1, Ch_RXMM0]),
(Ch: [Ch_Mop2, Ch_Rop1]),
(Ch: [Ch_Mop2, Ch_Rop1])
);


    type
      TOperandOrder = (op_intel,op_att);

      {Instruction flags }
      tinsflag = (
        { please keep these in order and in sync with IF_SMASK }
        IF_SM,                  { size match first two operands  }
        IF_SM2,
        IF_SB,                  { unsized operands can't be non-byte  }
        IF_SW,                  { unsized operands can't be non-word  }
        IF_SD,                  { unsized operands can't be nondword  }

        { unsized argument spec }
        { please keep these in order and in sync with IF_ARMASK }
        IF_AR0,                 { SB, SW, SD applies to argument 0  }
        IF_AR1,                 { SB, SW, SD applies to argument 1  }
        IF_AR2,                 { SB, SW, SD applies to argument 2  }

        IF_PRIV,                { it's a privileged instruction  }
        IF_SMM,                 { it's only valid in SMM  }
        IF_PROT,                { it's protected mode only  }
        IF_NOX86_64,            { removed instruction in x86_64  }
        IF_UNDOC,               { it's an undocumented instruction  }
        IF_FPU,                 { it's an FPU instruction  }
        IF_MMX,                 { it's an MMX instruction  }
        { it's a 3DNow! instruction  }
        IF_3DNOW,
        { it's a SSE (KNI, MMX2) instruction  }
        IF_SSE,
        { SSE2 instructions  }
        IF_SSE2,
        { SSE3 instructions  }
        IF_SSE3,
        { SSE64 instructions  }
        IF_SSE64,
        { SVM instructions  }
        IF_SVM,
        { SSE4 instructions  }
        IF_SSE4,
        IF_SSSE3,
        IF_SSE41,
        IF_SSE42,
        IF_MOVBE,
        IF_CLMUL,
        IF_AVX,
        IF_AVX2,
        IF_AVX512,
        IF_BMI1,
        IF_BMI2,
        { Intel ADX (Multi-Precision Add-Carry Instruction Extensions) }
        IF_ADX,
        IF_16BITONLY,
        IF_FMA,
        IF_FMA4,
        IF_TSX,
        IF_RAND,
        IF_XSAVE,
        IF_PREFETCHWT1,
        IF_SHA,

        { mask for processor level }
        { please keep these in order and in sync with IF_PLEVEL }
        IF_8086,                { 8086 instruction  }
        IF_186,                 { 186+ instruction  }
        IF_286,                 { 286+ instruction  }
        IF_386,                 { 386+ instruction  }
        IF_486,                 { 486+ instruction  }
        IF_PENT,                { Pentium instruction  }
        IF_P6,                  { P6 instruction  }
        IF_KATMAI,              { Katmai instructions  }
        IF_WILLAMETTE,          { Willamette instructions }
        IF_PRESCOTT,            { Prescott instructions }
        IF_X86_64,
        IF_SANDYBRIDGE,         { Sandybridge-specific instruction }
        IF_NEC,                 { NEC V20/V30 instruction }

        { the following are not strictly part of the processor level, because
          they are never used standalone, but always in combination with a
          separate processor level flag. Therefore, they use bits outside of
          IF_PLEVEL, otherwise they would mess up the processor level they're
          used in combination with.
          The following combinations are currently used:
          [IF_AMD, IF_P6],
          [IF_CYRIX, IF_486],
          [IF_CYRIX, IF_PENT],
          [IF_CYRIX, IF_P6] }
        IF_CYRIX,               { Cyrix, Centaur or VIA-specific instruction }
        IF_AMD,                 { AMD-specific instruction  }

        { added flags }
        IF_PRE,                 { it's a prefix instruction }
        IF_PASS2,               { if the instruction can change in a second pass }
        IF_IMM4,                { immediate operand is a nibble (must be in range [0..15]) }
        IF_IMM3,                { immediate operand is a triad (must be in range [0..7]) }

        { avx512 flags }
        IF_BCST2,
        IF_BCST4,
        IF_BCST8,
        IF_BCST16,
        IF_T2,                  { disp8 - tuple - 2 }
        IF_T4,                  { disp8 - tuple - 4 }
        IF_T8,                  { disp8 - tuple - 8 }
        IF_T1S,                 { disp8 - tuple - 1 scalar }
        IF_T1S8,                { disp8 - tuple - 1 scalar byte }
        IF_T1S16,               { disp8 - tuple - 1 scalar word }
        IF_T1F32,
        IF_T1F64,
        IF_TMDDUP,
        IF_TFV,                 { disp8 - tuple - full vector }
        IF_TFVM,                { disp8 - tuple - full vector memory }
        IF_TQVM,
        IF_TMEM128,
        IF_THV,
        IF_THVM,
        IF_TOVM

      );
      tinsflags=set of tinsflag;

    const
      IF_SMASK=[IF_SM,IF_SM2,IF_SB,IF_SW,IF_SD];
      IF_ARMASK=[IF_AR0,IF_AR1,IF_AR2];  { mask for unsized argument spec  }
      IF_PLEVEL=[IF_8086..IF_NEC]; { mask for processor level }

      IF_TUPLEMASK=[IF_T2..IF_TOVM]; { mask for AVX512 disp8-tuples }


    type
      tinsentry=packed record
        opcode  : tasmop;
        ops     : byte;
        optypes : array[0..max_operands-1] of int64;
        code    : array[0..maxinfolen] of char;
        flags   : tinsflags;
      end;
      pinsentry=^tinsentry;

      { alignment for operator }
      tai_align = class(tai_align_abstract)
         function calculatefillbuf(var buf : tfillbuffer;executable : boolean):pchar;override;
      end;

      { taicpu }

      taicpu = class(tai_cpu_abstract_sym)
         opsize    : topsize;
         constructor op_none(op : tasmop);
         constructor op_none(op : tasmop;_size : topsize);

         constructor op_reg(op : tasmop;_size : topsize;_op1 : tregister);
         constructor op_const(op : tasmop;_size : topsize;_op1 : aint);
         constructor op_ref(op : tasmop;_size : topsize;const _op1 : treference);

         constructor op_reg_reg(op : tasmop;_size : topsize;_op1,_op2 : tregister);
         constructor op_reg_ref(op : tasmop;_size : topsize;_op1 : tregister;const _op2 : treference);
         constructor op_reg_const(op:tasmop; _size: topsize; _op1: tregister; _op2: aint);

         constructor op_const_reg(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister);
         constructor op_const_const(op : tasmop;_size : topsize;_op1,_op2 : aint);
         constructor op_const_ref(op : tasmop;_size : topsize;_op1 : aint;const _op2 : treference);

         constructor op_ref_reg(op : tasmop;_size : topsize;const _op1 : treference;_op2 : tregister);

         constructor op_reg_reg_reg(op : tasmop;_size : topsize;_op1,_op2,_op3 : tregister);
         constructor op_const_reg_reg(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister;_op3 : tregister);
         constructor op_const_ref_reg(op : tasmop;_size : topsize;_op1 : aint;const _op2 : treference;_op3 : tregister);
         constructor op_ref_reg_reg(op : tasmop;_size : topsize;const _op1 : treference;_op2,_op3 : tregister);
         constructor op_const_reg_ref(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister;const _op3 : treference);
         constructor op_reg_reg_ref(op : tasmop;_size : topsize;_op1,_op2 : tregister;const _op3 : treference);
         constructor op_const_reg_reg_reg(op : tasmop;_size : topsize;_op1 : aint;_op2, _op3, _op4 : tregister);

         { this is for Jmp instructions }
         constructor op_cond_sym(op : tasmop;cond:TAsmCond;_size : topsize;_op1 : tasmsymbol);

         constructor op_sym(op : tasmop;_size : topsize;_op1 : tasmsymbol);
         constructor op_sym_ofs(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint);
         constructor op_sym_ofs_reg(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint;_op2 : tregister);
         constructor op_sym_ofs_ref(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint;const _op2 : treference);

         procedure changeopsize(siz:topsize); {$ifdef USEINLINE}inline;{$endif USEINLINE}

         { This is a workaround for the GAS non commutative fpu instruction braindamage.
           Early versions of the UnixWare assembler had a bug where some fpu instructions
           were reversed and GAS still keeps this "feature" for compatibility.
           for details: http://sourceware.org/binutils/docs/as/i386_002dBugs.html#i386_002dBugs
                        http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=372528
                        http://en.wikibooks.org/wiki/X86_Assembly/GAS_Syntax#Caveats

           Since FPC is "GAS centric" due to its history it generates instructions with the same operand order so
           when generating output for other assemblers, the opcodes must be fixed before writing them.
           This function returns the fixed opcodes. Changing the opcodes permanently is no good idea
           because in case of smartlinking assembler is generated twice so at the second run wrong
           assembler is generated.
           }
         function FixNonCommutativeOpcodes: tasmop;
      private
         FOperandOrder : TOperandOrder;
         procedure init(_size : topsize); { this need to be called by all constructor }
      public
         { the next will reset all instructions that can change in pass 2 }
         procedure ResetPass1;override;
         procedure ResetPass2;override;
         function  CheckIfValid:boolean; {$ifdef USEINLINE}inline;{$endif USEINLINE}
         function  Pass1(objdata:TObjData):longint;override;
         procedure Pass2(objdata:TObjData);override;
         procedure SetOperandOrder(order:TOperandOrder);
         function is_same_reg_move(regtype: Tregistertype):boolean;override;
         { register spilling code }
         function spilling_get_operation_type(opnr: longint): topertype;override;
         property OperandOrder : TOperandOrder read FOperandOrder;
      private
         { next fields are filled in pass1, so pass2 is faster }
         insentry  : PInsEntry;
         insoffset : longint;
         LastInsOffset : longint; { need to be public to be reset }
         inssize   : shortint;
         EVEXTupleState: TEVEXTupleState; { AVX512 disp8*N }
         rex       : byte;
         function  InsEnd:longint;
         procedure create_ot(objdata:TObjData);
         function  Matches(p:PInsEntry):boolean;
         function  calcsize(p:PInsEntry):shortint;
         procedure gencode(objdata:TObjData);
         function  NeedAddrPrefix(opidx:byte):boolean;
         function  NeedAddrPrefix:boolean;
         procedure write0x66prefix(objdata:TObjData);
         procedure write0x67prefix(objdata:TObjData);
         procedure Swapoperands;
         function  FindInsentry(objdata:TObjData):boolean;
         function  CheckUseEVEX: boolean;
         procedure CheckEVEXTuple(const aInput:toper; aInsEntry: pInsentry; aIsVector128, aIsVector256, aIsVector512, aIsEVEXW1: boolean);
      end;

    function is_64_bit_ref(const ref:treference):boolean;
    function is_32_bit_ref(const ref:treference):boolean;
    function is_16_bit_ref(const ref:treference):boolean;
    function get_ref_address_size(const ref:treference):byte;
    function get_default_segment_of_ref(const ref:treference):tregister;
    procedure optimize_ref(var ref:treference; inlineasm: boolean);

    function spilling_create_load(const ref:treference;r:tregister):Taicpu;
    function spilling_create_store(r:tregister; const ref:treference):Taicpu;

    function MemRefInfo(aAsmop: TAsmOp): TInsTabMemRefSizeInfoRec;
    function MightHaveExtension(AsmOp : TAsmOp) : Boolean;

{*****************************************************************************
                              External Symbol Chain
                              used for agx86nsm and agx86int
*****************************************************************************}

    type
      PExternChain = ^TExternChain;

      TExternChain = Record
        psym : pshortstring;
        is_defined : boolean;
        next : PExternChain;
      end;

    const
      FEC : PExternChain = nil;

    procedure AddSymbol(symname : string; defined : boolean);
    procedure FreeExternChainList;

implementation

     uses
       cutils,
       globals,
       systems,
       itcpugas,
       cpuinfo;

    procedure AddSymbol(symname : string; defined : boolean);
    var
       EC : PExternChain;
    begin
      EC:=FEC;
      while assigned(EC) do
        begin
          if EC^.psym^=symname then
            begin
              if defined then
                EC^.is_defined:=true;
              exit;
            end;
          EC:=EC^.next;
        end;
      New(EC);
      EC^.next:=FEC;
      FEC:=EC;
      FEC^.psym:=stringdup(symname);
      FEC^.is_defined := defined;
    end;

    procedure FreeExternChainList;
    var
       EC : PExternChain;
    begin
      EC:=FEC;
      while assigned(EC) do
        begin
          FEC:=EC^.next;
          stringdispose(EC^.psym);
          Dispose(EC);
          EC:=FEC;
        end;
    end;

{*****************************************************************************
                              Instruction table
*****************************************************************************}

     type
       TInsTabCache=array[TasmOp] of longint;
       PInsTabCache=^TInsTabCache;

       TInsTabMemRefSizeInfoCache=array[TasmOp] of TInsTabMemRefSizeInfoRec;
       PInsTabMemRefSizeInfoCache=^TInsTabMemRefSizeInfoCache;

     const
       InsTab:array[0..instabentries-1] of TInsEntry=
(
  (
    opcode  : A_NONE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #0;
    flags   : []
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#17#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#19#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#16#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#18#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#130#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#21#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#21#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#21#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#129#130#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#130#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#130#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#21#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#130#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#20#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_ADC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#130#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#1#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#3#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#0#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#2#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#128#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #208#1#5#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#5#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#5#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#129#128#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#128#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#128#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#5#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#128#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#4#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_ADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#128#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#33#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#35#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#32#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#34#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#132#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#37#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#37#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#37#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#129#132#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#132#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#132#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#37#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#132#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#36#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_AND;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#132#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_BSF;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#2#15#188#72;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BSR;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#2#15#189#72;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BSWAP;
    ops     : 1;
    optypes : (ot_reg32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#15#8#200;
    flags   : [if_486]
  ),
  (
    opcode  : A_BT;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#163#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BT;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#2#15#186#132#21;
    flags   : [if_386,if_sb]
  ),
  (
    opcode  : A_BTC;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#187#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BTC;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#2#15#186#135#21;
    flags   : [if_386,if_sb]
  ),
  (
    opcode  : A_BTR;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#179#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BTR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#2#15#186#134#21;
    flags   : [if_386,if_sb]
  ),
  (
    opcode  : A_BTS;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#171#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_BTS;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#2#15#186#133#21;
    flags   : [if_386,if_sb]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#255#130;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#255#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #208#1#232#52;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_immediate or ot_near,ot_none,ot_none,ot_none);
    code    : #208#1#232#52;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_memory or ot_near,ot_none,ot_none,ot_none);
    code    : #208#1#255#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CALL;
    ops     : 1;
    optypes : (ot_memory or ot_far,ot_none,ot_none,ot_none);
    code    : #208#1#255#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CBW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#152;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CDQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#153;
    flags   : [if_386]
  ),
  (
    opcode  : A_CLC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#248;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CLD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#252;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CLI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#250;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CLTS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#6;
    flags   : [if_286,if_priv]
  ),
  (
    opcode  : A_CMC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#245;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#57#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#59#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#56#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#58#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#135#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#61#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#61#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#61#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#129#135#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#135#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#135#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#61#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#135#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#60#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#135#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits32,ot_none,ot_none);
    code    : #213#1#129#135#33;
    flags   : [if_386,if_sd]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits16,ot_none,ot_none);
    code    : #212#1#129#135#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_CMP;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits8,ot_none,ot_none);
    code    : #1#128#135#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_CMPSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#1#166;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMPSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#213#1#167;
    flags   : [if_386]
  ),
  (
    opcode  : A_CMPSD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #220#2#15#194#72#22;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_CMPSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#212#1#167;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CMPXCHG;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#177#65;
    flags   : [if_pent,if_sm]
  ),
  (
    opcode  : A_CMPXCHG;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #2#15#176#65;
    flags   : [if_pent]
  ),
  (
    opcode  : A_CMPXCHG486;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#167#65;
    flags   : [if_486,if_sm]
  ),
  (
    opcode  : A_CMPXCHG486;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #2#15#166#65;
    flags   : [if_486,if_undoc]
  ),
  (
    opcode  : A_CMPXCHG8B;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#199#129;
    flags   : [if_pent]
  ),
  (
    opcode  : A_CPUID;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#162;
    flags   : [if_pent]
  ),
  (
    opcode  : A_CWD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#153;
    flags   : [if_8086]
  ),
  (
    opcode  : A_CWDE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#152;
    flags   : [if_386]
  ),
  (
    opcode  : A_DEC;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#255#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_DEC;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#254#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_DIV;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#134;
    flags   : [if_8086]
  ),
  (
    opcode  : A_DIV;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#134;
    flags   : [if_8086]
  ),
  (
    opcode  : A_EMMS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#119;
    flags   : [if_pent,if_mmx]
  ),
  (
    opcode  : A_ENTER;
    ops     : 2;
    optypes : (ot_immediate,ot_immediate,ot_none,ot_none);
    code    : #1#200#24#21;
    flags   : [if_186]
  ),
  (
    opcode  : A_F2XM1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FABS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#225;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#193;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADD;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADDP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#193;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADDP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FADDP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FBLD;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #1#223#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FBLD;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#223#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FBSTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #1#223#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FBSTP;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#223#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCHS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCLEX;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#155#219#226;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCMOVB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#218#193;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVB;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#218#8#192;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVB;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#218#9#192;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVBE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#218#209;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVBE;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#218#8#208;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVBE;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#218#9#208;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#218#201;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVE;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#218#8#200;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVE;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#218#9#200;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#193;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNB;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#192;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNB;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#192;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNBE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#209;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNBE;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#208;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNBE;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#208;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#201;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNE;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#200;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNE;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#200;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNU;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#217;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNU;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#216;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVNU;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#216;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVU;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#218#217;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVU;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#218#8#216;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCMOVU;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#218#9#216;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOM;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOM;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#216#209;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOM;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#208;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOM;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#208;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#241;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMI;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#240;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMI;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#240;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMIP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#223#241;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMIP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#223#8#240;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMIP;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#223#9#240;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FCOMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#216#217;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#216;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMP;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#216;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOMPP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#217;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FCOS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#255;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FDECSTP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#246;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDISI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#155#219#225;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#220#241;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIV;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#241;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#240;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#220#249;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVR;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVRP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#249;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVRP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FDIVRP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FEMMS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#14;
    flags   : [if_pent,if_3dnow]
  ),
  (
    opcode  : A_FENI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#155#219#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FFREE;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#221#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIADD;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIADD;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FICOM;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FICOM;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FICOMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FICOMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIDIV;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIDIV;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIDIVR;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIDIVR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FILD;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#219#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FILD;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#223#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FILD;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#223#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIMUL;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#129;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIMUL;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#129;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FINCSTP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#247;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FINIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#155#219#227;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIST;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#219#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FIST;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#223#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#219#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#223#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#223#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISTTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#219#129;
    flags   : [if_prescott,if_fpu]
  ),
  (
    opcode  : A_FISTTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#223#129;
    flags   : [if_prescott,if_fpu]
  ),
  (
    opcode  : A_FISTTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#221#129;
    flags   : [if_prescott,if_fpu]
  ),
  (
    opcode  : A_FISUB;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISUB;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISUBR;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #1#222#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FISUBR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#218#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLD;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#217#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLD;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#221#128;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLD;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #1#219#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLD;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#217#8#192;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLD1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDCW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#217#133;
    flags   : [if_8086,if_fpu,if_sw]
  ),
  (
    opcode  : A_FLDENV;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#217#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDL2E;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#234;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDL2T;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#233;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDLG2;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#236;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDLN2;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#237;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDPI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#235;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FLDZ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#238;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#129;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#129;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#220#201;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMUL;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMULP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#201;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMULP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FMULP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNCLEX;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#226;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNDISI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#225;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNENI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNINIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#227;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNOP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#208;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNSAVE;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#221#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNSTCW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#217#135;
    flags   : [if_8086,if_fpu,if_sw]
  ),
  (
    opcode  : A_FNSTENV;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#217#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FNSTSW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#221#135;
    flags   : [if_8086,if_fpu,if_sw]
  ),
  (
    opcode  : A_FNSTSW;
    ops     : 1;
    optypes : (ot_reg_ax,ot_none,ot_none,ot_none);
    code    : #2#223#224;
    flags   : [if_286,if_fpu]
  ),
  (
    opcode  : A_FPATAN;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#243;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FPREM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#248;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FPREM1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#245;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FPTAN;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#242;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FRNDINT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#252;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FRSTOR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #1#221#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSAVE;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#155#221#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSCALE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#253;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSETPM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#228;
    flags   : [if_286,if_fpu]
  ),
  (
    opcode  : A_FSIN;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#254;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FSINCOS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#251;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FSQRT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#250;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FST;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#217#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FST;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#221#130;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FST;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#221#8#208;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTCW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#155#217#135;
    flags   : [if_8086,if_fpu,if_sw]
  ),
  (
    opcode  : A_FSTENV;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#155#217#134;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#217#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#221#131;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTP;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #1#219#135;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#221#8#216;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSTSW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#155#221#135;
    flags   : [if_8086,if_fpu,if_sw]
  ),
  (
    opcode  : A_FSTSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#155#223#224;
    flags   : [if_286,if_fpu]
  ),
  (
    opcode  : A_FSTSW;
    ops     : 1;
    optypes : (ot_reg_ax,ot_none,ot_none,ot_none);
    code    : #3#155#223#224;
    flags   : [if_286,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#132;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#220#225;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUB;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#225;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#224;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #1#216#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #1#220#133;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#220#233;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 1;
    optypes : (ot_fpureg or ot_to,ot_none,ot_none,ot_none);
    code    : #1#220#8#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#220#8#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#216#8#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBR;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#216#9#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBRP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#222#233;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBRP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#222#8#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FSUBRP;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#222#8#232;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FTST;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#228;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FUCOM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#221#225;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOM;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#221#8#224;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOM;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#221#9#224;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOMI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#219#233;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMI;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#219#8#232;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMI;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#219#9#232;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMIP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#223#233;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMIP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#223#8#232;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMIP;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#223#9#232;
    flags   : [if_p6,if_fpu]
  ),
  (
    opcode  : A_FUCOMP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#221#233;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOMP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#221#8#232;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOMP;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#221#9#232;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FUCOMPP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#218#233;
    flags   : [if_386,if_fpu]
  ),
  (
    opcode  : A_FWAIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#155;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXAM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#229;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXCH;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#201;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXCH;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#217#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXCH;
    ops     : 2;
    optypes : (ot_fpureg,ot_fpu0,ot_none,ot_none);
    code    : #1#217#8#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXCH;
    ops     : 2;
    optypes : (ot_fpu0,ot_fpureg,ot_none,ot_none);
    code    : #1#217#9#200;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FXTRACT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#244;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FYL2X;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#241;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_FYL2XP1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#217#249;
    flags   : [if_8086,if_fpu]
  ),
  (
    opcode  : A_HLT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#244;
    flags   : [if_8086,if_priv]
  ),
  (
    opcode  : A_IBTS;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#167#65;
    flags   : [if_386,if_sm,if_undoc]
  ),
  (
    opcode  : A_ICEBP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#241;
    flags   : [if_386]
  ),
  (
    opcode  : A_IDIV;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IDIV;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#2#15#175#72;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_IMUL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IMUL;
    ops     : 3;
    optypes : (ot_reg32 or ot_bits64,ot_rm_gpr,ot_immediate or ot_bits8 or ot_signed,ot_none);
    code    : #208#1#107#72#14;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_IMUL;
    ops     : 3;
    optypes : (ot_reg32 or ot_bits64,ot_rm_gpr,ot_immediate,ot_none);
    code    : #208#1#105#72#34;
    flags   : [if_386,if_sm,if_sd,if_ar2]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#107#64#13;
    flags   : [if_386]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg32,ot_immediate,ot_none,ot_none);
    code    : #213#1#105#64#33;
    flags   : [if_386,if_sd]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#105#64#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg64,ot_immediate,ot_none,ot_none);
    code    : #214#1#105#64#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_IMUL;
    ops     : 3;
    optypes : (ot_reg16,ot_rm_gpr,ot_immediate or ot_bits8 or ot_signed,ot_none);
    code    : #212#1#107#72#14;
    flags   : [if_186,if_sm]
  ),
  (
    opcode  : A_IMUL;
    ops     : 3;
    optypes : (ot_reg16,ot_rm_gpr,ot_immediate,ot_none);
    code    : #212#1#105#72#26;
    flags   : [if_186,if_sm,if_sw,if_ar2]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg16,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #212#1#107#64#13;
    flags   : [if_186]
  ),
  (
    opcode  : A_IMUL;
    ops     : 2;
    optypes : (ot_reg16,ot_immediate,ot_none,ot_none);
    code    : #212#1#105#64#25;
    flags   : [if_186,if_sw]
  ),
  (
    opcode  : A_IMUL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IN;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#228#21;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_IN;
    ops     : 2;
    optypes : (ot_reg_ax or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #208#1#229#21;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_IN;
    ops     : 2;
    optypes : (ot_reg_al,ot_reg_dx,ot_none,ot_none);
    code    : #1#236;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IN;
    ops     : 2;
    optypes : (ot_reg_ax or ot_bits32,ot_reg_dx,ot_none,ot_none);
    code    : #208#1#237;
    flags   : [if_8086]
  ),
  (
    opcode  : A_INC;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#255#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_INC;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#254#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_INSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#108;
    flags   : [if_186]
  ),
  (
    opcode  : A_INSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#109;
    flags   : [if_386]
  ),
  (
    opcode  : A_INSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#109;
    flags   : [if_186]
  ),
  (
    opcode  : A_INT;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #1#205#20;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_INT01;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#241;
    flags   : [if_386]
  ),
  (
    opcode  : A_INT1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#241;
    flags   : [if_386]
  ),
  (
    opcode  : A_INT03;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#204;
    flags   : [if_8086]
  ),
  (
    opcode  : A_INT3;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#204;
    flags   : [if_8086]
  ),
  (
    opcode  : A_INVD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#8;
    flags   : [if_486,if_priv]
  ),
  (
    opcode  : A_INVLPG;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#1#135;
    flags   : [if_486,if_priv]
  ),
  (
    opcode  : A_IRET;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #215#1#207;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IRETD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#207;
    flags   : [if_386]
  ),
  (
    opcode  : A_IRETW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#207;
    flags   : [if_8086]
  ),
  (
    opcode  : A_IRETQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#207;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_JECXZ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #201#1#227#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_JRCXZ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #1#227#40;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_immediate or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#235#40;
    flags   : [if_8086,if_pass2]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_immediate or ot_bits16 or ot_bits32,ot_none,ot_none,ot_none);
    code    : #208#1#233#52;
    flags   : [if_8086,if_pass2]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#255#132;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#255#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_immediate or ot_short,ot_none,ot_none,ot_none);
    code    : #1#235#40;
    flags   : [if_8086,if_pass2]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_immediate or ot_near,ot_none,ot_none,ot_none);
    code    : #208#1#233#52;
    flags   : [if_8086,if_pass2]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_memory or ot_near,ot_none,ot_none,ot_none);
    code    : #208#1#255#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_JMP;
    ops     : 1;
    optypes : (ot_memory or ot_far,ot_none,ot_none,ot_none);
    code    : #208#1#255#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LAHF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#159;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LAR;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#2#15#2#72;
    flags   : [if_286,if_prot,if_sm]
  ),
  (
    opcode  : A_LCALL;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#255#131;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_LCALL;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#255#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LEA;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_memory,ot_none,ot_none);
    code    : #208#1#141#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LEAVE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#201;
    flags   : [if_186]
  ),
  (
    opcode  : A_LFS;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32,ot_memory,ot_none,ot_none);
    code    : #208#2#15#180#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_LGDT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#1#130;
    flags   : [if_286,if_priv]
  ),
  (
    opcode  : A_LGS;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32,ot_memory,ot_none,ot_none);
    code    : #208#2#15#181#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_LIDT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#1#131;
    flags   : [if_286,if_priv]
  ),
  (
    opcode  : A_LJMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#255#133;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_LJMP;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#255#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LLDT;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #2#15#0#130;
    flags   : [if_286,if_prot,if_priv]
  ),
  (
    opcode  : A_LMSW;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #2#15#1#134;
    flags   : [if_286,if_priv]
  ),
  (
    opcode  : A_LOADALL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#7;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_LOADALL286;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#5;
    flags   : [if_286,if_undoc]
  ),
  (
    opcode  : A_LOCK;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#240;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_LODSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#172;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LODSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#173;
    flags   : [if_386]
  ),
  (
    opcode  : A_LODSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#173;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOP;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #202#1#226#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOP;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ecx or ot_bits64,ot_none,ot_none);
    code    : #201#1#226#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_LOOPE;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #202#1#225#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOPE;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ecx or ot_bits64,ot_none,ot_none);
    code    : #201#1#225#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_LOOPNE;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #202#1#224#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOPNE;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ecx or ot_bits64,ot_none,ot_none);
    code    : #201#1#224#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_LOOPNZ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #202#1#224#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOPNZ;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ecx or ot_bits64,ot_none,ot_none);
    code    : #201#1#224#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_LOOPZ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #202#1#225#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LOOPZ;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ecx or ot_bits64,ot_none,ot_none);
    code    : #201#1#225#40;
    flags   : [if_386]
  ),
  (
    opcode  : A_LSL;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#2#15#3#72;
    flags   : [if_286,if_prot,if_sm]
  ),
  (
    opcode  : A_LSS;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_memory,ot_none,ot_none);
    code    : #208#2#15#178#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_LTR;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #2#15#0#131;
    flags   : [if_286,if_prot,if_priv]
  ),
  (
    opcode  : A_MONITOR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#200;
    flags   : [if_prescott]
  ),
  (
    opcode  : A_MONITOR;
    ops     : 3;
    optypes : (ot_reg_eax,ot_reg_ecx,ot_reg_edx,ot_none);
    code    : #3#15#1#200;
    flags   : [if_prescott]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_ax,ot_none,ot_none);
    code    : #212#1#163#36;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_eax,ot_none,ot_none);
    code    : #213#1#163#36;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_rax,ot_none,ot_none);
    code    : #214#1#163#36;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#137#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_ax,ot_mem_offs,ot_none,ot_none);
    code    : #212#1#161#37;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_eax,ot_mem_offs,ot_none,ot_none);
    code    : #213#1#161#37;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_rax,ot_mem_offs,ot_none,ot_none);
    code    : #214#1#161#37;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#139#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg32,ot_immediate,ot_none,ot_none);
    code    : #213#8#184#33;
    flags   : [if_386,if_sd]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#199#128#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg64,ot_immediate,ot_none,ot_none);
    code    : #214#8#184#45;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#199#128#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#199#128#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg16,ot_immediate,ot_none,ot_none);
    code    : #212#8#184#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#199#128#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_al,ot_none,ot_none);
    code    : #1#162#36;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#136#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_al,ot_mem_offs,ot_none,ot_none);
    code    : #1#160#37;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#138#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg8,ot_immediate,ot_none,ot_none);
    code    : #8#176#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#198#128#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32,ot_reg_cs,ot_none,ot_none);
    code    : #208#1#140#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32,ot_reg_dess,ot_none,ot_none);
    code    : #208#1#140#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32,ot_reg_fsgs,ot_none,ot_none);
    code    : #208#1#140#65;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_dess,ot_rm_gpr or ot_bits16 or ot_bits32,ot_none,ot_none);
    code    : #209#1#142#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_fsgs,ot_rm_gpr or ot_bits16 or ot_bits32,ot_none,ot_none);
    code    : #209#1#142#72;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg64,ot_reg_cr4,ot_none,ot_none);
    code    : #2#15#32#132;
    flags   : [if_pent,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg64,ot_reg_creg,ot_none,ot_none);
    code    : #2#15#32#65;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg64,ot_reg_dreg,ot_none,ot_none);
    code    : #2#15#33#65;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg64,ot_reg_treg,ot_none,ot_none);
    code    : #2#15#36#65;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_cr4,ot_reg64,ot_none,ot_none);
    code    : #2#15#34#140;
    flags   : [if_pent,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_creg,ot_reg64,ot_none,ot_none);
    code    : #2#15#34#72;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_dreg,ot_reg64,ot_none,ot_none);
    code    : #2#15#35#72;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOV;
    ops     : 2;
    optypes : (ot_reg_treg,ot_reg64,ot_none,ot_none);
    code    : #2#15#38#72;
    flags   : [if_386,if_priv,if_x86_64]
  ),
  (
    opcode  : A_MOVD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #2#15#110#72;
    flags   : [if_pent,if_mmx,if_sd]
  ),
  (
    opcode  : A_MOVD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#126#65;
    flags   : [if_pent,if_mmx,if_sd]
  ),
  (
    opcode  : A_MOVD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #241#2#15#110#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#126#65;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#111#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_mmxrm,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#127#65;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #214#2#15#110#72;
    flags   : [if_x86_64,if_mmx]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_mmxreg,ot_none,ot_none);
    code    : #214#2#15#126#65;
    flags   : [if_x86_64,if_mmx]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#2#15#126#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#214#65;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_reg64,ot_none,ot_none);
    code    : #241#214#2#15#110#72;
    flags   : [if_willamette,if_sse2,if_x86_64]
  ),
  (
    opcode  : A_MOVQ;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg,ot_none,ot_none);
    code    : #241#214#2#15#126#65;
    flags   : [if_willamette,if_sse2,if_x86_64]
  ),
  (
    opcode  : A_MOVSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#164;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MOVSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#165;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOVSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#16#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVSD;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #220#2#15#17#65;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVSQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#165;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_MOVSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#165;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MOVSX;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_rm_gpr or ot_bits16,ot_none,ot_none);
    code    : #208#2#15#191#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOVSX;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #208#2#15#190#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOVZX;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_rm_gpr or ot_bits16,ot_none,ot_none);
    code    : #208#2#15#183#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOVZX;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #208#2#15#182#72;
    flags   : [if_386]
  ),
  (
    opcode  : A_MUL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MUL;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_MWAIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#201;
    flags   : [if_prescott]
  ),
  (
    opcode  : A_MWAIT;
    ops     : 2;
    optypes : (ot_reg_eax,ot_reg_ecx,ot_none,ot_none);
    code    : #3#15#1#201;
    flags   : [if_prescott]
  ),
  (
    opcode  : A_NEG;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_NEG;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_NOP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#144;
    flags   : [if_8086]
  ),
  (
    opcode  : A_NOT;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#1#247#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_NOT;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#246#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#9#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#11#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#8#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#10#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#129#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#13#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#13#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#13#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#129#129#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#129#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#129#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#13#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#129#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#12#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_OR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#129#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_al,ot_none,ot_none);
    code    : #1#230#20;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_ax,ot_none,ot_none);
    code    : #212#1#231#20;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_immediate,ot_reg_eax,ot_none,ot_none);
    code    : #213#1#231#20;
    flags   : [if_386,if_sb]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_reg_dx,ot_reg_al,ot_none,ot_none);
    code    : #1#238;
    flags   : [if_8086]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_reg_dx,ot_reg_ax,ot_none,ot_none);
    code    : #212#1#239;
    flags   : [if_8086]
  ),
  (
    opcode  : A_OUT;
    ops     : 2;
    optypes : (ot_reg_dx,ot_reg_eax,ot_none,ot_none);
    code    : #213#1#239;
    flags   : [if_386]
  ),
  (
    opcode  : A_OUTSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#110;
    flags   : [if_186]
  ),
  (
    opcode  : A_OUTSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#111;
    flags   : [if_386]
  ),
  (
    opcode  : A_OUTSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#111;
    flags   : [if_186]
  ),
  (
    opcode  : A_PACKSSDW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#107#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PACKSSDW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#107#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PACKSSWB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#99#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PACKSSWB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#99#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PACKUSWB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#103#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PACKUSWB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#103#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#252#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#252#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#254#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#254#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#236#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#236#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDSIW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#81#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PADDSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#237#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#237#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDUSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#220#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDUSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#220#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDUSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#221#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDUSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#221#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#253#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PADDW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#253#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PAND;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#219#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PAND;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#219#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PANDN;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#223#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PANDN;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#223#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PAVEB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#80#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PAVGUSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#191;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PCMPEQB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#116#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPEQB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#116#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PCMPEQD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#118#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPEQD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#118#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PCMPEQW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#117#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPEQW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#117#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PCMPGTB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#100#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPGTB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#100#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PCMPGTD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#102#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPGTD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#102#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PCMPGTW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#101#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PCMPGTW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#101#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PDISTIB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#84#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PF2ID;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#29;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFACC;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#174;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFADD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#158;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFCMPEQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#176;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFCMPGE;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#144;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFCMPGT;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#160;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFMAX;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#164;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFMIN;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#148;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFMUL;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#180;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFRCP;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#150;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFRCPIT1;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#166;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFRCPIT2;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#182;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFRSQIT1;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#167;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFRSQRT;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#151;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFSUB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#154;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFSUBR;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#170;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PI2FD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#13;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PMACHRIW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#94#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMADDWD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#245#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMADDWD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#245#72;
    flags   : [if_willamette,if_sm,if_sse2]
  ),
  (
    opcode  : A_PMAGW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#82#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMULHRIW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#93#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMULHRWA;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#183;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PMULHRWC;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#89#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMULHW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#229#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMULHW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#229#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMULLW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#213#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMULLW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#213#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMVGEZB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#92#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMVLZB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#91#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMVNZB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#90#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PMVZB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#88#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_reg16,ot_none,ot_none,ot_none);
    code    : #212#8#88;
    flags   : [if_8086]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_reg64,ot_none,ot_none,ot_none);
    code    : #221#8#88;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#143#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#143#128;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_reg_cs,ot_none,ot_none,ot_none);
    code    : #1#15;
    flags   : [if_8086,if_undoc]
  ),
  (
    opcode  : A_POP;
    ops     : 1;
    optypes : (ot_reg_fsgs,ot_none,ot_none,ot_none);
    code    : #1#15#5#221;
    flags   : [if_386]
  ),
  (
    opcode  : A_POPF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#157;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_POPFW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#157;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_POPFQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#157;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_POR;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#235#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_POR;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#235#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PREFETCH;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#13#128;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PREFETCHW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#13#129;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PSLLD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#242#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSLLD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#114#134#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#242#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSLLD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#114#134#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#115#135#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#243#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSLLQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#115#134#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#243#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSLLQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#115#134#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#241#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSLLW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#113#134#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSLLW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#241#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSLLW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#113#134#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRAD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#226#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSRAD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#114#132#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRAD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#226#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSRAD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#114#132#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRAW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#225#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSRAW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#113#132#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRAW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#225#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSRAW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#113#132#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#210#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSRLD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#114#130#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#210#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSRLD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#114#130#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#211#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSRLQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#115#130#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#211#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSRLQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#115#130#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#209#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSRLW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_immediate,ot_none,ot_none);
    code    : #2#15#113#130#21;
    flags   : [if_pent,if_mmx,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSRLW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#209#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSRLW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#113#130#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSUBB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#248#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#248#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#250#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#250#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#232#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#232#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBSIW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory,ot_none,ot_none);
    code    : #2#15#85#72;
    flags   : [if_pent,if_mmx,if_sm,if_cyrix]
  ),
  (
    opcode  : A_PSUBSIW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#85#72;
    flags   : [if_pent,if_mmx,if_cyrix]
  ),
  (
    opcode  : A_PSUBSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#233#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#233#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBUSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#216#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBUSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#216#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBUSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#217#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBUSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#217#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#249#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSUBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#249#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKHBW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#104#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKHBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#104#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKHDQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#106#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKHDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#106#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKHWD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#105#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKHWD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#105#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKLBW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#96#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKLBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#96#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKLDQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#98#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKLDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#98#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKLWD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#97#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PUNPCKLWD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#97#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_reg16,ot_none,ot_none,ot_none);
    code    : #212#8#80;
    flags   : [if_8086]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_reg64,ot_none,ot_none,ot_none);
    code    : #221#8#80;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#255#134;
    flags   : [if_8086]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits64,ot_none,ot_none,ot_none);
    code    : #221#1#255#134;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_immediate or ot_bits32,ot_none,ot_none,ot_none);
    code    : #213#1#104#32#221;
    flags   : [if_386]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_immediate or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#104#24#221;
    flags   : [if_186]
  ),
  (
    opcode  : A_PUSH;
    ops     : 1;
    optypes : (ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none,ot_none);
    code    : #1#106#12#221;
    flags   : [if_186]
  ),
  (
    opcode  : A_PUSHF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #215#1#156;
    flags   : [if_8086]
  ),
  (
    opcode  : A_PUSHFW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#156;
    flags   : [if_8086]
  ),
  (
    opcode  : A_PUSHFQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#156;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_PXOR;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#239#72;
    flags   : [if_pent,if_mmx,if_sm]
  ),
  (
    opcode  : A_PXOR;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#239#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#130#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#130;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#130#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#131#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#131;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RCR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#131#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_RDSHR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#54;
    flags   : [if_p6,if_cyrix,if_smm]
  ),
  (
    opcode  : A_RDMSR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#50;
    flags   : [if_pent,if_priv]
  ),
  (
    opcode  : A_RDPMC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#51;
    flags   : [if_p6]
  ),
  (
    opcode  : A_RDTSC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#49;
    flags   : [if_pent]
  ),
  (
    opcode  : A_REP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#243;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_REPE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#243;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_REPNE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#242;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_REPNZ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#242;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_REPZ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#243;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_RET;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #215#1#195;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RET;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #215#1#194#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #215#1#203;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RETF;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #215#1#202#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETN;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #215#1#195;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RETN;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #215#1#194#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#195;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RETW;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #212#1#194#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETFW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#203;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RETFW;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #212#1#202#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETNW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#195;
    flags   : [if_8086]
  ),
  (
    opcode  : A_RETNW;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #212#1#194#24;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_RETFD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#203;
    flags   : [if_386]
  ),
  (
    opcode  : A_RETFD;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #213#1#202#24;
    flags   : [if_386,if_sw]
  ),
  (
    opcode  : A_RETQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #221#1#195;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_RETQ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #221#1#194#24;
    flags   : [if_x86_64,if_sw]
  ),
  (
    opcode  : A_RETFQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#203;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_RETFQ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #214#1#202#24;
    flags   : [if_x86_64,if_sw]
  ),
  (
    opcode  : A_RETNQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #221#1#195;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_RETNQ;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #221#1#194#24;
    flags   : [if_x86_64,if_sw]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#128#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#128;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#128#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#129#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#129;
    flags   : [if_8086]
  ),
  (
    opcode  : A_ROR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#129#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_RSDC;
    ops     : 2;
    optypes : (ot_reg_sreg,ot_memory or ot_bits80,ot_none,ot_none);
    code    : #2#15#121#65;
    flags   : [if_486,if_cyrix,if_smm]
  ),
  (
    opcode  : A_RSLDT;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #2#15#123#128;
    flags   : [if_486,if_cyrix,if_smm]
  ),
  (
    opcode  : A_RSM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#170;
    flags   : [if_pent,if_smm]
  ),
  (
    opcode  : A_SAHF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#158;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#132#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#132#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#135#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#135;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SAR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#135#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#25#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#27#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#131#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#24#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#26#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#29#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#29#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#29#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #208#1#129#131#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#131#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#131#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#29#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #208#1#129#131#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#28#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_SBB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#131#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_SCASB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#1#174;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SCASD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#213#1#175;
    flags   : [if_386]
  ),
  (
    opcode  : A_SCASQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#214#1#175;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_SCASW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #218#212#1#175;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SEGCS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#46;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_SEGDS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#62;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_SEGES;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#38;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_SEGFS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#100;
    flags   : [if_386,if_pre]
  ),
  (
    opcode  : A_SEGGS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#101;
    flags   : [if_386,if_pre]
  ),
  (
    opcode  : A_SEGSS;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#54;
    flags   : [if_8086,if_pre]
  ),
  (
    opcode  : A_SGDT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#1#128;
    flags   : [if_286]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#132#21;
    flags   : [if_186,if_sw]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#132;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHL;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#132#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SHLD;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg16 or ot_bits32 or ot_bits64,ot_immediate,ot_none);
    code    : #209#2#15#164#65#22;
    flags   : [if_386,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_SHLD;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none);
    code    : #209#2#15#165#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_unity,ot_none,ot_none);
    code    : #208#1#209#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none,ot_none);
    code    : #208#1#211#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #208#1#193#133#21;
    flags   : [if_186,if_sw]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_unity,ot_none,ot_none);
    code    : #1#208#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg_cl,ot_none,ot_none);
    code    : #1#210#133;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SHR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#192#133#21;
    flags   : [if_186,if_sb]
  ),
  (
    opcode  : A_SHRD;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg16 or ot_bits32 or ot_bits64,ot_immediate,ot_none);
    code    : #209#2#15#172#65#22;
    flags   : [if_386,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_SHRD;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_reg16 or ot_bits32 or ot_bits64,ot_reg_cl,ot_none);
    code    : #209#2#15#173#65;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SIDT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#1#129;
    flags   : [if_286]
  ),
  (
    opcode  : A_SLDT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#0#128;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_SLDT;
    ops     : 1;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#2#15#0#128;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_SMI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#241;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_SMINT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#56;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_SMINTOLD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#126;
    flags   : [if_486,if_cyrix]
  ),
  (
    opcode  : A_SMSW;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#2#15#1#132;
    flags   : [if_286]
  ),
  (
    opcode  : A_STC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#249;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#253;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#251;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STOSB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#170;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STOSD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #213#1#171;
    flags   : [if_386]
  ),
  (
    opcode  : A_STOSW;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #212#1#171;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#0#129;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_STR;
    ops     : 1;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#2#15#0#129;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#41#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#43#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#40#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#42#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#133#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#45#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#45#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#45#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #208#1#129#133#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#133#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#133#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#45#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#133#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#44#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_SUB;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#133#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_SVDC;
    ops     : 2;
    optypes : (ot_memory or ot_bits80,ot_reg_sreg,ot_none,ot_none);
    code    : #2#15#120#65;
    flags   : [if_486,if_cyrix,if_smm]
  ),
  (
    opcode  : A_SVLDT;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #2#15#122#128;
    flags   : [if_486,if_cyrix,if_smm]
  ),
  (
    opcode  : A_SVTS;
    ops     : 1;
    optypes : (ot_memory or ot_bits80,ot_none,ot_none,ot_none);
    code    : #2#15#124#128;
    flags   : [if_486,if_cyrix,if_smm]
  ),
  (
    opcode  : A_SYSCALL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#5;
    flags   : [if_p6,if_prot,if_amd]
  ),
  (
    opcode  : A_SYSENTER;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#52;
    flags   : [if_p6,if_prot]
  ),
  (
    opcode  : A_SYSEXIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#53;
    flags   : [if_p6,if_prot,if_priv]
  ),
  (
    opcode  : A_SYSRET;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#7;
    flags   : [if_p6,if_prot,if_priv,if_amd]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#133#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_memory,ot_none,ot_none);
    code    : #208#1#133#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg8,ot_reg8,ot_none,ot_none);
    code    : #1#132#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#132#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#169#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#169#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#169#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#169#25;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#168#17;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#247#128#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#247#128#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #213#1#247#128#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#247#128#25;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#246#128#17;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits32,ot_none,ot_none);
    code    : #213#1#247#128#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits16,ot_none,ot_none);
    code    : #212#1#247#128#25;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_TEST;
    ops     : 2;
    optypes : (ot_memory,ot_immediate or ot_bits8,ot_none,ot_none);
    code    : #1#246#128#17;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_UD1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#185;
    flags   : [if_286,if_undoc]
  ),
  (
    opcode  : A_UD2;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#11;
    flags   : [if_286]
  ),
  (
    opcode  : A_UMOV;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#17#65;
    flags   : [if_386,if_undoc,if_sm]
  ),
  (
    opcode  : A_UMOV;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_memory,ot_none,ot_none);
    code    : #208#2#15#19#72;
    flags   : [if_386,if_undoc,if_sm]
  ),
  (
    opcode  : A_UMOV;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #2#15#16#65;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_UMOV;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #2#15#18#72;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_VERR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#0#132;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_VERR;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #2#15#0#132;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_VERR;
    ops     : 1;
    optypes : (ot_reg16,ot_none,ot_none,ot_none);
    code    : #2#15#0#132;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_VERW;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#0#133;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_VERW;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #2#15#0#133;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_VERW;
    ops     : 1;
    optypes : (ot_reg16,ot_none,ot_none,ot_none);
    code    : #2#15#0#133;
    flags   : [if_286,if_prot]
  ),
  (
    opcode  : A_WAIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#155;
    flags   : [if_8086]
  ),
  (
    opcode  : A_WBINVD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#9;
    flags   : [if_486,if_priv]
  ),
  (
    opcode  : A_WRSHR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#55;
    flags   : [if_p6,if_cyrix,if_smm]
  ),
  (
    opcode  : A_WRMSR;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#48;
    flags   : [if_pent,if_priv]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_memory,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#193#65;
    flags   : [if_486,if_sm]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #2#15#192#65;
    flags   : [if_486]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_reg8,ot_reg8,ot_none,ot_none);
    code    : #2#15#192#65;
    flags   : [if_486]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_reg16,ot_reg16,ot_none,ot_none);
    code    : #208#2#15#193#65;
    flags   : [if_486]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_reg32,ot_reg32,ot_none,ot_none);
    code    : #209#2#15#193#65;
    flags   : [if_486]
  ),
  (
    opcode  : A_XADD;
    ops     : 2;
    optypes : (ot_reg64,ot_reg64,ot_none,ot_none);
    code    : #208#2#15#193#65;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_XBTS;
    ops     : 2;
    optypes : (ot_reg16,ot_memory,ot_none,ot_none);
    code    : #212#2#15#166#72;
    flags   : [if_386,if_sw,if_undoc]
  ),
  (
    opcode  : A_XBTS;
    ops     : 2;
    optypes : (ot_reg16,ot_reg16,ot_none,ot_none);
    code    : #212#2#15#166#72;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_XBTS;
    ops     : 2;
    optypes : (ot_reg32,ot_memory,ot_none,ot_none);
    code    : #213#2#15#166#72;
    flags   : [if_386,if_sd,if_undoc]
  ),
  (
    opcode  : A_XBTS;
    ops     : 2;
    optypes : (ot_reg32,ot_reg32,ot_none,ot_none);
    code    : #213#2#15#166#72;
    flags   : [if_386,if_undoc]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg_ax,ot_reg16,ot_none,ot_none);
    code    : #212#9#144;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg_eax,ot_reg32,ot_none,ot_none);
    code    : #213#9#144;
    flags   : [if_386]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg_rax,ot_reg64,ot_none,ot_none);
    code    : #214#9#144;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg16,ot_reg_ax,ot_none,ot_none);
    code    : #212#8#144;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg32,ot_reg_eax,ot_none,ot_none);
    code    : #213#8#144;
    flags   : [if_386]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg64,ot_reg_rax,ot_none,ot_none);
    code    : #214#8#144;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#135#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_memory,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#135#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#134#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XCHG;
    ops     : 2;
    optypes : (ot_memory or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#134#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XLAT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#215;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XLAT;
    ops     : 1;
    optypes : (ot_memory or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#215;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XLATB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#215;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#1#49#65;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#51#72;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_reg8,ot_none,ot_none);
    code    : #1#48#65;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg8,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #1#50#72;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16 or ot_bits32 or ot_bits64,ot_immediate or ot_bits8 or ot_signed,ot_none,ot_none);
    code    : #208#1#131#134#13;
    flags   : [if_8086]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg_eax,ot_immediate,ot_none,ot_none);
    code    : #213#1#53#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#53#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg_rax,ot_immediate,ot_none,ot_none);
    code    : #214#1#53#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_immediate,ot_none,ot_none);
    code    : #208#1#129#134#33;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate or ot_signed,ot_none,ot_none);
    code    : #214#1#129#134#173;
    flags   : [if_x86_64,if_sd]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_immediate,ot_none,ot_none);
    code    : #214#1#129#134#173;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg_ax,ot_immediate,ot_none,ot_none);
    code    : #212#1#53#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits16,ot_immediate,ot_none,ot_none);
    code    : #212#1#129#134#25;
    flags   : [if_8086,if_sw]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_reg_al,ot_immediate,ot_none,ot_none);
    code    : #1#52#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_XOR;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits8,ot_immediate,ot_none,ot_none);
    code    : #1#128#134#17;
    flags   : [if_8086,if_sb]
  ),
  (
    opcode  : A_XSTORE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#167#192;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XCRYPTECB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#167#200;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XCRYPTCBC;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#167#208;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XCRYPTCFB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#167#224;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XCRYPTOFB;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#167#232;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_CMOVcc;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#1#15#11#64#72;
    flags   : [if_p6,if_sm]
  ),
  (
    opcode  : A_Jcc;
    ops     : 1;
    optypes : (ot_immediate or ot_bits8,ot_none,ot_none,ot_none);
    code    : #11#112#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_Jcc;
    ops     : 1;
    optypes : (ot_immediate or ot_bits16 or ot_bits32,ot_none,ot_none,ot_none);
    code    : #208#1#15#11#128#52;
    flags   : [if_386,if_pass2]
  ),
  (
    opcode  : A_Jcc;
    ops     : 1;
    optypes : (ot_immediate or ot_short,ot_none,ot_none,ot_none);
    code    : #11#112#40;
    flags   : [if_8086]
  ),
  (
    opcode  : A_Jcc;
    ops     : 1;
    optypes : (ot_immediate or ot_near,ot_none,ot_none,ot_none);
    code    : #208#1#15#11#128#52;
    flags   : [if_386,if_pass2]
  ),
  (
    opcode  : A_SETcc;
    ops     : 1;
    optypes : (ot_rm_gpr or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#15#11#144#128;
    flags   : [if_386]
  ),
  (
    opcode  : A_MOVS;
    ops     : 2;
    optypes : (ot_memory or ot_bits8,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #1#164;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOVS;
    ops     : 2;
    optypes : (ot_memory or ot_bits16,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #212#1#165;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_MOVS;
    ops     : 2;
    optypes : (ot_memory or ot_bits32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #213#1#165;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_MOVS;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #214#1#165;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_CMPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits8,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #218#1#166;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_CMPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits16,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #218#212#1#167;
    flags   : [if_8086,if_sm]
  ),
  (
    opcode  : A_CMPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #218#213#1#167;
    flags   : [if_386,if_sm]
  ),
  (
    opcode  : A_CMPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #214#1#167;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_SCAS;
    ops     : 1;
    optypes : (ot_memory or ot_bits8,ot_none,ot_none,ot_none);
    code    : #218#1#174;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SCAS;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #218#212#1#175;
    flags   : [if_8086]
  ),
  (
    opcode  : A_SCAS;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #218#213#1#175;
    flags   : [if_386]
  ),
  (
    opcode  : A_SCAS;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #218#214#1#175;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_LODS;
    ops     : 1;
    optypes : (ot_memory or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#172;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LODS;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#173;
    flags   : [if_8086]
  ),
  (
    opcode  : A_LODS;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #213#1#173;
    flags   : [if_386]
  ),
  (
    opcode  : A_LODS;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #214#1#173;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_STOS;
    ops     : 1;
    optypes : (ot_memory or ot_bits8,ot_none,ot_none,ot_none);
    code    : #1#170;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STOS;
    ops     : 1;
    optypes : (ot_memory or ot_bits16,ot_none,ot_none,ot_none);
    code    : #212#1#171;
    flags   : [if_8086]
  ),
  (
    opcode  : A_STOS;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #213#1#171;
    flags   : [if_386]
  ),
  (
    opcode  : A_STOS;
    ops     : 1;
    optypes : (ot_memory or ot_bits64,ot_none,ot_none,ot_none);
    code    : #214#1#171;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_INS;
    ops     : 2;
    optypes : (ot_memory or ot_bits8,ot_reg_dx,ot_none,ot_none);
    code    : #1#108;
    flags   : [if_186]
  ),
  (
    opcode  : A_INS;
    ops     : 2;
    optypes : (ot_memory or ot_bits16,ot_reg_dx,ot_none,ot_none);
    code    : #212#1#109;
    flags   : [if_186]
  ),
  (
    opcode  : A_INS;
    ops     : 2;
    optypes : (ot_memory or ot_bits32,ot_reg_dx,ot_none,ot_none);
    code    : #213#1#109;
    flags   : [if_386]
  ),
  (
    opcode  : A_OUTS;
    ops     : 2;
    optypes : (ot_reg_dx,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #1#110;
    flags   : [if_186]
  ),
  (
    opcode  : A_OUTS;
    ops     : 2;
    optypes : (ot_reg_dx,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #212#1#111;
    flags   : [if_186]
  ),
  (
    opcode  : A_OUTS;
    ops     : 2;
    optypes : (ot_reg_dx,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #213#1#111;
    flags   : [if_386]
  ),
  (
    opcode  : A_ADDPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#88#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_ADDSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#88#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_ANDNPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#85#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_ANDPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#84#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPEQPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#0;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPEQSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#0;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPLEPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#2;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPLESS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#2;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPLTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#1;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPLTSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#1;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNEQPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#4;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNEQSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#4;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNLEPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#6;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNLESS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#6;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNLTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#5;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPNLTSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#5;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPORDPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#7;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPORDSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#7;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPUNORDPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#194#72#1#3;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPUNORDSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#194#72#1#3;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CMPPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #217#2#15#194#72#18;
    flags   : [if_katmai,if_sse,if_sb,if_ar2]
  ),
  (
    opcode  : A_CMPSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #222#219#2#15#194#72#18;
    flags   : [if_katmai,if_sse,if_sb,if_ar2]
  ),
  (
    opcode  : A_COMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#2#15#47#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CVTPI2PS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#2#15#42#72;
    flags   : [if_katmai,if_sse,if_mmx]
  ),
  (
    opcode  : A_CVTPS2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #217#2#15#45#72;
    flags   : [if_katmai,if_sse,if_mmx]
  ),
  (
    opcode  : A_CVTPS2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_xmmreg,ot_none,ot_none);
    code    : #217#2#15#45#72;
    flags   : [if_katmai,if_sse,if_mmx]
  ),
  (
    opcode  : A_CVTSI2SS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #219#209#2#15#42#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CVTSI2SS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #219#209#2#15#42#72;
    flags   : [if_katmai,if_sse,if_x86_64]
  ),
  (
    opcode  : A_CVTSS2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#208#2#15#45#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CVTSS2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #219#208#2#15#45#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CVTTPS2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #217#2#15#44#72;
    flags   : [if_katmai,if_sse,if_mmx]
  ),
  (
    opcode  : A_CVTTPS2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_xmmreg,ot_none,ot_none);
    code    : #217#2#15#44#72;
    flags   : [if_katmai,if_sse,if_mmx]
  ),
  (
    opcode  : A_CVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#208#2#15#44#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_CVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #219#208#2#15#44#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_DIVPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#94#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_DIVSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#94#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_LDMXCSR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#130;
    flags   : [if_katmai,if_sse,if_sd]
  ),
  (
    opcode  : A_MAXPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#95#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MAXSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#95#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MINPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#93#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MINSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#93#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVAPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#40#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVAPS;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#41#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVHPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #2#15#22#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVHPS;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#23#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVLHPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#22#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVLPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #2#15#18#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVLPS;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#19#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVHLPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#18#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVMSKPS;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#80#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVNTPS;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#43#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#16#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#2#15#16#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#17#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVSS;
    ops     : 2;
    optypes : (ot_memory or ot_bits32,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#17#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVUPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#16#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MOVUPS;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #217#2#15#17#65;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MULPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#89#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_MULSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#89#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_ORPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#86#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_RCPPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#83#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_RCPSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#83#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_RSQRTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#82#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_RSQRTSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#82#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_SHUFPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #2#15#198#72#18;
    flags   : [if_katmai,if_sse,if_sb,if_ar2]
  ),
  (
    opcode  : A_SQRTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#81#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_SQRTSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#81#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_STMXCSR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#131;
    flags   : [if_katmai,if_sse,if_sd]
  ),
  (
    opcode  : A_SUBPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #217#2#15#92#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_SUBSS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#219#2#15#92#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_UCOMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #222#2#15#46#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_UNPCKHPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#21#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_UNPCKLPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#20#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_XORPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#87#72;
    flags   : [if_katmai,if_sse]
  ),
  (
    opcode  : A_FXRSTOR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#129;
    flags   : [if_p6,if_sse,if_fpu]
  ),
  (
    opcode  : A_FXSAVE;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#128;
    flags   : [if_p6,if_sse,if_fpu]
  ),
  (
    opcode  : A_PREFETCHNTA;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#24#128;
    flags   : [if_katmai]
  ),
  (
    opcode  : A_PREFETCHT0;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#24#129;
    flags   : [if_katmai]
  ),
  (
    opcode  : A_PREFETCHT1;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#24#130;
    flags   : [if_katmai]
  ),
  (
    opcode  : A_PREFETCHT2;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#24#131;
    flags   : [if_katmai]
  ),
  (
    opcode  : A_SFENCE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#174#248;
    flags   : [if_katmai]
  ),
  (
    opcode  : A_MASKMOVQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#247#72;
    flags   : [if_katmai,if_mmx]
  ),
  (
    opcode  : A_MOVNTQ;
    ops     : 2;
    optypes : (ot_memory,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#231#65;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PAVGB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#224#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PAVGB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#224#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PAVGW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#227#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PAVGW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#227#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PEXTRW;
    ops     : 3;
    optypes : (ot_reg32,ot_mmxreg,ot_immediate,ot_none);
    code    : #2#15#197#72#18;
    flags   : [if_katmai,if_mmx,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRW;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#2#15#197#72#22;
    flags   : [if_sse41]
  ),
  (
    opcode  : A_PEXTRW;
    ops     : 3;
    optypes : (ot_memory or ot_bits32,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#21#65#22;
    flags   : [if_sse41]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_mmxreg,ot_reg16,ot_immediate,ot_none);
    code    : #2#15#196#72#18;
    flags   : [if_katmai,if_mmx,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_mmxreg,ot_reg32,ot_immediate,ot_none);
    code    : #2#15#196#72#18;
    flags   : [if_katmai,if_mmx,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_mmxreg,ot_memory,ot_immediate,ot_none);
    code    : #2#15#196#72#18;
    flags   : [if_katmai,if_mmx,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_mmxreg,ot_memory or ot_bits16,ot_immediate,ot_none);
    code    : #2#15#196#72#18;
    flags   : [if_katmai,if_mmx,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_reg16,ot_immediate,ot_none);
    code    : #241#2#15#196#72#22;
    flags   : [if_willamette,if_sse2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_reg32,ot_immediate,ot_none);
    code    : #241#2#15#196#72#22;
    flags   : [if_willamette,if_sse2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory,ot_immediate,ot_none);
    code    : #241#2#15#196#72#22;
    flags   : [if_willamette,if_sse2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits16,ot_immediate,ot_none);
    code    : #241#2#15#196#72#22;
    flags   : [if_willamette,if_sse2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PMAXSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#238#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMAXSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#238#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMAXUB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#222#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMAXUB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#222#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMINSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#234#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMINSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#234#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMINUB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#218#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMINUB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#218#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMOVMSKB;
    ops     : 2;
    optypes : (ot_reg32,ot_mmxreg,ot_none,ot_none);
    code    : #2#15#215#72;
    flags   : [if_katmai,if_mmx]
  ),
  (
    opcode  : A_PMOVMSKB;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#215#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_PMULHUW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#228#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMULHUW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#228#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSADBW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#246#72;
    flags   : [if_katmai,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSADBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#246#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSHUFW;
    ops     : 3;
    optypes : (ot_mmxreg,ot_mmxrm,ot_immediate,ot_none);
    code    : #2#15#112#72#18;
    flags   : [if_katmai,if_mmx,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PFNACC;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#138;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PFPNACC;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#142;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PI2FW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#12;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PF2IW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#28;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_PSWAPD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#15#72#1#187;
    flags   : [if_pent,if_3dnow,if_sm]
  ),
  (
    opcode  : A_FFREEP;
    ops     : 1;
    optypes : (ot_fpureg,ot_none,ot_none,ot_none);
    code    : #1#223#8#192;
    flags   : [if_pent,if_3dnow,if_fpu]
  ),
  (
    opcode  : A_MASKMOVDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#247#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CLFLUSH;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#135;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVNTDQ;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#231#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVNTI;
    ops     : 2;
    optypes : (ot_memory,ot_reg32 or ot_bits64,ot_none,ot_none);
    code    : #208#2#15#195#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVNTPD;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#43#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PAUSE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#1#144;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_LFENCE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#174#232;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MFENCE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#174#240;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVDQA;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#127#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVDQA;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#111#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVDQU;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#127#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#2#15#111#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVDQ2Q;
    ops     : 2;
    optypes : (ot_mmxreg,ot_xmmreg,ot_none,ot_none);
    code    : #220#2#15#214#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVQ2DQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_mmxreg,ot_none,ot_none);
    code    : #219#2#15#214#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_PADDQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#212#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PADDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#212#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMULUDQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#244#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PMULUDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#244#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSHUFD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#2#15#112#72#18;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PSHUFHW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #219#2#15#112#72#18;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PSHUFLW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #220#2#15#112#72#18;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PSRLDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_immediate,ot_none,ot_none);
    code    : #241#2#15#115#131#21;
    flags   : [if_willamette,if_sse2,if_sb,if_ar1]
  ),
  (
    opcode  : A_PSUBQ;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #2#15#251#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PSUBQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#251#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKHQDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#109#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_PUNPCKLQDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#108#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_ADDPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#88#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_ADDSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#221#220#2#15#88#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_ANDNPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#85#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_ANDPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#84#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPEQPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#0;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPEQSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#0;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPLEPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#2;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPLESD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#2;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPLTPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#1;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPLTSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#1;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPNEQPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#4;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPNEQSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#4;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPNLEPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#6;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPNLESD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#6;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPNLTPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#5;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPNLTSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#5;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPORDPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#7;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPORDSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#7;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPUNORDPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#194#72#1#3;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CMPUNORDSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#194#72#1#3;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CMPPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#2#15#194#72#22;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_COMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#241#2#15#47#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#230#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #219#2#15#230#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTDQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #2#15#91#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#230#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTPD2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#45#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#90#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTPI2PD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_mmxrm,ot_none,ot_none);
    code    : #241#2#15#42#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#91#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTPS2PD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTPS2PD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #220#2#15#45#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#2#15#45#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg,ot_none,ot_none);
    code    : #220#208#2#15#45#72;
    flags   : [if_willamette,if_sse2,if_x86_64]
  ),
  (
    opcode  : A_CVTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#208#2#15#45#72;
    flags   : [if_willamette,if_sse2,if_x86_64]
  ),
  (
    opcode  : A_CVTSD2SS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #220#2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSD2SS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSI2SD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #220#2#15#42#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSI2SD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #214#220#209#2#15#42#72;
    flags   : [if_willamette,if_sse2,if_x86_64]
  ),
  (
    opcode  : A_CVTSS2SD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTSS2SD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#2#15#90#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTTPD2PI;
    ops     : 2;
    optypes : (ot_mmxreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#44#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#230#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#2#15#91#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_CVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #220#208#2#15#44#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_CVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#208#2#15#44#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_DIVPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#94#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_DIVSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#94#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MAXPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#95#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MAXSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#95#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MINPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#93#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MINSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#93#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVAPD;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#41#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVAPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#40#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVHPD;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#23#65;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVHPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #241#2#15#22#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVLPD;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#19#65;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVLPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #241#2#15#18#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVMSKPD;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#80#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_MOVUPD;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#17#65;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MOVUPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#16#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MULPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#89#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_MULSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#89#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_ORPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#86#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_SHUFPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#2#15#198#72#22;
    flags   : [if_willamette,if_sse2,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_SQRTPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#81#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_SQRTSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#81#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_SUBPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#92#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_SUBSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#220#2#15#92#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_UCOMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #223#241#2#15#46#72;
    flags   : [if_willamette,if_sse2]
  ),
  (
    opcode  : A_UNPCKHPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#21#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_UNPCKLPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#20#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_XORPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#87#72;
    flags   : [if_willamette,if_sse2,if_sm]
  ),
  (
    opcode  : A_ADDSUBPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#208#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_ADDSUBPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#208#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_HADDPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#124#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_HADDPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#124#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_HSUBPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#2#15#125#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_HSUBPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#125#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_LDDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #220#2#15#240#72;
    flags   : [if_prescott,if_sse3]
  ),
  (
    opcode  : A_MOVDDUP;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #220#2#15#18#72;
    flags   : [if_prescott,if_sse3]
  ),
  (
    opcode  : A_MOVSHDUP;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#2#15#22#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_MOVSLDUP;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#2#15#18#72;
    flags   : [if_prescott,if_sse3,if_sm]
  ),
  (
    opcode  : A_VMREAD;
    ops     : 2;
    optypes : (ot_reg32,ot_reg32,ot_none,ot_none);
    code    : #2#15#120#65;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMREAD;
    ops     : 2;
    optypes : (ot_memory,ot_reg32,ot_none,ot_none);
    code    : #2#15#120#65;
    flags   : [if_386,if_priv,if_prot,if_sm]
  ),
  (
    opcode  : A_VMWRITE;
    ops     : 2;
    optypes : (ot_reg32,ot_reg32,ot_none,ot_none);
    code    : #2#15#121#72;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMWRITE;
    ops     : 2;
    optypes : (ot_reg32,ot_memory,ot_none,ot_none);
    code    : #2#15#121#72;
    flags   : [if_386,if_priv,if_prot,if_sm]
  ),
  (
    opcode  : A_VMCALL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#193;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMLAUNCH;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#194;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMRESUME;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#195;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMXOFF;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#196;
    flags   : [if_386,if_priv,if_prot]
  ),
  (
    opcode  : A_VMXON;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #219#2#15#199#134;
    flags   : [if_priv,if_prot]
  ),
  (
    opcode  : A_VMCLEAR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #241#2#15#199#134;
    flags   : [if_priv,if_prot]
  ),
  (
    opcode  : A_VMPTRLD;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#199#134;
    flags   : [if_priv,if_prot]
  ),
  (
    opcode  : A_VMPTRST;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#199#135;
    flags   : [if_priv,if_prot]
  ),
  (
    opcode  : A_VMRUN;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#216;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_VMMCALL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#217;
    flags   : [if_386,if_svm]
  ),
  (
    opcode  : A_VMLOAD;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#218;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_VMSAVE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#219;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_STGI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#220;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_CLGI;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#221;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_SKINIT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#222;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_INVLPGA;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#223;
    flags   : [if_386,if_svm,if_priv,if_prot]
  ),
  (
    opcode  : A_MONTMUL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#166#192;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XSHA1;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#166#200;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_XSHA256;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #219#3#15#166#208;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_DMINT;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#57;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_RDM;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #2#15#58;
    flags   : [if_p6,if_cyrix]
  ),
  (
    opcode  : A_MOVABS;
    ops     : 2;
    optypes : (ot_reg_al,ot_mem_offs,ot_none,ot_none);
    code    : #1#160#37;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOVABS;
    ops     : 2;
    optypes : (ot_reg_ax or ot_bits32 or ot_bits64,ot_mem_offs,ot_none,ot_none);
    code    : #208#1#161#37;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOVABS;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_al,ot_none,ot_none);
    code    : #1#162#36;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOVABS;
    ops     : 2;
    optypes : (ot_mem_offs,ot_reg_ax or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #209#163#36;
    flags   : [if_x86_64,if_sm]
  ),
  (
    opcode  : A_MOVABS;
    ops     : 2;
    optypes : (ot_reg64,ot_immediate,ot_none,ot_none);
    code    : #214#8#184#45;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_MOVSXD;
    ops     : 2;
    optypes : (ot_reg64,ot_memory,ot_none,ot_none);
    code    : #214#1#99#72;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_MOVSXD;
    ops     : 2;
    optypes : (ot_reg64,ot_reg32,ot_none,ot_none);
    code    : #214#1#99#72;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_CQO;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#153;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_CDQE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#152;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_CMPXCHG16B;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #214#2#15#199#129;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_MOVNTSS;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #222#219#2#15#43#65;
    flags   : [if_sse4,if_sd]
  ),
  (
    opcode  : A_MOVNTSD;
    ops     : 2;
    optypes : (ot_memory,ot_xmmreg,ot_none,ot_none);
    code    : #223#220#213#2#15#43#65;
    flags   : [if_sse4]
  ),
  (
    opcode  : A_INSERTQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #220#2#15#121#72;
    flags   : [if_sse4]
  ),
  (
    opcode  : A_INSERTQ;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_immediate,ot_immediate);
    code    : #220#2#15#120#72#22#23;
    flags   : [if_sse4,if_sb]
  ),
  (
    opcode  : A_EXTRQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_immediate,ot_immediate,ot_none);
    code    : #241#2#15#120#128#21#22;
    flags   : [if_sse4,if_sb]
  ),
  (
    opcode  : A_EXTRQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #241#2#15#121#72;
    flags   : [if_sse4]
  ),
  (
    opcode  : A_LZCNT;
    ops     : 2;
    optypes : (ot_reg16,ot_rm_gpr,ot_none,ot_none);
    code    : #208#219#2#15#189#72;
    flags   : [if_386,if_sm,if_sse4]
  ),
  (
    opcode  : A_LZCNT;
    ops     : 2;
    optypes : (ot_reg32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #209#219#2#15#189#72;
    flags   : [if_386,if_sm,if_sse4]
  ),
  (
    opcode  : A_PABSB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#28#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PABSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#28#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PABSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#29#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PABSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#29#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PABSD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#30#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PABSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#30#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PALIGNR;
    ops     : 3;
    optypes : (ot_mmxreg,ot_mmxrm,ot_immediate,ot_none);
    code    : #217#3#15#58#15#72#22;
    flags   : [if_ssse3,if_mmx,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PALIGNR;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#15#72#22;
    flags   : [if_ssse3,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PHADDW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#1#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHADDW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#1#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PHADDD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#2#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHADDD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#2#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PHADDSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#3#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHADDSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#3#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PHSUBW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#5#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHSUBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#5#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PHSUBD;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#6#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHSUBD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#6#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PHSUBSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#7#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PHSUBSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#7#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PMADDUBSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#4#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMADDUBSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#4#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PMULHRSW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#11#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PMULHRSW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#11#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PSHUFB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#0#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSHUFB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#0#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PSIGNB;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#8#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSIGNB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#8#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PSIGNW;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#9#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSIGNW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#9#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_PSIGND;
    ops     : 2;
    optypes : (ot_mmxreg,ot_mmxrm,ot_none,ot_none);
    code    : #217#3#15#56#10#72;
    flags   : [if_ssse3,if_mmx,if_sm]
  ),
  (
    opcode  : A_PSIGND;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#10#72;
    flags   : [if_ssse3,if_sm]
  ),
  (
    opcode  : A_BLENDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#12#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_BLENDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#13#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_BLENDVPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#20#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_BLENDVPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#21#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_DPPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#64#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_DPPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#65#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_EXTRACTPS;
    ops     : 3;
    optypes : (ot_memory,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#213#3#15#58#23#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_EXTRACTPS;
    ops     : 3;
    optypes : (ot_reg32 or ot_bits64,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#23#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_INSERTPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#33#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_MOVNTDQA;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory,ot_none,ot_none);
    code    : #241#3#15#56#42#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_MPSADBW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#66#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PACKUSDW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#43#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PBLENDVB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#16#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PBLENDW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#14#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PCMPEQQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#41#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PEXTRB;
    ops     : 3;
    optypes : (ot_reg32 or ot_bits64,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#20#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRB;
    ops     : 3;
    optypes : (ot_memory or ot_bits8,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#20#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRD;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#22#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRD;
    ops     : 3;
    optypes : (ot_memory or ot_bits32,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#3#15#58#22#65#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRQ;
    ops     : 3;
    optypes : (ot_reg64,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#214#3#15#58#22#65#22;
    flags   : [if_sse41,if_x86_64,if_sb,if_ar2]
  ),
  (
    opcode  : A_PEXTRQ;
    ops     : 3;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_immediate,ot_none);
    code    : #241#214#3#15#58#22#65#22;
    flags   : [if_sse41,if_x86_64,if_sb,if_ar2]
  ),
  (
    opcode  : A_PHMINPOSUW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#65#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PINSRB;
    ops     : 3;
    optypes : (ot_xmmreg,ot_reg32 or ot_bits64,ot_immediate,ot_none);
    code    : #241#3#15#58#32#72#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRB;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits8,ot_immediate,ot_none);
    code    : #241#3#15#58#32#72#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_reg32,ot_immediate,ot_none);
    code    : #241#3#15#58#34#72#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits32,ot_immediate,ot_none);
    code    : #241#3#15#58#34#72#22;
    flags   : [if_sse41,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_reg64,ot_immediate,ot_none);
    code    : #241#214#3#15#58#34#72#22;
    flags   : [if_sse41,if_x86_64,if_sb,if_ar2]
  ),
  (
    opcode  : A_PINSRQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_immediate,ot_none);
    code    : #241#214#3#15#58#34#72#22;
    flags   : [if_sse41,if_x86_64,if_sb,if_ar2]
  ),
  (
    opcode  : A_PMAXSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#60#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMAXSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#61#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMAXUD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#63#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMAXUW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#62#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMINSB;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#56#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMINSD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#57#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMINUW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#58#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMINUD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#59#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#32#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXBD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#33#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXBQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#34#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXWD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#35#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXWQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#36#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVSXDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#37#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXBW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#48#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXBD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#49#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXBQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#50#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXWD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#51#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXWQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#52#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMOVZXDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#53#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMULDQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#40#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PMULLD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#64#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_PTEST;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#23#72;
    flags   : [if_sse41,if_sm]
  ),
  (
    opcode  : A_ROUNDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#8#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_ROUNDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#9#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_ROUNDSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #222#241#3#15#58#10#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_ROUNDSD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #223#241#3#15#58#11#72#22;
    flags   : [if_sse41,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_CRC32;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #220#3#15#56#240#72;
    flags   : [if_sse42]
  ),
  (
    opcode  : A_CRC32;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits16 or ot_bits32,ot_none,ot_none);
    code    : #209#220#3#15#56#241#72;
    flags   : [if_sse42]
  ),
  (
    opcode  : A_CRC32;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits8,ot_none,ot_none);
    code    : #220#214#3#15#56#240#72;
    flags   : [if_sse42,if_x86_64]
  ),
  (
    opcode  : A_CRC32;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #220#214#3#15#56#241#72;
    flags   : [if_sse42,if_x86_64]
  ),
  (
    opcode  : A_PCMPESTRI;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#97#72#22;
    flags   : [if_sse42,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PCMPESTRM;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#96#72#22;
    flags   : [if_sse42,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PCMPISTRI;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#99#72#22;
    flags   : [if_sse42,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PCMPISTRM;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#98#72#22;
    flags   : [if_sse42,if_sm2,if_sb,if_ar2]
  ),
  (
    opcode  : A_PCMPGTQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#55#72;
    flags   : [if_sse42,if_sm]
  ),
  (
    opcode  : A_POPCNT;
    ops     : 2;
    optypes : (ot_reg16,ot_rm_gpr or ot_bits16,ot_none,ot_none);
    code    : #219#208#2#15#184#72;
    flags   : [if_386,if_sm,if_sse4]
  ),
  (
    opcode  : A_POPCNT;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #219#208#2#15#184#72;
    flags   : [if_386,if_sm,if_sse4]
  ),
  (
    opcode  : A_POPCNT;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #219#208#2#15#184#72;
    flags   : [if_386,if_sm,if_sse4,if_x86_64]
  ),
  (
    opcode  : A_AESENC;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#220#72;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_AESENCLAST;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#221#72;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_AESDEC;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#222#72;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_AESDECLAST;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#223#72;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_AESIMC;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#3#15#56#219#72;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_AESKEYGENASSIST;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #241#3#15#58#223#72#22;
    flags   : [if_sse4,if_sb,if_ar2]
  ),
  (
    opcode  : A_RDTSCP;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#249;
    flags   : [if_sse4,if_sm]
  ),
  (
    opcode  : A_STOSQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#171;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_LODSQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#173;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_CMPSQ;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #214#1#167;
    flags   : [if_x86_64]
  ),
  (
    opcode  : A_MOVBE;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_memory or ot_bits16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #208#3#15#56#240#72;
    flags   : [if_movbe,if_sm]
  ),
  (
    opcode  : A_MOVBE;
    ops     : 2;
    optypes : (ot_memory or ot_bits16 or ot_bits32 or ot_bits64,ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none);
    code    : #209#3#15#56#241#65;
    flags   : [if_movbe,if_sm]
  ),
  (
    opcode  : A_PCLMULQDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#3#15#58#68#72#22;
    flags   : [if_clmul,if_sandybridge]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#248#1#88#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#248#1#88#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#88#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VADDSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #223#220#232#234#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VADDSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VADDSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VADDSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #219#232#242#248#1#88#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VADDSUBPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#208#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VADDSUBPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#208#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VADDSUBPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #220#242#244#248#1#208#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VADDSUBPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #220#242#248#1#208#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VAESDEC;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#222#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESDEC;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#222#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VAESDEC;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#222#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESDECLAST;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESDECLAST;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#223#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VAESDECLAST;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESENC;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#220#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESENC;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#220#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VAESENC;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#220#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESENCLAST;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#221#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESENCLAST;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#221#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VAESENCLAST;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#221#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VAESIMC;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#249#1#219#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VAESKEYGENASSIST;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#223#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#85#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#85#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#85#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#85#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDNPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#85#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#84#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#84#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#84#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#84#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VANDPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#84#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#13#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDPD;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#13#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#12#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDPS;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#12#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDVPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_xmmreg);
    code    : #241#242#250#1#75#61#80#247;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDVPD;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_ymmreg);
    code    : #241#242#244#250#1#75#61#80#247;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDVPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_xmmreg);
    code    : #241#242#250#1#74#61#80#247;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBLENDVPS;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_ymmreg);
    code    : #241#242#244#250#1#74#61#80#247;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBROADCASTF128;
    ops     : 2;
    optypes : (ot_ymmreg,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #241#242#244#249#1#26#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VBROADCASTSD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#244#249#1#25#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VBROADCASTSD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#249#1#25#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VBROADCASTSD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#25#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VBROADCASTSD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#25#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#24#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#24#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#244#249#1#24#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#24#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#233#241#249#1#24#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VBROADCASTSS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#24#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCMPEQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#0;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#0;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLTPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#1;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLTPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#1;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#2;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#2;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#3;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORDPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#3;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#4;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#4;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLTPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#5;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLTPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#5;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#6;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#6;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#7;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORDPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#7;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_UQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#8;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_UQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#8;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#9;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#9;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGTPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#10;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGTPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#10;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#11;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#11;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#12;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#12;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#13;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#13;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGTPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#14;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGTPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#14;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUEPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#15;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUEPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#15;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_OSPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#16;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_OSPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#16;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLT_OQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#17;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLT_OQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#17;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLE_OQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#18;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLE_OQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#18;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORD_SPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#19;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORD_SPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#19;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_USPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#20;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_USPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#20;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLT_UQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#21;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLT_UQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#21;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLE_UQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLE_UQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORD_SPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORD_SPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_USPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#24;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_USPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#24;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGE_UQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#25;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGE_UQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#25;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGT_UQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#26;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGT_UQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#26;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSE_OSPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#27;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSE_OSPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#27;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OSPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#28;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OSPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#28;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGE_OQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#29;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGE_OQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#29;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGT_OQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#30;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGT_OQPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#30;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUE_USPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #242#248#1#194#61#80#1#31;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUE_USPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #242#244#248#1#194#61#80#1#31;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#0;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#0;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLTPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#1;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLTPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#1;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#2;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#2;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#3;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORDPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#3;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#4;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#4;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLTPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#5;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLTPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#5;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#6;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#6;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#7;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORDPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#7;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_UQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#8;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_UQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#8;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#9;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#9;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGTPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#10;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGTPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#10;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#11;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#11;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#12;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#12;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#13;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#13;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGTPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#14;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGTPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#14;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUEPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#15;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUEPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#15;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_OSPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#16;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_OSPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#16;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLT_OQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#17;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLT_OQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#17;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLE_OQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#18;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPLE_OQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#18;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORD_SPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#19;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPUNORD_SPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#19;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_USPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#20;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_USPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#20;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLT_UQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#21;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLT_UQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#21;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLE_UQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNLE_UQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORD_SPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPORD_SPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_USPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#24;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPEQ_USPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#24;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGE_UQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#25;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGE_UQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#25;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGT_UQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#26;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNGT_UQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#26;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSE_OSPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#27;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPFALSE_OSPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#27;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OSPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#28;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPNEQ_OSPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#28;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGE_OQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#29;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGE_OQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#29;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGT_OQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#30;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPGT_OQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#30;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUE_USPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#194#61#80#1#31;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPTRUE_USPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#194#61#80#1#31;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#234#241#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#234#241#248#1#194#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPPD;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#244#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#244#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#248#1#194#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#248#1#194#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #242#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPPS;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #242#244#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPSD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #220#232#234#248#1#194#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCMPSD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #220#232#234#248#1#194#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCMPSD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8);
    code    : #220#242#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPSD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #220#242#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCMPSS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #219#232#248#1#194#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCMPSS;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #219#232#248#1#194#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCMPSS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #222#219#242#248#1#194#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCOMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#248#1#47#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCOMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg_sae,ot_none,ot_none);
    code    : #232#234#241#242#248#1#47#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCOMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#242#248#1#47#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCOMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg_sae,ot_none,ot_none);
    code    : #232#242#248#1#47#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #219#232#242#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#242#248#1#230#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#248#1#230#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #219#232#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#244#248#1#230#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #219#232#233#248#1#230#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#233#248#1#230#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#233#248#1#230#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#242#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#242#244#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#244#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#248#1#91#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #220#232#234#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #220#232#234#242#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #220#232#234#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#234#248#1#230#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#234#244#248#1#230#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_er,ot_none,ot_none);
    code    : #220#232#233#234#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#233#234#248#1#230#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#90#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#90#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#242#248#1#90#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#90#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#90#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_er,ot_none,ot_none);
    code    : #232#233#234#241#248#1#90#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#90#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#19#72;
    flags   : [if_avx,if_thvm]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#19#72;
    flags   : [if_avx]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#242#244#249#1#19#72;
    flags   : [if_avx,if_thvm]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#19#72;
    flags   : [if_avx]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#249#1#19#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VCVTPH2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_sae,ot_none,ot_none);
    code    : #232#233#241#249#1#19#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#242#244#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#241#248#1#91#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#242#248#1#90#72;
    flags   : [if_avx,if_sandybridge,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#90#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#248#1#90#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#242#244#248#1#90#72;
    flags   : [if_avx,if_sandybridge,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#244#248#1#90#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#248#1#90#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_sae,ot_none,ot_none);
    code    : #232#233#248#1#90#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#248#1#90#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_mem128_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#250#1#29#65#22;
    flags   : [if_avx,if_thvm]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_mem256_m,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#29#65#22;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_mem64_m,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#29#65#22;
    flags   : [if_avx,if_thvm]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#29#65#22;
    flags   : [if_avx]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#250#1#29#65#22;
    flags   : [if_avx]
  ),
  (
    opcode  : A_VCVTPS2PH;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#29#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#242#248#1#45#72;
    flags   : [if_avx,if_sandybridge,if_t1f64]
  ),
  (
    opcode  : A_VCVTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_er,ot_none,ot_none);
    code    : #220#232#242#248#1#45#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#242#243#248#1#45#72;
    flags   : [if_avx,if_sandybridge,if_t1f64]
  ),
  (
    opcode  : A_VCVTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_er,ot_none,ot_none);
    code    : #220#232#234#242#243#248#1#45#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSD2SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#90#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSD2SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#90#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg32,ot_none);
    code    : #220#232#242#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg64,ot_none);
    code    : #220#232#234#242#243#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_x86_64]
  ),
  (
    opcode  : A_VCVTSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits32,ot_none);
    code    : #220#232#242#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#243#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg32,ot_none);
    code    : #219#232#242#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg64,ot_none);
    code    : #219#232#234#242#243#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_x86_64]
  ),
  (
    opcode  : A_VCVTSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits64,ot_none);
    code    : #219#232#234#242#243#248#1#42#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSS2SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#90#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VCVTSS2SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #219#232#242#248#1#90#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSS2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#242#248#1#45#72;
    flags   : [if_avx,if_sandybridge,if_t1f32]
  ),
  (
    opcode  : A_VCVTSS2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_er,ot_none,ot_none);
    code    : #219#232#242#248#1#45#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTSS2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#234#242#243#248#1#45#72;
    flags   : [if_avx,if_sandybridge,if_t1f32]
  ),
  (
    opcode  : A_VCVTSS2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_er,ot_none,ot_none);
    code    : #219#232#234#242#243#248#1#45#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#230#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#242#248#1#230#72;
    flags   : [if_avx,if_sandybridge,if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#230#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#230#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_sae,ot_none,ot_none);
    code    : #232#233#234#241#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#230#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#242#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#242#244#248#1#91#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#244#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #219#232#233#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#233#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2DQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #219#232#233#248#1#91#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#242#248#1#44#72;
    flags   : [if_avx,if_sandybridge,if_t1f64]
  ),
  (
    opcode  : A_VCVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_sae,ot_none,ot_none);
    code    : #220#232#242#248#1#44#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#242#243#248#1#44#72;
    flags   : [if_avx,if_sandybridge,if_t1f64]
  ),
  (
    opcode  : A_VCVTTSD2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_sae,ot_none,ot_none);
    code    : #220#232#234#242#243#248#1#44#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#242#248#1#44#72;
    flags   : [if_avx,if_sandybridge,if_t1f32]
  ),
  (
    opcode  : A_VCVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_sae,ot_none,ot_none);
    code    : #219#232#242#248#1#44#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VCVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#234#242#243#248#1#44#72;
    flags   : [if_avx,if_sandybridge,if_t1f32]
  ),
  (
    opcode  : A_VCVTTSS2SI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_sae,ot_none,ot_none);
    code    : #219#232#234#242#243#248#1#44#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#248#1#94#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#94#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VDIVPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#248#1#94#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VDIVSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VDIVSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VDIVSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VDIVSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #219#232#242#248#1#94#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VDPPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#65#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VDPPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#64#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VDPPS;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#64#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VEXTRACTF128;
    ops     : 3;
    optypes : (ot_xmmrm,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#244#250#1#25#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VEXTRACTPS;
    ops     : 3;
    optypes : (ot_memory or ot_bits32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#23#65#22;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VEXTRACTPS;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#23#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VEXTRACTPS;
    ops     : 3;
    optypes : (ot_reg64,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#23#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VHADDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#124#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHADDPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#124#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHADDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #220#242#248#1#124#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHADDPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #220#242#244#248#1#124#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHSUBPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#125#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHSUBPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#125#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHSUBPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #220#242#248#1#125#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VHSUBPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #220#242#244#248#1#125#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VINSERTF128;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#24#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VINSERTPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#33#61#80#23;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VINSERTPS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#33#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VLDDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #220#242#248#1#240#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VLDDQU;
    ops     : 2;
    optypes : (ot_ymmreg,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #220#242#244#248#1#240#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VLDMXCSR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #242#248#1#174#130;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#247#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPD;
    ops     : 3;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_ymmreg,ot_none);
    code    : #241#242#244#249#1#47#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPD;
    ops     : 3;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #241#242#249#1#47#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_memory or ot_bits256,ot_none);
    code    : #241#242#244#249#1#45#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits128,ot_none);
    code    : #241#242#249#1#45#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPS;
    ops     : 3;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_ymmreg,ot_none);
    code    : #241#242#244#249#1#46#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPS;
    ops     : 3;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #241#242#249#1#46#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_memory or ot_bits256,ot_none);
    code    : #241#242#244#249#1#44#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMASKMOVPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits128,ot_none);
    code    : #241#242#249#1#44#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_none);
    code    : #232#233#234#241#248#1#95#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#95#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMAXPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_none);
    code    : #232#233#248#1#95#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMAXSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMAXSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #220#232#234#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMAXSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMAXSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #219#232#242#248#1#95#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_none);
    code    : #232#233#234#241#248#1#93#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#93#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMINPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_none);
    code    : #232#233#248#1#93#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMINSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMINSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #220#232#234#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMINSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMINSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #219#232#242#248#1#93#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#41#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#41#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#248#1#41#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#242#248#1#40#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#40#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#248#1#40#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#41#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#242#244#248#1#41#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#248#1#41#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#242#248#1#40#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#242#244#248#1#40#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVAPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#248#1#40#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVD;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits32,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#248#1#126#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#248#1#110#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVDDUP;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#242#248#1#18#72;
    flags   : [if_avx,if_sandybridge,if_tmddup]
  ),
  (
    opcode  : A_VMOVDDUP;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #220#232#234#242#248#1#18#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDDUP;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #220#232#234#242#244#248#1#18#72;
    flags   : [if_avx,if_sandybridge,if_tmddup]
  ),
  (
    opcode  : A_VMOVDDUP;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #220#232#233#234#248#1#18#72;
    flags   : [if_avx512,if_tmddup]
  ),
  (
    opcode  : A_VMOVDQA;
    ops     : 2;
    optypes : (ot_ymmrm,ot_ymmreg,ot_none,ot_none);
    code    : #241#242#244#248#1#127#65;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQA;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#248#1#111#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQA;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#127#65;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQA;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #241#242#244#248#1#111#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQU;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #219#242#244#248#1#111#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQU;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #219#242#248#1#111#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQU;
    ops     : 2;
    optypes : (ot_xmmrm,ot_xmmreg,ot_none,ot_none);
    code    : #219#242#248#1#127#65;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVDQU;
    ops     : 2;
    optypes : (ot_ymmrm,ot_ymmreg,ot_none,ot_none);
    code    : #219#242#244#248#1#127#65;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVHLPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#242#248#1#18#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVHPD;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#23#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVHPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#242#248#1#22#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVHPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#23#65;
    flags   : [if_avx,if_sandybridge,if_t2]
  ),
  (
    opcode  : A_VMOVHPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#242#248#1#22#61#80;
    flags   : [if_avx,if_sandybridge,if_t2]
  ),
  (
    opcode  : A_VMOVLHPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#242#248#1#22#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVLPD;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#19#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVLPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#242#248#1#18#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVLPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#19#65;
    flags   : [if_avx,if_sandybridge,if_t2]
  ),
  (
    opcode  : A_VMOVLPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#242#248#1#18#61#80;
    flags   : [if_avx,if_sandybridge,if_t2]
  ),
  (
    opcode  : A_VMOVMSKPD;
    ops     : 2;
    optypes : (ot_reg32,ot_ymmreg,ot_none,ot_none);
    code    : #241#242#244#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPD;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPD;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPD;
    ops     : 2;
    optypes : (ot_reg64,ot_ymmreg,ot_none,ot_none);
    code    : #241#242#244#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPS;
    ops     : 2;
    optypes : (ot_reg32,ot_ymmreg,ot_none,ot_none);
    code    : #242#244#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPS;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg,ot_none,ot_none);
    code    : #242#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPS;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #242#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVMSKPS;
    ops     : 2;
    optypes : (ot_reg64,ot_ymmreg,ot_none,ot_none);
    code    : #242#244#248#1#80#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVNTDQ;
    ops     : 2;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#248#1#231#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTDQ;
    ops     : 2;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#242#244#248#1#231#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTDQ;
    ops     : 2;
    optypes : (ot_memory or ot_bits512,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#248#1#231#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTDQA;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#242#249#1#42#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTDQA;
    ops     : 2;
    optypes : (ot_ymmreg,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#241#242#244#249#1#42#72;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTDQA;
    ops     : 2;
    optypes : (ot_zmmreg,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#42#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPD;
    ops     : 2;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#43#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPD;
    ops     : 2;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#43#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPD;
    ops     : 2;
    optypes : (ot_memory or ot_bits512,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#248#1#43#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#43#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_none,ot_none);
    code    : #232#242#244#248#1#43#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVNTPS;
    ops     : 2;
    optypes : (ot_memory or ot_bits512,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#248#1#43#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#214#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_rm_gpr or ot_bits64,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#243#248#1#126#65;
    flags   : [if_avx,if_sandybridge,if_x86_64,if_t1s]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #219#232#234#242#248#1#126#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#243#248#1#110#72;
    flags   : [if_avx,if_sandybridge,if_x86_64,if_t1s]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#234#242#248#1#126#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVQ;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#214#65;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVSD;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #220#232#234#242#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVSD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#242#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #220#232#234#242#248#1#17#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #220#232#234#242#248#1#16#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVSHDUP;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#242#248#1#22#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVSHDUP;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#242#244#248#1#22#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVSHDUP;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #219#232#233#248#1#22#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVSLDUP;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#242#248#1#18#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVSLDUP;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#242#244#248#1#18#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVSLDUP;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #219#232#233#248#1#18#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVSS;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#242#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVSS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#242#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMOVSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #219#232#242#248#1#17#61#66;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #219#232#242#248#1#16#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#248#1#17#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#242#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#248#1#16#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#242#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#242#244#248#1#17#65;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#248#1#17#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#242#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#242#244#248#1#16#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VMOVUPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#248#1#16#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMPSADBW;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#66#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMPSADBW;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#66#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#248#1#89#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#89#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VMULPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#248#1#89#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VMULSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMULSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VMULSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VMULSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #219#232#242#248#1#89#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#86#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#86#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#86#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#86#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VORPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#86#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#249#1#28#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPABSB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#28#72;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPABSB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#28#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#249#1#30#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#30#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#30#72;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#30#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#30#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#30#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#249#1#29#72;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPABSW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#29#72;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPABSW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#29#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#107#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#107#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#107#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#107#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#107#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKSSDW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#107#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKSSWB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#99#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPACKSSWB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#99#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPACKSSWB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#99#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#43#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#43#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#43#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#43#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#43#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKUSDW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#43#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPACKUSWB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#103#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPACKUSWB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#103#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPACKUSWB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#103#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#252#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#252#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#252#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#254#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#254#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#254#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#254#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#254#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#254#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#212#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#212#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#212#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#212#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#212#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#212#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPADDSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#236#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#236#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#236#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#237#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#237#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#237#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#220#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#220#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#220#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#221#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#221#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDUSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#221#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPADDW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#253#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPADDW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#253#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPADDW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#253#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPALIGNR;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#15#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPALIGNR;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#242#244#250#1#15#61#80#23;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPALIGNR;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#15#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPAND;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#219#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPAND;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#219#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPANDN;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#223#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPANDN;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#223#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPAVGB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#224#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPAVGB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#224#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPAVGB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#224#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPAVGW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#227#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPAVGW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#227#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPAVGW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#227#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDVB;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_xmmreg);
    code    : #241#242#250#1#76#61#80#247;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPBLENDVB;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_ymmreg);
    code    : #241#242#244#250#1#76#61#80#247;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBLENDW;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#14#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPBLENDW;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#14#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCLMULQDQ;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#68#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPCLMULQDQ;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#68#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCLMULQDQ;
    ops     : 4;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#68#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#116#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#116#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#116#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQB;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#116#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPEQB;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#116#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#118#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPEQD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#118#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#41#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#41#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPEQQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#41#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPEQW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPEQW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#117#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPEQW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#117#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPESTRI;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#97#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPESTRM;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#96#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPGTB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#100#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#100#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#100#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTB;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#100#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPGTB;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#100#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#102#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#102#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPGTD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#102#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#55#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#55#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPGTQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#55#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPGTW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#101#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#101#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#101#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPGTW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#101#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPGTW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#101#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPCMPISTRI;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#99#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPCMPISTRM;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#98#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPERM2F128;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#6#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#249#1#13#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#13#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#250#1#5#72#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#5#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#249#1#13#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#13#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#244#250#1#5#72#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#5#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#13#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#13#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#5#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#5#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#12#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#12#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#4#72#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#4#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#12#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#12#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#250#1#4#72#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#4#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#12#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#12#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#4#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMILPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#4#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPEXTRB;
    ops     : 3;
    optypes : (ot_memory or ot_bits8,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#20#65#22;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPEXTRB;
    ops     : 3;
    optypes : (ot_reg16,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#20#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXTRB;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#20#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRB;
    ops     : 3;
    optypes : (ot_reg64,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#20#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRB;
    ops     : 3;
    optypes : (ot_reg8,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#20#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXTRD;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#22#65#22;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPEXTRQ;
    ops     : 3;
    optypes : (ot_rm_gpr or ot_bits64,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#243#250#1#22#65#22;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_memory or ot_bits16,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#21#65#22;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg16,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#21#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg16,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#197#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#197#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg32,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#21#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg64,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#250#1#21#65#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPEXTRW;
    ops     : 3;
    optypes : (ot_reg64,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#197#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHADDD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#2#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHADDD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#2#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPHADDSW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#3#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHADDSW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#3#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPHADDW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#1#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHADDW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#1#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPHMINPOSUW;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#249#1#65#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHSUBD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#6#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHSUBD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#6#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPHSUBSW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#7#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHSUBSW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#7#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPHSUBW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#5#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPHSUBW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#5#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPINSRB;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_reg32,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#32#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPINSRB;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits8,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#32#61#80#23;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPINSRD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_rm_gpr or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#242#250#1#34#61#80#23;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPINSRQ;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_rm_gpr or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#242#243#250#1#34#61#80#23;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPINSRW;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits16,ot_immediate or ot_bits8);
    code    : #232#241#242#248#1#196#61#80#23;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VPINSRW;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_reg32,ot_immediate or ot_bits8);
    code    : #232#241#242#248#1#196#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMADDUBSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#4#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMADDUBSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#4#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMADDUBSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#4#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMADDWD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#245#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMADDWD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#245#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMADDWD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#245#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#60#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#60#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#60#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#61#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#61#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#238#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#238#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMAXSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#238#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#222#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#222#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#222#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#63#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#63#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#62#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#62#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMAXUW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#62#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMINSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#56#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMINSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#56#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMINSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#56#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#57#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#57#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#234#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMINSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#234#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMINSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#234#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMINUB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#218#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMINUB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#218#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMINUB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#218#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#59#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#59#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#58#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMINUW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#58#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMINUW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#58#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMOVMSKB;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#215#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVMSKB;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg,ot_none,ot_none);
    code    : #241#242#248#1#215#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVMSKB;
    ops     : 2;
    optypes : (ot_reg32,ot_ymmreg,ot_none,ot_none);
    code    : #241#242#244#248#1#215#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVMSKB;
    ops     : 2;
    optypes : (ot_reg64,ot_ymmreg,ot_none,ot_none);
    code    : #241#242#244#248#1#215#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#33#72;
    flags   : [if_avx,if_sandybridge,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#33#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#244#249#1#33#72;
    flags   : [if_avx2,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#33#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#241#249#1#33#72;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSXBD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#33#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #232#241#242#249#1#34#72;
    flags   : [if_avx,if_sandybridge,if_tovm]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#34#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#244#249#1#34#72;
    flags   : [if_avx2,if_tovm]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#34#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#241#249#1#34#72;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVSXBQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#34#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSXBW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#32#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXBW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#32#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXBW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#32#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXBW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#32#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#37#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#37#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#37#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#37#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXWD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#35#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXWD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#35#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXWD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#35#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXWD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#35#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSXWQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#36#72;
    flags   : [if_avx,if_sandybridge,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSXWQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#36#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVSXWQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#244#249#1#36#72;
    flags   : [if_avx2,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSXWQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#36#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVSXWQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#36#72;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXBD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#49#72;
    flags   : [if_avx,if_sandybridge,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXBD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#49#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXBD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#244#249#1#49#72;
    flags   : [if_avx2,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXBD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#49#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVZXBD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#49#72;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #232#241#242#249#1#50#72;
    flags   : [if_avx,if_sandybridge,if_tovm]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#50#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#244#249#1#50#72;
    flags   : [if_avx2,if_tovm]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#50#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#241#249#1#50#72;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVZXBQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#50#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVZXBW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#48#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXBW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#48#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXBW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#48#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXBW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#48#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#53#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#53#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#53#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#53#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXWD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#249#1#51#72;
    flags   : [if_avx,if_sandybridge,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXWD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#51#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXWD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#242#244#249#1#51#72;
    flags   : [if_avx2,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXWD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#51#72;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#52#72;
    flags   : [if_avx,if_sandybridge,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#52#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#242#244#249#1#52#72;
    flags   : [if_avx2,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#52#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#241#249#1#52#72;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVZXWQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#52#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#249#1#40#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#40#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#249#1#40#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#40#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#40#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#40#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULHRSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#11#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMULHRSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#11#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMULHRSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#11#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMULHUW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#228#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMULHUW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#228#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMULHUW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#228#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMULHW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#229#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMULHW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#229#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMULHW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#229#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#64#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#64#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#213#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPMULLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#213#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPMULLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#213#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#244#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#244#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#244#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#244#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#244#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULUDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#244#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOR;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#235#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPOR;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#235#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSADBW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#246#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSADBW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#246#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSADBW;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#246#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#0#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#0#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#0#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#112#72#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#112#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#112#72#22;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#112#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#112#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHUFD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#112#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHUFHW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #219#232#242#248#1#112#72#22;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFHW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #219#232#242#244#248#1#112#72#22;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFHW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #219#232#233#248#1#112#72#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #220#232#242#248#1#112#72#22;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #220#232#242#244#248#1#112#72#22;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #220#232#233#248#1#112#72#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSIGNB;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#8#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSIGNB;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#8#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSIGND;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#10#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSIGND;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#10#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSIGNW;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#249#1#9#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSIGNW;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#249#1#9#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#114#60#142#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#242#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#114#60#142#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#242#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#242#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#115#60#143#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#115#60#143#22;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#115#60#143#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#115#60#143#22;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#115#60#143#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLDQ;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#115#60#143#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#243#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#248#1#115#60#142#22;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#244#248#1#243#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#244#248#1#115#60#142#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#234#241#248#1#243#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#115#60#142#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#241#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#113#60#142#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#113#60#142#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#241#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#113#60#142#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#113#60#142#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#241#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSLLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#113#60#142#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#114#60#140#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#226#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#114#60#140#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#226#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#226#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#113#60#140#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#113#60#140#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#225#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#113#60#140#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#113#60#140#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#225#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#225#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRAW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#113#60#140#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#114#60#138#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#210#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#210#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#114#60#138#22;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#138#22;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#210#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#115#60#139#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#115#60#139#22;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSRLDQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#115#60#139#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLDQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#115#60#139#22;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSRLDQ;
    ops     : 3;
    optypes : (ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#115#60#139#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#211#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#248#1#115#60#138#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#244#248#1#115#60#138#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#244#248#1#211#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#234#241#248#1#211#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#115#60#138#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#248#1#113#60#138#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#113#60#138#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#209#61#80;
    flags   : [if_avx,if_sandybridge,if_tmem128]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#242#244#248#1#113#60#138#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#113#60#138#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#244#248#1#209#61#80;
    flags   : [if_avx2,if_tmem128]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#241#248#1#209#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRLW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#113#60#138#22;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#248#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#248#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#248#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#250#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#250#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#250#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#250#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#250#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#250#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#251#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#251#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#251#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#251#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#251#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#251#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSUBSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#232#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#232#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#232#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#233#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#233#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#233#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#216#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#216#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#216#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#217#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#217#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBUSW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#217#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSUBW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#249#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPSUBW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#249#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPSUBW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#249#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTEST;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #241#242#244#249#1#23#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPTEST;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#249#1#23#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPUNPCKHBW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#104#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKHBW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#104#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKHBW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#104#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#106#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#106#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#106#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#106#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#106#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#106#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#109#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#109#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#109#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#109#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#109#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHQDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#109#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKHWD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#105#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKHWD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#105#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKHWD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#105#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLBW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#96#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLBW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#96#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLBW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#96#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#98#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#98#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#98#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#98#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#98#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#98#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#108#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#108#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#108#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#108#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#108#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLQDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#108#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPUNPCKLWD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#248#1#97#61#80;
    flags   : [if_avx,if_sandybridge,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLWD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#248#1#97#61#80;
    flags   : [if_avx2,if_tfvm]
  ),
  (
    opcode  : A_VPUNPCKLWD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#97#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPXOR;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #241#242#248#1#239#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VPXOR;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #241#242#244#248#1#239#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VRCPPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #242#248#1#83#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRCPPS;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #242#244#248#1#83#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRCPSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#242#248#1#83#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRCPSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #219#242#248#1#83#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#244#250#1#9#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#9#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#244#250#1#8#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#8#72#22;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDSD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #241#242#250#1#11#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDSD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8);
    code    : #241#242#250#1#11#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDSS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #241#242#250#1#10#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VROUNDSS;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_immediate or ot_bits8);
    code    : #241#242#250#1#10#61#80#23;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRSQRTPS;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #242#244#248#1#82#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRSQRTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #242#248#1#82#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRSQRTSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#242#248#1#82#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VRSQRTSS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #219#242#248#1#82#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#242#248#1#198#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#242#244#248#1#198#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#242#248#1#198#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#242#244#248#1#198#61#80#23;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#244#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#248#1#198#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#242#248#1#81#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#242#244#248#1#81#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#234#241#248#1#81#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#242#248#1#81#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#242#244#248#1#81#72;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#244#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#248#1#81#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSQRTPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#248#1#81#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSQRTSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#81#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VSQRTSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#81#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VSQRTSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #219#232#242#248#1#81#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VSQRTSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#81#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VSTMXCSR;
    ops     : 1;
    optypes : (ot_memory or ot_bits32,ot_none,ot_none,ot_none);
    code    : #242#248#1#174#131;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#92#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSUBPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#248#1#92#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSUBSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VSUBSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #220#232#234#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VSUBSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #219#232#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VSUBSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #219#232#242#248#1#92#61#80;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VTESTPD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#249#1#15#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VTESTPD;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #241#242#244#249#1#15#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VTESTPS;
    ops     : 2;
    optypes : (ot_ymmreg,ot_ymmrm,ot_none,ot_none);
    code    : #241#242#244#249#1#14#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VTESTPS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #241#242#249#1#14#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VUCOMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#248#1#46#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VUCOMISD;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg_sae,ot_none,ot_none);
    code    : #232#234#241#242#248#1#46#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VUCOMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#242#248#1#46#72;
    flags   : [if_avx,if_sandybridge,if_t1s]
  ),
  (
    opcode  : A_VUCOMISS;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmreg_sae,ot_none,ot_none);
    code    : #232#242#248#1#46#72;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#21#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#21#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#21#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#21#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKHPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#20#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#20#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#20#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#20#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VUNPCKLPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#248#1#87#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#244#248#1#87#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#242#248#1#87#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#242#244#248#1#87#61#80;
    flags   : [if_avx,if_sandybridge,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#244#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VXORPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#248#1#87#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VZEROALL;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #242#244#248#1#119;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_VZEROUPPER;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #242#248#1#119;
    flags   : [if_avx,if_sandybridge]
  ),
  (
    opcode  : A_ANDN;
    ops     : 3;
    optypes : (ot_reg32,ot_reg32,ot_rm_gpr or ot_bits32,ot_none);
    code    : #242#249#1#242#61#80;
    flags   : [if_bmi1,if_prot]
  ),
  (
    opcode  : A_ANDN;
    ops     : 3;
    optypes : (ot_reg64,ot_reg64,ot_rm_gpr or ot_bits64,ot_none);
    code    : #242#243#249#1#242#61#80;
    flags   : [if_bmi1,if_prot,if_x86_64]
  ),
  (
    opcode  : A_BEXTR;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_reg32,ot_none);
    code    : #242#249#1#247#62#72;
    flags   : [if_bmi1,if_prot]
  ),
  (
    opcode  : A_BEXTR;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_reg64,ot_none);
    code    : #242#243#249#1#247#62#72;
    flags   : [if_bmi1,if_prot,if_x86_64]
  ),
  (
    opcode  : A_BLSI;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #242#249#1#243#60#139;
    flags   : [if_bmi1,if_prot]
  ),
  (
    opcode  : A_BLSI;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #242#243#249#1#243#60#139;
    flags   : [if_bmi1,if_prot,if_x86_64]
  ),
  (
    opcode  : A_BLSMSK;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #242#249#1#243#60#138;
    flags   : [if_bmi1,if_prot]
  ),
  (
    opcode  : A_BLSMSK;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #242#243#249#1#243#60#138;
    flags   : [if_bmi1,if_prot,if_x86_64]
  ),
  (
    opcode  : A_BLSR;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #242#249#1#243#60#137;
    flags   : [if_bmi1,if_prot]
  ),
  (
    opcode  : A_BLSR;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #242#243#249#1#243#60#137;
    flags   : [if_bmi1,if_prot,if_x86_64]
  ),
  (
    opcode  : A_TZCNT;
    ops     : 2;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_rm_gpr,ot_none,ot_none);
    code    : #208#219#2#15#188#72;
    flags   : [if_bmi1,if_sm]
  ),
  (
    opcode  : A_BZHI;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_reg32,ot_none);
    code    : #242#249#1#245#62#72;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_BZHI;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_reg64,ot_none);
    code    : #242#243#249#1#245#62#72;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_MULX;
    ops     : 3;
    optypes : (ot_reg32,ot_reg32,ot_rm_gpr or ot_bits32,ot_none);
    code    : #220#242#249#1#246#61#80;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_MULX;
    ops     : 3;
    optypes : (ot_reg64,ot_reg64,ot_rm_gpr or ot_bits64,ot_none);
    code    : #220#242#243#249#1#246#61#80;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_PDEP;
    ops     : 3;
    optypes : (ot_reg32,ot_reg32,ot_rm_gpr or ot_bits32,ot_none);
    code    : #220#242#249#1#245#61#80;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_PDEP;
    ops     : 3;
    optypes : (ot_reg64,ot_reg64,ot_rm_gpr or ot_bits64,ot_none);
    code    : #220#242#243#249#1#245#61#80;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_PEXT;
    ops     : 3;
    optypes : (ot_reg32,ot_reg32,ot_rm_gpr or ot_bits32,ot_none);
    code    : #219#242#249#1#245#61#80;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_PEXT;
    ops     : 3;
    optypes : (ot_reg64,ot_reg64,ot_rm_gpr or ot_bits64,ot_none);
    code    : #219#242#243#249#1#245#61#80;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_RORX;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_immediate or ot_bits8,ot_none);
    code    : #220#242#250#1#240#72#22;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_RORX;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_immediate or ot_bits8,ot_none);
    code    : #220#242#243#250#1#240#72#22;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_SARX;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_reg32,ot_none);
    code    : #219#242#249#1#247#62#72;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_SARX;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_reg64,ot_none);
    code    : #219#242#243#249#1#247#62#72;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_SHLX;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_reg32,ot_none);
    code    : #241#242#249#1#247#62#72;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_SHLX;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_reg64,ot_none);
    code    : #241#242#243#249#1#247#62#72;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_SHRX;
    ops     : 3;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_reg32,ot_none);
    code    : #220#242#249#1#247#62#72;
    flags   : [if_bmi2,if_prot]
  ),
  (
    opcode  : A_SHRX;
    ops     : 3;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_reg64,ot_none);
    code    : #220#242#243#249#1#247#62#72;
    flags   : [if_bmi2,if_prot,if_x86_64]
  ),
  (
    opcode  : A_ADCX;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #241#3#15#56#246#72;
    flags   : [if_adx]
  ),
  (
    opcode  : A_ADCX;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #241#214#3#15#56#246#72;
    flags   : [if_adx,if_x86_64]
  ),
  (
    opcode  : A_ADOX;
    ops     : 2;
    optypes : (ot_reg32,ot_rm_gpr or ot_bits32,ot_none,ot_none);
    code    : #219#3#15#56#246#72;
    flags   : [if_adx]
  ),
  (
    opcode  : A_ADOX;
    ops     : 2;
    optypes : (ot_reg64,ot_rm_gpr or ot_bits64,ot_none,ot_none);
    code    : #219#214#3#15#56#246#72;
    flags   : [if_adx,if_x86_64]
  ),
  (
    opcode  : A_VBROADCASTI128;
    ops     : 2;
    optypes : (ot_ymmreg,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #241#242#244#249#1#90#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VEXTRACTI128;
    ops     : 3;
    optypes : (ot_xmmrm,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#244#250#1#57#65#22;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VINSERTI128;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#56#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBLENDD;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#2#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBLENDD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #241#242#250#1#2#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #232#241#242#249#1#120#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#120#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg8,ot_none,ot_none);
    code    : #232#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #232#241#242#244#249#1#120#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#120#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#241#244#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#244#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#241#244#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg8,ot_none,ot_none);
    code    : #232#241#244#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #232#233#241#249#1#120#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#233#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#233#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#233#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg8,ot_none,ot_none);
    code    : #232#233#241#249#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#249#1#88#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#88#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#241#242#244#249#1#88#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#88#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#244#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #232#233#241#249#1#88#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#233#241#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#88#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#249#1#89#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#249#1#89#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#234#241#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#234#241#242#244#249#1#89#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#242#244#249#1#89#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#89#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#124#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#89#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #232#241#242#249#1#121#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#249#1#121#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #232#241#242#244#249#1#121#72;
    flags   : [if_avx2,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#242#244#249#1#121#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#241#244#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#241#244#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#241#244#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #232#233#241#249#1#121#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg16,ot_none,ot_none);
    code    : #232#233#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg32,ot_none,ot_none);
    code    : #232#233#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_reg64,ot_none,ot_none);
    code    : #232#233#241#249#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPBROADCASTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPERM2I128;
    ops     : 4;
    optypes : (ot_ymmreg,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #241#242#244#250#1#70#61#80#23;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPERMD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#54#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPERMD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#243#244#250#1#1#72#22;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#1#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#1#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#1#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#22#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPERMPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#22#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#242#243#244#250#1#0#72#22;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#0#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#0#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#0#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#54#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMASKMOVD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits128,ot_none);
    code    : #241#242#249#1#140#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVD;
    ops     : 3;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #241#242#249#1#142#61#66;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_memory or ot_bits256,ot_none);
    code    : #241#242#244#249#1#140#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVD;
    ops     : 3;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_ymmreg,ot_none);
    code    : #241#242#244#249#1#142#61#66;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_memory or ot_bits128,ot_none);
    code    : #241#242#243#249#1#140#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVQ;
    ops     : 3;
    optypes : (ot_memory or ot_bits128,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #241#242#243#249#1#142#61#66;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVQ;
    ops     : 3;
    optypes : (ot_memory or ot_bits256,ot_ymmreg,ot_ymmreg,ot_none);
    code    : #241#242#243#244#249#1#142#61#66;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPMASKMOVQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymmreg,ot_memory or ot_bits256,ot_none);
    code    : #241#242#243#244#249#1#140#61#80;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#71#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#71#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#71#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#71#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSLLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#71#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#70#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#70#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#69#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#69#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#69#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#69#61#80;
    flags   : [if_avx2,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#69#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGATHERDPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem32,ot_xmmreg,ot_none);
    code    : #241#242#243#249#1#146#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERDPD;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#234#241#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERDPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_xmem32,ot_ymmreg,ot_none);
    code    : #241#242#243#244#249#1#146#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERDPD;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#234#241#244#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERDPD;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_ymem32,ot_none,ot_none);
    code    : #232#233#234#241#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERDPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem32,ot_xmmreg,ot_none);
    code    : #241#242#249#1#146#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERDPS;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#241#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERDPS;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymem32,ot_ymmreg,ot_none);
    code    : #241#242#244#249#1#146#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERDPS;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_ymem32,ot_none,ot_none);
    code    : #232#241#244#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERDPS;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_zmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#146#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem64,ot_xmmreg,ot_none);
    code    : #241#242#243#249#1#147#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERQPD;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymem64,ot_ymmreg,ot_none);
    code    : #241#242#243#244#249#1#147#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERQPD;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_ymem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPD;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_zmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem32,ot_xmmreg,ot_none);
    code    : #241#242#249#1#147#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERQPS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_ymem32,ot_xmmreg,ot_none);
    code    : #241#242#244#249#1#147#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VGATHERQPS;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#241#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPS;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_ymem32,ot_none,ot_none);
    code    : #232#241#244#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGATHERQPS;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_zmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#147#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem32,ot_xmmreg,ot_none);
    code    : #241#242#249#1#144#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERDD;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#241#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDD;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymem32,ot_ymmreg,ot_none);
    code    : #241#242#244#249#1#144#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERDD;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_ymem32,ot_none,ot_none);
    code    : #232#241#244#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDD;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_zmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem64,ot_xmmreg,ot_none);
    code    : #241#242#243#249#1#144#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERDQ;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_xmem64,ot_ymmreg,ot_none);
    code    : #241#242#243#244#249#1#144#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERDQ;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_xmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERDQ;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_ymem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#144#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem32,ot_xmmreg,ot_none);
    code    : #241#242#249#1#145#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERQD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_ymem32,ot_xmmreg,ot_none);
    code    : #241#242#244#249#1#145#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERQD;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem32,ot_none,ot_none);
    code    : #232#241#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQD;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_ymem32,ot_none,ot_none);
    code    : #232#241#244#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQD;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_zmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQQ;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmem64,ot_xmmreg,ot_none);
    code    : #241#242#243#249#1#145#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERQQ;
    ops     : 2;
    optypes : (ot_xmmreg_m,ot_xmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQQ;
    ops     : 3;
    optypes : (ot_ymmreg,ot_ymem64,ot_ymmreg,ot_none);
    code    : #241#242#243#244#249#1#145#62#72;
    flags   : [if_avx2]
  ),
  (
    opcode  : A_VPGATHERQQ;
    ops     : 2;
    optypes : (ot_ymmreg_m,ot_ymem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPGATHERQQ;
    ops     : 2;
    optypes : (ot_zmmreg_m,ot_zmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#145#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VFMADDPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_xmmreg);
    code    : #241#242#250#1#105#61#80#247;
    flags   : [if_fma4]
  ),
  (
    opcode  : A_VFMADDPD;
    ops     : 4;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_xmmrm);
    code    : #241#242#250#243#1#105#61#88#246;
    flags   : [if_fma4]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#152#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#152#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#152#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#152#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#152#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#152#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#152#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#153#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADD132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#153#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#153#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#153#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#168#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#168#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#168#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#168#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#168#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#168#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#168#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#169#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADD213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#169#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#169#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#169#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#184#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#184#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#184#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#184#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#184#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#184#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#184#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADD231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#185#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#185#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADD231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#185#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMADD231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#185#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#150#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#150#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#150#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#150#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#150#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#150#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#150#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#166#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#166#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#166#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#166#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#166#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#166#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#166#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#182#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#182#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#182#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#182#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#182#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#182#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMADDSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#182#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#154#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#154#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#154#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#154#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#154#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#154#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#154#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#155#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUB132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#155#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#155#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#155#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#170#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#170#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#170#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#170#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#170#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#170#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#170#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#171#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUB213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#171#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#171#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUB213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#171#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#186#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#186#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#186#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#186#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#186#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#186#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#186#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUB231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#187#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUB231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#187#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#187#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFMSUB231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#187#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#151#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#151#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#151#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#151#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#151#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#151#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#151#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#167#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#167#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#167#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#167#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#167#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#167#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#167#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#183#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#183#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#183#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#183#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#183#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#183#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFMSUBADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#183#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#156#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#156#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#156#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#156#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#156#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#156#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#156#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#157#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMADD132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#157#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#157#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMADD132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#157#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#172#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#172#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#172#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#172#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#172#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#172#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#172#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#173#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#173#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMADD213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#173#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#173#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#188#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#188#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#188#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#188#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#188#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#188#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#188#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMADD231SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#189#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMADD231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#189#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD231SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #241#242#249#1#189#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMADD231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#189#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#158#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#158#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#158#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#158#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#158#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#158#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#158#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#159#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB132SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#159#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#159#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB132SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#159#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#174#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#174#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#174#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#174#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#174#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#174#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB213PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#174#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#175#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB213SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#175#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#175#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB213SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#175#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#242#243#249#1#190#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#242#243#244#249#1#190#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#190#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#242#249#1#190#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#242#244#249#1#190#61#80;
    flags   : [if_fma,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#190#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#190#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFNMSUB231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#242#243#249#1#191#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB231SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #223#232#234#241#242#243#249#1#191#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_VFNMSUB231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#242#249#1#191#61#80;
    flags   : [if_fma]
  ),
  (
    opcode  : A_VFNMSUB231SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #222#232#241#242#249#1#191#61#80;
    flags   : [if_fma,if_t1s]
  ),
  (
    opcode  : A_XACQUIRE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#242;
    flags   : [if_tsx]
  ),
  (
    opcode  : A_XRELEASE;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #1#243;
    flags   : [if_tsx]
  ),
  (
    opcode  : A_XBEGIN;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #208#2#199#248#52;
    flags   : [if_tsx]
  ),
  (
    opcode  : A_XABORT;
    ops     : 1;
    optypes : (ot_immediate,ot_none,ot_none,ot_none);
    code    : #2#198#248#20;
    flags   : [if_tsx,if_sb]
  ),
  (
    opcode  : A_XEND;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#213;
    flags   : [if_tsx]
  ),
  (
    opcode  : A_XTEST;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#214;
    flags   : [if_tsx]
  ),
  (
    opcode  : A_RDRAND;
    ops     : 1;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#2#15#199#134;
    flags   : [if_rand]
  ),
  (
    opcode  : A_RDSEED;
    ops     : 1;
    optypes : (ot_reg16 or ot_bits32 or ot_bits64,ot_none,ot_none,ot_none);
    code    : #208#2#15#199#135;
    flags   : [if_rand]
  ),
  (
    opcode  : A_XGETBV;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#208;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XSETBV;
    ops     : 0;
    optypes : (ot_none,ot_none,ot_none,ot_none);
    code    : #3#15#1#209;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XSAVE;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#132;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XSAVE64;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #214#2#15#174#132;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XRSTOR;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#133;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XRSTOR64;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #214#2#15#174#133;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XSAVEOPT;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#174#134;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_XSAVEOPT64;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #214#2#15#174#134;
    flags   : [if_xsave]
  ),
  (
    opcode  : A_PREFETCHWT1;
    ops     : 1;
    optypes : (ot_memory,ot_none,ot_none,ot_none);
    code    : #2#15#13#130;
    flags   : [if_prefetchwt1]
  ),
  (
    opcode  : A_KADDB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#74#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KADDD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#74#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KADDQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#74#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KADDW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#74#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#65#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#65#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDNB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#66#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDND;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#66#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDNQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#66#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDNW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#66#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#65#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KANDW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#65#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVB;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVB;
    ops     : 2;
    optypes : (ot_kreg,ot_memory or ot_bits8,ot_none,ot_none);
    code    : #241#242#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVB;
    ops     : 2;
    optypes : (ot_kreg,ot_reg32,ot_none,ot_none);
    code    : #241#242#248#1#146#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVB;
    ops     : 2;
    optypes : (ot_memory or ot_bits8,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#145#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVB;
    ops     : 2;
    optypes : (ot_reg32,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#147#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVD;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#243#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVD;
    ops     : 2;
    optypes : (ot_kreg,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #241#242#243#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVD;
    ops     : 2;
    optypes : (ot_kreg,ot_reg32,ot_none,ot_none);
    code    : #220#242#248#1#146#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVD;
    ops     : 2;
    optypes : (ot_memory or ot_bits32,ot_kreg,ot_none,ot_none);
    code    : #241#242#243#248#1#145#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVD;
    ops     : 2;
    optypes : (ot_reg32,ot_kreg,ot_none,ot_none);
    code    : #220#242#248#1#147#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVQ;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#243#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVQ;
    ops     : 2;
    optypes : (ot_kreg,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #242#243#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVQ;
    ops     : 2;
    optypes : (ot_kreg,ot_reg64,ot_none,ot_none);
    code    : #220#242#243#248#1#146#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVQ;
    ops     : 2;
    optypes : (ot_memory or ot_bits64,ot_kreg,ot_none,ot_none);
    code    : #242#243#248#1#145#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVQ;
    ops     : 2;
    optypes : (ot_reg64,ot_kreg,ot_none,ot_none);
    code    : #220#242#243#248#1#147#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVW;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVW;
    ops     : 2;
    optypes : (ot_kreg,ot_memory or ot_bits16,ot_none,ot_none);
    code    : #242#248#1#144#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVW;
    ops     : 2;
    optypes : (ot_kreg,ot_reg32,ot_none,ot_none);
    code    : #242#248#1#146#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVW;
    ops     : 2;
    optypes : (ot_memory or ot_bits16,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#145#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KMOVW;
    ops     : 2;
    optypes : (ot_reg32,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#147#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KNOTB;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#68#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KNOTD;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#243#248#1#68#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KNOTQ;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#243#248#1#68#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KNOTW;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#68#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#69#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#69#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#69#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORTESTB;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#152#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORTESTD;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#243#248#1#152#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORTESTQ;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#243#248#1#152#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORTESTW;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#152#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KORW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#69#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTLB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#50#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTLD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#51#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTLQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#243#250#1#51#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTLW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#243#250#1#50#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTRB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#48#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTRD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#250#1#49#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTRQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#243#250#1#49#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KSHIFTRW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_immediate or ot_bits8,ot_none);
    code    : #241#242#243#250#1#48#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KTESTB;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#248#1#153#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KTESTD;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #241#242#243#248#1#153#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KTESTQ;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#243#248#1#153#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KTESTW;
    ops     : 2;
    optypes : (ot_kreg,ot_kreg,ot_none,ot_none);
    code    : #242#248#1#153#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KUNPCKBW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#75#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KUNPCKDQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#75#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KUNPCKWD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#75#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXNORB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#70#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXNORD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#70#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXNORQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#70#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXNORW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#70#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXORB;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#244#248#1#71#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXORD;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #241#242#243#244#248#1#71#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXORQ;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#243#244#248#1#71#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_KXORW;
    ops     : 3;
    optypes : (ot_kreg,ot_kreg,ot_kreg,ot_none);
    code    : #242#244#248#1#71#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGND;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VALIGNQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#3#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBLENDMPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#101#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VBROADCASTF32X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#25#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTF32X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#244#249#1#25#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTF32X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#25#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTF32X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#241#249#1#25#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTF32X4;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#244#249#1#26#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VBROADCASTF32X4;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#241#249#1#26#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VBROADCASTF32X8;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#249#1#27#72;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VBROADCASTF64X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#234#241#244#249#1#26#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTF64X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#234#241#249#1#26#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTF64X4;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#234#241#249#1#27#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#89#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#249#1#89#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#89#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#244#249#1#89#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#89#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VBROADCASTI32X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#233#241#249#1#89#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTI32X4;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#244#249#1#90#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VBROADCASTI32X4;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#241#249#1#90#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VBROADCASTI32X8;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#249#1#91#72;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VBROADCASTI64X2;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#234#241#244#249#1#90#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTI64X2;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#233#234#241#249#1#90#72;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VBROADCASTI64X4;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#234#241#249#1#91#72;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VCOMPRESSPD;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPD;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPD;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#138#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#138#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#138#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCOMPRESSPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#138#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#123#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#234#241#248#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#244#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#248#1#121#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#244#248#1#121#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#244#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_er,ot_none,ot_none);
    code    : #232#233#234#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#248#1#121#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#121#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#121#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#121#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#234#241#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#248#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#248#1#123#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#248#1#123#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#244#248#1#123#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#248#1#123#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#248#1#123#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#248#1#123#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_er,ot_none,ot_none);
    code    : #232#233#241#248#1#123#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#248#1#121#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#244#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#244#248#1#121#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#248#1#121#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#248#1#121#72;
    flags   : [if_avx512,if_bcst16,if_tfv]
  ),
  (
    opcode  : A_VCVTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #232#233#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#248#1#121#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#248#1#121#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#244#248#1#121#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#248#1#121#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#248#1#121#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#248#1#121#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_er,ot_none,ot_none);
    code    : #232#233#241#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#234#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#234#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#234#244#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#234#244#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #219#232#233#234#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#233#234#248#1#230#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #219#232#233#234#248#1#230#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#244#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#248#1#91#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#244#248#1#91#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#244#248#1#91#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_er,ot_none,ot_none);
    code    : #232#233#234#248#1#91#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTQQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#248#1#91#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTSD2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#248#1#121#72;
    flags   : [if_avx512,if_t1f64]
  ),
  (
    opcode  : A_VCVTSD2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_er,ot_none,ot_none);
    code    : #220#232#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTSD2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#248#1#121#72;
    flags   : [if_avx512,if_t1f64]
  ),
  (
    opcode  : A_VCVTSD2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_er,ot_none,ot_none);
    code    : #220#232#234#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTSS2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#248#1#121#72;
    flags   : [if_avx512,if_t1f32]
  ),
  (
    opcode  : A_VCVTSS2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_er,ot_none,ot_none);
    code    : #219#232#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTSS2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#234#248#1#121#72;
    flags   : [if_avx512,if_t1f32]
  ),
  (
    opcode  : A_VCVTSS2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_er,ot_none,ot_none);
    code    : #219#232#234#248#1#121#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#244#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#248#1#120#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#244#248#1#120#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#244#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_sae,ot_none,ot_none);
    code    : #232#233#234#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#248#1#120#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPD2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#248#1#122#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#244#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#248#1#122#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#248#1#122#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2QQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_sae,ot_none,ot_none);
    code    : #232#233#241#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#244#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#244#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#248#1#120#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTTPS2UDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #232#241#248#1#120#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#248#1#120#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#244#248#1#120#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#248#1#120#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#233#241#248#1#120#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#248#1#120#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTTPS2UQQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg_sae,ot_none,ot_none);
    code    : #232#233#241#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTSD2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#248#1#120#72;
    flags   : [if_avx512,if_t1f64]
  ),
  (
    opcode  : A_VCVTTSD2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_sae,ot_none,ot_none);
    code    : #220#232#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTSD2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #220#232#234#248#1#120#72;
    flags   : [if_avx512,if_t1f64]
  ),
  (
    opcode  : A_VCVTTSD2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_sae,ot_none,ot_none);
    code    : #220#232#234#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTSS2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#248#1#120#72;
    flags   : [if_avx512,if_t1f32]
  ),
  (
    opcode  : A_VCVTTSS2USI;
    ops     : 2;
    optypes : (ot_reg32,ot_xmmreg_sae,ot_none,ot_none);
    code    : #219#232#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTTSS2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_memory or ot_bits32,ot_none,ot_none);
    code    : #219#232#234#248#1#120#72;
    flags   : [if_avx512,if_t1f32]
  ),
  (
    opcode  : A_VCVTTSS2USI;
    ops     : 2;
    optypes : (ot_reg64,ot_xmmreg_sae,ot_none,ot_none);
    code    : #219#232#234#248#1#120#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits64,ot_none,ot_none);
    code    : #219#232#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#248#1#122#72;
    flags   : [if_avx512,if_bcst2,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#244#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#244#248#1#122#72;
    flags   : [if_avx512,if_bcst4,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #219#232#233#248#1#122#72;
    flags   : [if_avx512,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #219#232#233#248#1#122#72;
    flags   : [if_avx512,if_bcst8,if_thv]
  ),
  (
    opcode  : A_VCVTUDQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#233#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #220#232#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #220#232#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #220#232#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #220#232#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #220#232#233#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #220#232#233#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUDQ2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #220#232#233#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#234#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#234#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #219#232#233#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #219#232#233#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_er,ot_none,ot_none);
    code    : #219#232#233#234#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #220#232#234#244#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #220#232#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#234#248#1#122#72;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#234#244#248#1#122#72;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #220#232#234#244#248#1#122#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmrm_er,ot_none,ot_none);
    code    : #220#232#233#234#248#1#122#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VCVTUQQ2PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #220#232#233#234#248#1#122#72;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VCVTUSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg32,ot_none);
    code    : #220#232#248#1#123#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg64,ot_none);
    code    : #220#232#234#248#1#123#61#80;
    flags   : [if_avx512,if_x86_64]
  ),
  (
    opcode  : A_VCVTUSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits32,ot_none);
    code    : #220#232#248#1#123#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCVTUSI2SD;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits64,ot_none);
    code    : #220#232#234#248#1#123#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCVTUSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg32,ot_none);
    code    : #219#232#248#1#123#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VCVTUSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_reg64,ot_none);
    code    : #219#232#234#248#1#123#61#80;
    flags   : [if_avx512,if_x86_64]
  ),
  (
    opcode  : A_VCVTUSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits32,ot_none);
    code    : #219#232#248#1#123#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VCVTUSI2SS;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmreg_er,ot_memory or ot_bits64,ot_none);
    code    : #219#232#234#248#1#123#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VDBPSADBW;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#66#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VDBPSADBW;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#66#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VDBPSADBW;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#66#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VEXP2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#200#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VEXP2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#200#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VEXP2PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#249#1#200#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXP2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#200#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VEXP2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#200#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VEXP2PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#241#249#1#200#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#234#241#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#241#244#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#241#244#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#136#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VEXPANDPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#136#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF32X4;
    ops     : 3;
    optypes : (ot_mem128_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#25#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTF32X4;
    ops     : 3;
    optypes : (ot_mem128_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#25#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTF32X4;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#25#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF32X4;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#25#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF32X8;
    ops     : 3;
    optypes : (ot_mem256_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#27#65#22;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VEXTRACTF32X8;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#27#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF64X2;
    ops     : 3;
    optypes : (ot_mem128_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#25#65#22;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VEXTRACTF64X2;
    ops     : 3;
    optypes : (ot_mem128_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#25#65#22;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VEXTRACTF64X2;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#25#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF64X2;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#25#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTF64X4;
    ops     : 3;
    optypes : (ot_mem256_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#27#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTF64X4;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#27#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI32X4;
    ops     : 3;
    optypes : (ot_mem128_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#57#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTI32X4;
    ops     : 3;
    optypes : (ot_mem128_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#57#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTI32X4;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#57#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI32X4;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#57#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI32X8;
    ops     : 3;
    optypes : (ot_mem256_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#59#65#22;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VEXTRACTI32X8;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#59#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI64X2;
    ops     : 3;
    optypes : (ot_mem128_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#57#65#22;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VEXTRACTI64X2;
    ops     : 3;
    optypes : (ot_mem128_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#57#65#22;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VEXTRACTI64X2;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#57#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI64X2;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#57#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VEXTRACTI64X4;
    ops     : 3;
    optypes : (ot_mem256_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#59#65#22;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VEXTRACTI64X4;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#59#65#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#84#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#84#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFIXUPIMMPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#84#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFIXUPIMMSD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#85#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VFIXUPIMMSD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#85#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFIXUPIMMSS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#85#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VFIXUPIMMSS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#241#250#1#85#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#102#72#22;
    flags   : [if_avx512,if_bcst2,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#102#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#102#72#22;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#102#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSPD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#102#72#22;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits256,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#102#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#102#72#22;
    flags   : [if_avx512,if_bcst4,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#102#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#102#72#22;
    flags   : [if_avx512,if_bcst8,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#102#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSPS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#102#72#22;
    flags   : [if_avx512,if_bcst16,if_tfv]
  ),
  (
    opcode  : A_VFPCLASSSD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#103#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSSD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#103#72#22;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VFPCLASSSS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#103#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VFPCLASSSS;
    ops     : 3;
    optypes : (ot_kreg_m,ot_memory or ot_bits32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#103#72#22;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#249#1#66#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#66#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETEXPPS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#241#249#1#66#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETEXPSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#67#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGETEXPSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#234#241#249#1#67#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETEXPSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#67#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGETEXPSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#241#249#1#67#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#38#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#38#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VGETMANTPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#38#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETMANTSD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#39#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGETMANTSD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#39#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VGETMANTSS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#39#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VGETMANTSS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#241#250#1#39#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VINSERTF32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#24#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VINSERTF32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#24#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VINSERTF32X8;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#26#61#80#23;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VINSERTF64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#24#61#80#23;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VINSERTF64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#24#61#80#23;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VINSERTF64X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#26#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VINSERTI32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#56#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VINSERTI32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#56#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VINSERTI32X8;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#58#61#80#23;
    flags   : [if_avx512,if_t8]
  ),
  (
    opcode  : A_VINSERTI64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#56#61#80#23;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VINSERTI64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#56#61#80#23;
    flags   : [if_avx512,if_t2]
  ),
  (
    opcode  : A_VINSERTI64X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#58#61#80#23;
    flags   : [if_avx512,if_t4]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA32;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQA64;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #220#232#234#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #220#232#234#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #220#232#234#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #220#232#234#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #220#232#233#234#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU16;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #220#232#233#234#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #219#232#233#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU32;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #219#232#234#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#234#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #219#232#234#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#234#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #219#232#233#234#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU64;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#234#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #220#232#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_xmmrm_mz,ot_xmmreg,ot_none,ot_none);
    code    : #220#232#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #220#232#244#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_ymmrm_mz,ot_ymmreg,ot_none,ot_none);
    code    : #220#232#244#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #220#232#233#248#1#111#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VMOVDQU8;
    ops     : 2;
    optypes : (ot_zmmrm_mz,ot_zmmreg,ot_none,ot_none);
    code    : #220#232#233#248#1#127#65;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPABSQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#31#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDND;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDNQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#223#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPANDQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#219#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDMB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDMB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#100#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPBLENDMW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDMW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPBLENDMW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#102#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#31#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPUB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPUB;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUD;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUQ;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#30#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCMPUW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPUW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPUW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#62#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCMPW;
    ops     : 4;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#63#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#139#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#139#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCONFLICTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#196#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2B;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2B;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2B;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2D;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#119#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2Q;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#118#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMI2W;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2W;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMI2W;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#117#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2B;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2B;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2B;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2D;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2PS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#127#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2Q;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#126#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPERMT2W;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2W;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMT2W;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#125#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPERMW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#141#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#241#244#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#234#241#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#241#244#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#137#72;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPEXPANDQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#137#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPLZCNTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#68#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52HUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#181#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMADD52LUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#180#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXSQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#61#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMAXUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#63#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINSQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#57#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMINUQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#59#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMOVB2M;
    ops     : 2;
    optypes : (ot_kreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVB2M;
    ops     : 2;
    optypes : (ot_kreg,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVB2M;
    ops     : 2;
    optypes : (ot_kreg,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVD2M;
    ops     : 2;
    optypes : (ot_kreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVD2M;
    ops     : 2;
    optypes : (ot_kreg,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVD2M;
    ops     : 2;
    optypes : (ot_kreg,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#49#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#49#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#49#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#49#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#49#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#49#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#51#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#51#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#51#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#51#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#51#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVDW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#51#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2B;
    ops     : 2;
    optypes : (ot_xmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2B;
    ops     : 2;
    optypes : (ot_ymmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#244#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2B;
    ops     : 2;
    optypes : (ot_zmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#233#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2D;
    ops     : 2;
    optypes : (ot_xmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2D;
    ops     : 2;
    optypes : (ot_ymmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#244#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2D;
    ops     : 2;
    optypes : (ot_zmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#233#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2Q;
    ops     : 2;
    optypes : (ot_xmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#234#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2Q;
    ops     : 2;
    optypes : (ot_ymmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#234#244#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2Q;
    ops     : 2;
    optypes : (ot_zmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#233#234#249#1#56#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2W;
    ops     : 2;
    optypes : (ot_xmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#234#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2W;
    ops     : 2;
    optypes : (ot_ymmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#234#244#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVM2W;
    ops     : 2;
    optypes : (ot_zmmreg,ot_kreg,ot_none,ot_none);
    code    : #219#232#233#234#249#1#40#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQ2M;
    ops     : 2;
    optypes : (ot_kreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#234#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQ2M;
    ops     : 2;
    optypes : (ot_kreg,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#234#244#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQ2M;
    ops     : 2;
    optypes : (ot_kreg,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#234#249#1#57#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_mem16_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#50#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#50#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#50#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#50#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#50#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#50#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#53#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#53#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#53#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#53#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#53#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#53#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#52#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#52#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#52#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#52#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#52#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#52#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#33#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#33#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#33#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#33#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#33#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#33#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#35#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#35#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#35#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#35#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#35#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSDW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#35#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_mem16_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#34#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#34#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#34#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#34#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#34#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#34#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#37#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#37#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#37#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#37#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#37#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#37#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#36#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#36#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#36#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#36#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#36#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#36#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#32#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#32#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#32#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#32#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#32#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVSWB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#32#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#17#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#17#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#17#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#17#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#17#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#17#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#19#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#19#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#19#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#19#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#19#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSDW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#19#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_mem16_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#18#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_mem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#18#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#18#65;
    flags   : [if_avx512,if_tovm]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#18#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#18#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#18#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#21#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#21#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#21#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#21#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#21#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#21#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#20#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_mem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#20#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_mem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#20#65;
    flags   : [if_avx512,if_tqvm]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#20#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#20#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSQW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#20#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#16#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#16#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#16#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#16#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#16#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVUSWB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#16#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVW2M;
    ops     : 2;
    optypes : (ot_kreg,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#234#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVW2M;
    ops     : 2;
    optypes : (ot_kreg,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#234#244#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVW2M;
    ops     : 2;
    optypes : (ot_kreg,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#234#249#1#41#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#48#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_mem256_m,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#48#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_mem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#48#65;
    flags   : [if_avx512,if_thvm]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #219#232#249#1#48#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #219#232#244#249#1#48#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMOVWB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #219#232#233#249#1#48#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#64#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPMULTISHIFTQB;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#131#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#235#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#137#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPROLVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#21#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#136#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPRORVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#20#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSCATTERDD;
    ops     : 2;
    optypes : (ot_xmem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERDD;
    ops     : 2;
    optypes : (ot_ymem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERDD;
    ops     : 2;
    optypes : (ot_zmem32_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERDQ;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERDQ;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERDQ;
    ops     : 2;
    optypes : (ot_ymem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#160#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQD;
    ops     : 2;
    optypes : (ot_xmem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQD;
    ops     : 2;
    optypes : (ot_ymem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQD;
    ops     : 2;
    optypes : (ot_zmem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQQ;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQQ;
    ops     : 2;
    optypes : (ot_ymem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSCATTERQQ;
    ops     : 2;
    optypes : (ot_zmem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#161#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VPSLLVW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#18#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLVW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#18#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSLLVW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#18#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#248#1#226#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#244#248#1#226#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_xmmrm,ot_none);
    code    : #232#233#234#241#248#1#226#61#80;
    flags   : [if_avx512,if_tmem128]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#248#1#114#60#140#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#70#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSRAVW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#17#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAVW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#17#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRAVW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#17#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLVW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#16#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLVW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#16#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSRLVW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#16#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTERNLOGQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#37#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #219#232#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #219#232#244#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #219#232#233#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #219#232#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem32,ot_none);
    code    : #219#232#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #219#232#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem32,ot_none);
    code    : #219#232#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #219#232#233#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMD;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem32,ot_none);
    code    : #219#232#233#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #219#232#234#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_bmem64,ot_none);
    code    : #219#232#234#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #219#232#234#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_bmem64,ot_none);
    code    : #219#232#234#244#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #219#232#233#234#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMQ;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_bmem64,ot_none);
    code    : #219#232#233#234#249#1#39#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPTESTNMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #219#232#234#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #219#232#234#244#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPTESTNMW;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #219#232#233#234#249#1#38#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPXORQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#248#1#239#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#80#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#80#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRANGEPS;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#80#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRANGESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#81#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRANGESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#81#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRANGESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#81#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRANGESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#241#250#1#81#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#76#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP14SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#77#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRCP14SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#234#241#249#1#77#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP14SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#77#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRCP14SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#241#249#1#77#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#202#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#202#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#249#1#202#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#202#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#202#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRCP28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#241#249#1#202#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP28SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#203#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRCP28SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#234#241#249#1#203#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRCP28SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#203#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRCP28SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#241#249#1#203#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#86#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#86#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VREDUCEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#86#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VREDUCESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#87#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VREDUCESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#87#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VREDUCESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#87#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VREDUCESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#241#250#1#87#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#234#241#244#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#9#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#234#241#250#1#9#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#241#244#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#8#72#22;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRNDSCALEPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_immediate or ot_bits8,ot_none);
    code    : #232#233#241#250#1#8#72#22;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRNDSCALESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#11#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRNDSCALESD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#11#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRNDSCALESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#10#61#80#23;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRNDSCALESS;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_immediate or ot_bits8);
    code    : #232#241#250#1#10#61#80#23;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#234#241#244#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#241#244#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#78#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT14SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#234#241#249#1#79#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT14SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#79#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRSQRT14SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg,ot_none);
    code    : #232#241#249#1#79#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT14SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#79#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRSQRT28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#204#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem64,ot_none,ot_none);
    code    : #232#233#234#241#249#1#204#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT28PD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#234#241#249#1#204#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#204#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_bmem32,ot_none,ot_none);
    code    : #232#233#241#249#1#204#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VRSQRT28PS;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg_sae,ot_none,ot_none);
    code    : #232#233#241#249#1#204#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT28SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#205#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRSQRT28SD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#234#241#249#1#205#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VRSQRT28SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#205#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VRSQRT28SS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_sae,ot_none);
    code    : #232#241#249#1#205#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#234#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#234#241#249#1#44#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_memory or ot_bits512,ot_none);
    code    : #232#233#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#44#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSCALEFPS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmreg_er,ot_none);
    code    : #232#233#241#249#1#44#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSCALEFSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits64,ot_none);
    code    : #232#234#241#249#1#45#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCALEFSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#234#241#249#1#45#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSCALEFSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_memory or ot_bits32,ot_none);
    code    : #232#241#249#1#45#61#80;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCALEFSS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmreg_er,ot_none);
    code    : #232#241#249#1#45#61#80;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VSCATTERDPD;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERDPD;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERDPD;
    ops     : 2;
    optypes : (ot_ymem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERDPS;
    ops     : 2;
    optypes : (ot_xmem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERDPS;
    ops     : 2;
    optypes : (ot_ymem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERDPS;
    ops     : 2;
    optypes : (ot_zmem32_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#162#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPD;
    ops     : 2;
    optypes : (ot_xmem64_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPD;
    ops     : 2;
    optypes : (ot_ymem64_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPD;
    ops     : 2;
    optypes : (ot_zmem64_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPS;
    ops     : 2;
    optypes : (ot_xmem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPS;
    ops     : 2;
    optypes : (ot_ymem32_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSCATTERQPS;
    ops     : 2;
    optypes : (ot_zmem32_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#163#65;
    flags   : [if_avx512,if_t1s]
  ),
  (
    opcode  : A_VSHUFF32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFF64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#35#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI32X4;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI32X4;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI64X2;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VSHUFI64X2;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#67#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#99#65;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#99#65;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#99#65;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_mem128_m,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#99#65;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_mem256_m,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#99#65;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_mem512_m,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#99#65;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPCOMPRESSW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#99#65;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#80#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPBUSDS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#81#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#82#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPDPWSSDS;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#83#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#241#249#1#98#72;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#241#244#249#1#98#72;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#241#249#1#98#72;
    flags   : [if_avx512,if_t1s8]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#241#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#241#244#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#241#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_memory or ot_bits128,ot_none,ot_none);
    code    : #232#234#241#249#1#98#72;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_memory or ot_bits256,ot_none,ot_none);
    code    : #232#234#241#244#249#1#98#72;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_memory or ot_bits512,ot_none,ot_none);
    code    : #232#233#234#241#249#1#98#72;
    flags   : [if_avx512,if_t1s16]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_none,ot_none);
    code    : #232#234#241#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_none,ot_none);
    code    : #232#234#241#244#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPEXPANDW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_none,ot_none);
    code    : #232#233#234#241#249#1#98#72;
    flags   : [if_avx512]
  ),
  (
    opcode  : A_VPOPCNTB;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPOPCNTB;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPOPCNTB;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPOPCNTD;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#241#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTD;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#241#244#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTD;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#241#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTQ;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTQ;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTQ;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#85#72;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPOPCNTW;
    ops     : 2;
    optypes : (ot_xmmreg_mz,ot_xmmrm,ot_none,ot_none);
    code    : #232#234#241#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPOPCNTW;
    ops     : 2;
    optypes : (ot_ymmreg_mz,ot_ymmrm,ot_none,ot_none);
    code    : #232#234#241#244#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPOPCNTW;
    ops     : 2;
    optypes : (ot_zmmreg_mz,ot_zmmrm,ot_none,ot_none);
    code    : #232#233#234#241#249#1#84#72;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#113#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#113#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHLDVW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#112#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDVW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#112#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDVW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#112#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDW;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#112#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDW;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#112#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHLDW;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#112#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#241#244#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDD;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_immediate or ot_bits8);
    code    : #232#233#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDQ;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#115#61#80#23;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem32,ot_none);
    code    : #232#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem32,ot_none);
    code    : #232#241#244#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVD;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem32,ot_none);
    code    : #232#233#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_bmem64,ot_none);
    code    : #232#234#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_bmem64,ot_none);
    code    : #232#234#241#244#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVQ;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_bmem64,ot_none);
    code    : #232#233#234#241#249#1#115#61#80;
    flags   : [if_avx512,if_tfv]
  ),
  (
    opcode  : A_VPSHRDVW;
    ops     : 3;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#234#241#249#1#114#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDVW;
    ops     : 3;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#234#241#244#249#1#114#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDVW;
    ops     : 3;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#234#241#249#1#114#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDW;
    ops     : 4;
    optypes : (ot_xmmreg_mz,ot_xmmreg,ot_xmmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#250#1#114#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDW;
    ops     : 4;
    optypes : (ot_ymmreg_mz,ot_ymmreg,ot_ymmrm,ot_immediate or ot_bits8);
    code    : #232#234#241#244#250#1#114#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHRDW;
    ops     : 4;
    optypes : (ot_zmmreg_mz,ot_zmmreg,ot_zmmrm,ot_immediate or ot_bits8);
    code    : #232#233#234#241#250#1#114#61#80#23;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFBITQMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_xmmreg,ot_xmmrm,ot_none);
    code    : #232#241#249#1#143#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFBITQMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_ymmreg,ot_ymmrm,ot_none);
    code    : #232#241#244#249#1#143#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_VPSHUFBITQMB;
    ops     : 3;
    optypes : (ot_kreg_m,ot_zmmreg,ot_zmmrm,ot_none);
    code    : #232#233#241#249#1#143#61#80;
    flags   : [if_avx512,if_tfvm]
  ),
  (
    opcode  : A_SHA1RNDS4;
    ops     : 3;
    optypes : (ot_xmmreg,ot_xmmrm,ot_immediate,ot_none);
    code    : #3#15#58#204#72#22;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA1NEXTE;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#200#72;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA1MSG1;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#201#72;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA1MSG2;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#12#72;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA256RNDS2;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#203#72;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA256MSG1;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#204#72;
    flags   : [if_sha]
  ),
  (
    opcode  : A_SHA256MSG2;
    ops     : 2;
    optypes : (ot_xmmreg,ot_xmmrm,ot_none,ot_none);
    code    : #3#15#56#205#72;
    flags   : [if_sha]
  )
);


     var
       InsTabCache : PInsTabCache;
       InsTabMemRefSizeInfoCache: PInsTabMemRefSizeInfoCache;
     const
       { Intel style operands ! }
         opsize_2_type:array[0..2,topsize] of int64=(
         (OT_NONE,
          OT_BITS8,OT_BITS16,OT_BITS32,OT_BITS64,OT_BITS16,OT_BITS32,OT_BITS32,OT_BITS64,OT_BITS64,OT_BITS64,
          OT_BITS16,OT_BITS32,OT_BITS64,
          OT_BITS32,OT_BITS64,OT_BITS80,OT_BITS64,OT_NONE,
          OT_BITS64,
          OT_NEAR,OT_FAR,OT_SHORT,
          OT_NONE,
          OT_BITS128,
          OT_BITS256,
          OT_BITS512
         ),
         (OT_NONE,
          OT_BITS8,OT_BITS16,OT_BITS32,OT_BITS64,OT_BITS8,OT_BITS8,OT_BITS16,OT_BITS8,OT_BITS16,OT_BITS32,
          OT_BITS16,OT_BITS32,OT_BITS64,
          OT_BITS32,OT_BITS64,OT_BITS80,OT_BITS64,OT_NONE,
          OT_BITS64,
          OT_NEAR,OT_FAR,OT_SHORT,
          OT_NONE,
          OT_BITS128,
          OT_BITS256,
          OT_BITS512
         ),
         (OT_NONE,
          OT_BITS8,OT_BITS16,OT_BITS32,OT_BITS64,OT_NONE,OT_NONE,OT_NONE,OT_NONE,OT_NONE,OT_NONE,
          OT_BITS16,OT_BITS32,OT_BITS64,
          OT_BITS32,OT_BITS64,OT_BITS80,OT_BITS64,OT_NONE,
          OT_BITS64,
          OT_NEAR,OT_FAR,OT_SHORT,
          OT_NONE,
          OT_BITS128,
          OT_BITS256,
          OT_BITS512
         )
       );

      reg_ot_table : array[tregisterindex] of longint = (
OT_NONE,
OT_REG_AL,
OT_REG8,
OT_REG_AX,
OT_REG_EAX,
OT_REG_RAX,
OT_REG_CL,
OT_REG8,
OT_REG_CX,
OT_REG_ECX,
OT_REG_RCX,
OT_REG8,
OT_REG8,
OT_REG_DX,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG64,
OT_REG8,
OT_REG16,
OT_REG32,
OT_REG_DESS,
OT_REG_CS,
OT_REG_DESS,
OT_REG_DESS,
OT_REG_FSGS,
OT_REG_FSGS,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_NONE,
OT_REG_DREG,
OT_REG_DREG,
OT_REG_DREG,
OT_REG_DREG,
OT_REG_DREG,
OT_REG_DREG,
OT_REG_CREG,
OT_REG_CREG,
OT_REG_CREG,
OT_REG_CR4,
OT_REG_TREG,
OT_REG_TREG,
OT_REG_TREG,
OT_REG_TREG,
OT_REG_TREG,
OT_FPU0,
OT_FPUREG,
OT_FPUREG,
OT_FPUREG,
OT_FPUREG,
OT_FPUREG,
OT_FPUREG,
OT_FPUREG,
OT_FPU0,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_MMXREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_XMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_YMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_ZMMREG,
OT_KREG,
OT_KREG,
OT_KREG,
OT_KREG,
OT_KREG,
OT_KREG,
OT_KREG,
OT_KREG
      );

    function MemRefInfo(aAsmop: TAsmOp): TInsTabMemRefSizeInfoRec;
    begin
      result := InsTabMemRefSizeInfoCache^[aAsmop];
    end;


    function MightHaveExtension(AsmOp : TAsmOp): Boolean;
      var
        i,j: LongInt;
        insentry: pinsentry;
      begin
        Result:=true;
        i:=InsTabCache^[AsmOp];
        if i>=0 then
          begin
            insentry:=@instab[i];
            while insentry^.opcode=AsmOp do
              begin
                for j:=0 to insentry^.ops-1 do
                  begin
                    if (insentry^.optypes[j] and OT_VECTOR_EXT)<>0 then
                      exit;
                  end;
                inc(i);
                insentry:=@instab[i];
              end;
          end;
        Result:=false;
      end;

    { Operation type for spilling code }
    type
      toperation_type_table=array[tasmop,0..Max_Operands] of topertype;
    var
      operation_type_table : ^toperation_type_table;


{****************************************************************************
                              TAI_ALIGN
 ****************************************************************************}

    function tai_align.calculatefillbuf(var buf : tfillbuffer;executable : boolean):pchar;
      const
        { Updated according to
          Software Optimization Guide for AMD Family 15h Processors, Verison 3.08, January 2014
          and
          Intel 64 and IA-32 Architectures Software Developer’s Manual
            Volume 2B: Instruction Set Reference, N-Z, January 2015
        }
        alignarray_cmovcpus:array[0..10] of string[11]=(
          #$66#$66#$66#$0F#$1F#$84#$00#$00#$00#$00#$00,
          #$66#$66#$0F#$1F#$84#$00#$00#$00#$00#$00,
          #$66#$0F#$1F#$84#$00#$00#$00#$00#$00,
          #$0F#$1F#$84#$00#$00#$00#$00#$00,
          #$0F#$1F#$80#$00#$00#$00#$00,
          #$66#$0F#$1F#$44#$00#$00,
          #$0F#$1F#$44#$00#$00,
          #$0F#$1F#$40#$00,
          #$0F#$1F#$00,
          #$66#$90,
          #$90);
        alignarray:array[0..5] of string[8]=(
          #$8D#$B4#$26#$00#$00#$00#$00,
          #$8D#$B6#$00#$00#$00#$00,
          #$8D#$74#$26#$00,
          #$8D#$76#$00,
          #$89#$F6,
          #$90);
      var
        bufptr : pchar;
        j : longint;
        localsize: byte;
      begin
        inherited calculatefillbuf(buf,executable);
        if not(use_op) and executable then
         begin
           bufptr:=pchar(@buf);
           { fillsize may still be used afterwards, so don't modify }
           { e.g. writebytes(hp.calculatefillbuf(buf)^,hp.fillsize) }
           localsize:=fillsize;
           while (localsize>0) do
            begin
                  for j:=low(alignarray_cmovcpus) to high(alignarray_cmovcpus) do
                   if (localsize>=length(alignarray_cmovcpus[j])) then
                    break;
                  move(alignarray_cmovcpus[j][1],bufptr^,length(alignarray_cmovcpus[j]));
                  inc(bufptr,length(alignarray_cmovcpus[j]));
                  dec(localsize,length(alignarray_cmovcpus[j]));
            end;
         end;
        calculatefillbuf:=pchar(@buf);
      end;


{*****************************************************************************
                                 Taicpu Constructors
*****************************************************************************}

    procedure taicpu.changeopsize(siz:topsize); {$ifdef USEINLINE}inline;{$endif USEINLINE}
      begin
        opsize:=siz;
      end;


    procedure taicpu.init(_size : topsize);
      begin
         { default order is att }
         FOperandOrder:=op_att;
         segprefix:=NR_NO;
         opsize:=_size;
         insentry:=nil;
         LastInsOffset:=-1;
         InsOffset:=0;
         InsSize:=0;
         EVEXTupleState := etsUnknown;
      end;


    constructor taicpu.op_none(op : tasmop);
      begin
         inherited create(op);
         init(S_NO);
      end;


    constructor taicpu.op_none(op : tasmop;_size : topsize);
      begin
         inherited create(op);
         init(_size);
      end;


    constructor taicpu.op_reg(op : tasmop;_size : topsize;_op1 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=1;
         loadreg(0,_op1);
      end;


    constructor taicpu.op_const(op : tasmop;_size : topsize;_op1 : aint);
      begin
         inherited create(op);
         init(_size);
         ops:=1;
         loadconst(0,_op1);
      end;


    constructor taicpu.op_ref(op : tasmop;_size : topsize;const _op1 : treference);
      begin
         inherited create(op);
         init(_size);
         ops:=1;
         loadref(0,_op1);
      end;


    constructor taicpu.op_reg_reg(op : tasmop;_size : topsize;_op1,_op2 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadreg(0,_op1);
         loadreg(1,_op2);
      end;


    constructor taicpu.op_reg_const(op:tasmop; _size: topsize; _op1: tregister; _op2: aint);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadreg(0,_op1);
         loadconst(1,_op2);
      end;


    constructor taicpu.op_reg_ref(op : tasmop;_size : topsize;_op1 : tregister;const _op2 : treference);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadreg(0,_op1);
         loadref(1,_op2);
      end;


    constructor taicpu.op_const_reg(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadconst(0,_op1);
         loadreg(1,_op2);
      end;


    constructor taicpu.op_const_const(op : tasmop;_size : topsize;_op1,_op2 : aint);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadconst(0,_op1);
         loadconst(1,_op2);
      end;


    constructor taicpu.op_const_ref(op : tasmop;_size : topsize;_op1 : aint;const _op2 : treference);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadconst(0,_op1);
         loadref(1,_op2);
      end;


    constructor taicpu.op_ref_reg(op : tasmop;_size : topsize;const _op1 : treference;_op2 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadref(0,_op1);
         loadreg(1,_op2);
      end;


    constructor taicpu.op_reg_reg_reg(op : tasmop;_size : topsize;_op1,_op2,_op3 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=3;
         loadreg(0,_op1);
         loadreg(1,_op2);
         loadreg(2,_op3);
      end;


    constructor taicpu.op_const_reg_reg(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister;_op3 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=3;
         loadconst(0,_op1);
         loadreg(1,_op2);
         loadreg(2,_op3);
      end;


    constructor taicpu.op_ref_reg_reg(op : tasmop;_size : topsize;const _op1 : treference;_op2,_op3 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=3;
         loadref(0,_op1);
         loadreg(1,_op2);
         loadreg(2,_op3);
      end;


    constructor taicpu.op_const_ref_reg(op : tasmop;_size : topsize;_op1 : aint;const _op2 : treference;_op3 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=3;
         loadconst(0,_op1);
         loadref(1,_op2);
         loadreg(2,_op3);
      end;


    constructor taicpu.op_const_reg_ref(op : tasmop;_size : topsize;_op1 : aint;_op2 : tregister;const _op3 : treference);
      begin
         inherited create(op);
         init(_size);
         ops:=3;
         loadconst(0,_op1);
         loadreg(1,_op2);
         loadref(2,_op3);
      end;


    constructor taicpu.op_reg_reg_ref(op : tasmop;_size : topsize;_op1,_op2 : tregister;const _op3 : treference);
      begin
        inherited create(op);
        init(_size);
        ops:=3;
        loadreg(0,_op1);
        loadreg(1,_op2);
        loadref(2,_op3);
      end;

    constructor taicpu.op_const_reg_reg_reg(op : tasmop; _size : topsize; _op1 : aint; _op2, _op3, _op4 : tregister);
      begin
        inherited create(op);
        init(_size);
        ops:=4;
        loadconst(0,_op1);
        loadreg(1,_op2);
        loadreg(2,_op3);
        loadreg(3,_op4);
      end;


    constructor taicpu.op_cond_sym(op : tasmop;cond:TAsmCond;_size : topsize;_op1 : tasmsymbol);
      begin
         inherited create(op);
         init(_size);
         condition:=cond;
         ops:=1;
         loadsymbol(0,_op1,0);
      end;


    constructor taicpu.op_sym(op : tasmop;_size : topsize;_op1 : tasmsymbol);
      begin
         inherited create(op);
         init(_size);
         ops:=1;
         loadsymbol(0,_op1,0);
      end;

    constructor taicpu.op_sym_ofs(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint);
      begin
         inherited create(op);
         init(_size);
         ops:=1;
         loadsymbol(0,_op1,_op1ofs);
      end;

    constructor taicpu.op_sym_ofs_reg(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint;_op2 : tregister);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadsymbol(0,_op1,_op1ofs);
         loadreg(1,_op2);
      end;

    constructor taicpu.op_sym_ofs_ref(op : tasmop;_size : topsize;_op1 : tasmsymbol;_op1ofs:longint;const _op2 : treference);
      begin
         inherited create(op);
         init(_size);
         ops:=2;
         loadsymbol(0,_op1,_op1ofs);
         loadref(1,_op2);
      end;

    procedure taicpu.Swapoperands;
      var
        p : POper;
      begin
        { Fix the operands which are in AT&T style and we need them in Intel style }
        case ops of
          0,1:
            ;
          2 : begin
                { 0,1 -> 1,0 }
                p:=oper[0];
                oper[0]:=oper[1];
                oper[1]:=p;
              end;
          3 : begin
                { 0,1,2 -> 2,1,0 }
                p:=oper[0];
                oper[0]:=oper[2];
                oper[2]:=p;
              end;
          4 : begin
                { 0,1,2,3 -> 3,2,1,0 }
                p:=oper[0];
                oper[0]:=oper[3];
                oper[3]:=p;
                p:=oper[1];
                oper[1]:=oper[2];
                oper[2]:=p;
              end;
          else
            internalerror(201108141);
        end;
      end;


    procedure taicpu.SetOperandOrder(order:TOperandOrder);
      begin
        if FOperandOrder<>order then
         begin
           Swapoperands;
           FOperandOrder:=order;
         end;
      end;


    function taicpu.FixNonCommutativeOpcodes: tasmop;
      begin
        result:=opcode;

        { we need ATT order }
        SetOperandOrder(op_att);

        if (
            (ops=2) and
            (oper[0]^.typ=top_reg) and
            (oper[1]^.typ=top_reg) and
           { if the first is ST and the second is also a register
             it is necessarily ST1 .. ST7 }
            ((oper[0]^.reg=NR_ST) or
             (oper[0]^.reg=NR_ST0))
           ) or
           { ((ops=1) and
            (oper[0]^.typ=top_reg) and
            (oper[0]^.reg in [R_ST1..R_ST7]))  or}
           (ops=0) then
          begin
            if opcode=A_FSUBR then
              result:=A_FSUB
            else if opcode=A_FSUB then
              result:=A_FSUBR
            else if opcode=A_FDIVR then
              result:=A_FDIV
            else if opcode=A_FDIV then
              result:=A_FDIVR
            else if opcode=A_FSUBRP then
              result:=A_FSUBP
            else if opcode=A_FSUBP then
              result:=A_FSUBRP
            else if opcode=A_FDIVRP then
              result:=A_FDIVP
            else if opcode=A_FDIVP then
              result:=A_FDIVRP;
          end;
        if (
            (ops=1) and
            (oper[0]^.typ=top_reg) and
            (getregtype(oper[0]^.reg)=R_FPUREGISTER) and
            (oper[0]^.reg<>NR_ST)
           ) then
         begin
           if opcode=A_FSUBRP then
             result:=A_FSUBP
           else if opcode=A_FSUBP then
             result:=A_FSUBRP
           else if opcode=A_FDIVRP then
             result:=A_FDIVP
           else if opcode=A_FDIVP then
             result:=A_FDIVRP;
         end;
      end;


{*****************************************************************************
                                Assembler
*****************************************************************************}

    type
      ea = packed record
        sib_present : boolean;
        bytes : byte;
        size  : byte;
        modrm : byte;
        sib   : byte;
        rex   : byte;
      end;

    procedure taicpu.create_ot(objdata:TObjData);
      {
        this function will also fix some other fields which only needs to be once
      }
      var
        i,l,relsize : longint;
        currsym : TObjSymbol;
      begin
        if ops=0 then
         exit;
        { update oper[].ot field }
        for i:=0 to ops-1 do
         with oper[i]^ do
          begin
            case typ of
              top_reg :
                begin
                  ot:=reg_ot_table[findreg_by_number(reg)];
                end;
              top_ref :
                begin
                  if (ref^.refaddr in [addr_no{$ifdef x86_64},addr_tpoff{$endif x86_64}{$ifdef i386},addr_ntpoff{$endif i386}])
                     or (
                         (ref^.refaddr in [addr_pic,addr_pic_no_got,addr_tlsgd]) and
                         (ref^.base<>NR_NO)
                        )
                     then
                    begin

                      { create ot field }
                      if (reg_ot_table[findreg_by_number(ref^.base)] and OT_REG_GPR = OT_REG_GPR) and
                         ((reg_ot_table[findreg_by_number(ref^.index)] = OT_XMMREG) or
                          (reg_ot_table[findreg_by_number(ref^.index)] = OT_YMMREG) or
                          (reg_ot_table[findreg_by_number(ref^.index)] = OT_ZMMREG)
                         ) then
                        // AVX2 - vector-memory-referenz (e.g. vgatherdpd xmm0, [rax  xmm1], xmm2)
                        ot := (reg_ot_table[findreg_by_number(ref^.base)] and OT_REG_GPR) or
                              (reg_ot_table[findreg_by_number(ref^.index)])
                      else if (ref^.base = NR_NO) and
                              ((reg_ot_table[findreg_by_number(ref^.index)] = OT_XMMREG) or
                               (reg_ot_table[findreg_by_number(ref^.index)] = OT_YMMREG) or
                               (reg_ot_table[findreg_by_number(ref^.index)] = OT_ZMMREG)
                              ) then
                        // AVX2 - vector-memory-referenz without base-register (e.g. vgatherdpd xmm0, [xmm1], xmm2)
                        ot := (OT_REG_GPR) or
                              (reg_ot_table[findreg_by_number(ref^.index)])

                      else if (ot and OT_SIZE_MASK)=0 then
                        ot:=OT_MEMORY_ANY or opsize_2_type[i,opsize]
                      else
                        ot:=OT_MEMORY_ANY or (ot and OT_SIZE_MASK);
                      if (ref^.base=NR_NO) and (ref^.index=NR_NO) then
                        ot:=ot or OT_MEM_OFFS;
                      { fix scalefactor }
                      if (ref^.index=NR_NO) then
                       ref^.scalefactor:=0
                      else
                       if (ref^.scalefactor=0) then
                        ref^.scalefactor:=1;
                    end
                  else
                    begin
                      { Jumps use a relative offset which can be 8bit,
                        for other opcodes we always need to generate the full
                        32bit address }
                      if assigned(objdata) and
                         is_jmp then
                        begin
                          currsym:=objdata.symbolref(ref^.symbol);
                          l:=ref^.offset;
{$push}
{$r-,q-} { disable also overflow as address returns a qword for x86_64 }
                          if assigned(currsym) then
                            inc(l,currsym.address);
{$pop}
                          { when it is a forward jump we need to compensate the
                            offset of the instruction since the previous time,
                            because the symbol address is then still using the
                            'old-style' addressing.
                            For backwards jumps this is not required because the
                            address of the symbol is already adjusted to the
                            new offset }
                          if (l>InsOffset) and (LastInsOffset<>-1) then
                            inc(l,InsOffset-LastInsOffset);
                          { instruction size will then always become 2 (PFV) }
                          relsize:=(InsOffset+2)-l;
                          if (relsize>=-128) and (relsize<=127) and
                             (
                              not assigned(currsym) or
                              (currsym.objsection=objdata.currobjsec)
                             ) then
                            ot:=OT_IMM8 or OT_SHORT
                          else
                            ot:=OT_IMM32 or OT_NEAR;
                        end
                      else
                        ot:=OT_IMM32 or OT_NEAR;
                    end;
                end;
              top_local :
                begin
                  if (ot and OT_SIZE_MASK)=0 then
                    ot:=OT_MEMORY or opsize_2_type[i,opsize]
                  else
                    ot:=OT_MEMORY or (ot and OT_SIZE_MASK);
                end;
              top_const :
                begin
                  // if opcode is a SSE or AVX-instruction then we need a
                  // special handling (opsize can different from const-size)
                  // (e.g. "pextrw  reg/m16, xmmreg, imm8" =>> opsize (16 bit), const-size (8 bit)
                  if (InsTabMemRefSizeInfoCache^[opcode].ExistsSSEAVX) and
                     (not(InsTabMemRefSizeInfoCache^[opcode].ConstSize in [csiMultiple, csiUnknown])) then
                  begin
                    case InsTabMemRefSizeInfoCache^[opcode].ConstSize of
                      csiNoSize: ot := ot and OT_NON_SIZE or OT_IMMEDIATE;
                        csiMem8: ot := ot and OT_NON_SIZE or OT_IMMEDIATE or OT_BITS8;
                       csiMem16: ot := ot and OT_NON_SIZE or OT_IMMEDIATE or OT_BITS16;
                       csiMem32: ot := ot and OT_NON_SIZE or OT_IMMEDIATE or OT_BITS32;
                       csiMem64: ot := ot and OT_NON_SIZE or OT_IMMEDIATE or OT_BITS64;
                       else
                         ;
                    end;
                  end
                  else
                  begin
                    { allow 2nd, 3rd or 4th operand being a constant and expect no size for shuf* etc. }
                    { further, allow AAD and AAM with imm. operand }
                    if (opsize=S_NO) and not((i in [1,2,3])
                      ) then
                      message(asmr_e_invalid_opcode_and_operand);
                    if
                       (opsize<>S_W) and
                       (aint(val)>=-128) and (val<=127) then
                      ot:=OT_IMM8 or OT_SIGNED
                    else
                      ot:=OT_IMMEDIATE or opsize_2_type[i,opsize];
                    if (val=1) and (i=1) then
                      ot := ot or OT_ONENESS;
                  end;
                end;
              top_none :
                begin
                  { generated when there was an error in the
                    assembler reader. It never happends when generating
                    assembler }
                end;
              else
                internalerror(200402266);
            end;
          end;
      end;


    function taicpu.InsEnd:longint;
      begin
        InsEnd:=InsOffset+InsSize;
      end;


      function taicpu.Matches(p:PInsEntry):boolean;
      { * IF_SM stands for Size Match: any operand whose size is not
       * explicitly specified by the template is `really' intended to be
       * the same size as the first size-specified operand.
       * Non-specification is tolerated in the input instruction, but
       * _wrong_ specification is not.
       *
       * IF_SM2 invokes Size Match on only the first _two_ operands, for
       * three-operand instructions such as SHLD: it implies that the
       * first two operands must match in size, but that the third is
       * required to be _unspecified_.
       *
       * IF_SB invokes Size Byte: operands with unspecified size in the
       * template are really bytes, and so no non-byte specification in
       * the input instruction will be tolerated. IF_SW similarly invokes
       * Size Word, and IF_SD invokes Size Doubleword.
       *
       * (The default state if neither IF_SM nor IF_SM2 is specified is
       * that any operand with unspecified size in the template is
       * required to have unspecified size in the instruction too...)
      }
      var
        insot,
        currot: int64;
        i,j,asize,oprs : longint;
        insflags:tinsflags;
        vopext: int64;
        siz : array[0..max_operands-1] of longint;
      begin
        result:=false;

        { Check the opcode and operands }
        if (p^.opcode<>opcode) or (p^.ops<>ops) then
          exit;

        for i:=0 to p^.ops-1 do
         begin
           insot:=p^.optypes[i];
           currot:=oper[i]^.ot;

           { Check the operand flags }
           if (insot and (not currot) and OT_NON_SIZE)<>0 then
             exit;

           // IGNORE VECTOR-MEMORY-SIZE
           if insot and OT_TYPE_MASK = OT_MEMORY then
            insot := insot and not(int64(OT_BITS128 or OT_BITS256 or OT_BITS512));


           { Check if the passed operand size matches with one of
             the supported operand sizes }
           if ((insot and OT_SIZE_MASK)<>0) and
              ((insot and currot and OT_SIZE_MASK)<>(currot and OT_SIZE_MASK)) then
             exit;
           { "far" matches only with "far" }
           if (insot and OT_FAR)<>(currot and OT_FAR) then
             exit;
         end;

        { Check operand sizes }
        insflags:=p^.flags;
        if (insflags*IF_SMASK)<>[] then
          begin
            { as default an untyped size can get all the sizes, this is different
              from nasm, but else we need to do a lot checking which opcodes want
              size or not with the automatic size generation }
            asize:=-1;
            if IF_SB in insflags then
              asize:=OT_BITS8
            else if IF_SW in insflags then
              asize:=OT_BITS16
            else if IF_SD in insflags then
              asize:=OT_BITS32;
            if insflags*IF_ARMASK<>[] then
             begin
               siz[0]:=-1;
               siz[1]:=-1;
               siz[2]:=-1;
               if IF_AR0 in insflags then
                 siz[0]:=asize
               else if IF_AR1 in insflags then
                 siz[1]:=asize
               else if IF_AR2 in insflags then
                 siz[2]:=asize
               else
                 internalerror(2017092101);
             end
            else
             begin
               siz[0]:=asize;
               siz[1]:=asize;
               siz[2]:=asize;
             end;

            if insflags*[IF_SM,IF_SM2]<>[] then
             begin
               if IF_SM2 in insflags then
                oprs:=2
               else
                oprs:=p^.ops;
               for i:=0 to oprs-1 do
                if ((p^.optypes[i] and OT_SIZE_MASK) <> 0) then
                 begin
                   for j:=0 to oprs-1 do
                    siz[j]:=p^.optypes[i] and OT_SIZE_MASK;
                   break;
                 end;
              end
             else
              oprs:=2;

            { Check operand sizes }
            for i:=0 to p^.ops-1 do
              begin
                insot:=p^.optypes[i];
                currot:=oper[i]^.ot;
                if ((insot and OT_SIZE_MASK)=0) and
                   ((currot and OT_SIZE_MASK and (not siz[i]))<>0) and
                   { Immediates can always include smaller size }
                   ((currot and OT_IMMEDIATE)=0) and
                    (((insot and OT_SIZE_MASK) or siz[i])<(currot and OT_SIZE_MASK)) then
                  exit;
                if (insot and OT_FAR)<>(currot and OT_FAR) then
                  exit;
              end;
          end;

        if (InsTabMemRefSizeInfoCache^[opcode].MemRefSize in MemRefMultiples) and
           (InsTabMemRefSizeInfoCache^[opcode].ExistsSSEAVX) then
        begin
          for i:=0 to p^.ops-1 do
           begin
             insot:=p^.optypes[i];
             currot:=oper[i]^.ot;

             { Check the operand flags }
             if (insot and (not currot) and OT_NON_SIZE)<>0 then
               exit;
             { Check if the passed operand size matches with one of
               the supported operand sizes }
             if ((insot and OT_SIZE_MASK)<>0) and
                ((insot and currot and OT_SIZE_MASK)<>(currot and OT_SIZE_MASK)) then
               exit;
           end;
        end;

        if (InsTabMemRefSizeInfoCache^[opcode].ExistsSSEAVX) then
        begin
          for i:=0 to p^.ops-1 do
           begin
             // check vectoroperand-extention e.g. {k1} {z}

             vopext := 0;
             if (oper[i]^.vopext and OTVE_VECTOR_WRITEMASK) = OTVE_VECTOR_WRITEMASK then
             begin
               vopext := vopext or OT_VECTORMASK;

               if (oper[i]^.vopext and OTVE_VECTOR_ZERO) = OTVE_VECTOR_ZERO then
                vopext := vopext or OT_VECTORZERO;
             end;

             if (oper[i]^.vopext and OTVE_VECTOR_BCST) = OTVE_VECTOR_BCST then
             begin
               vopext := vopext or OT_VECTORBCST;

               if (InsTabMemRefSizeInfoCache^[opcode].BCSTTypes <> []) then
               begin
                 // any opcodes needs a special handling

                 // default broadcast calculation is
                 // bmem32
                 //         xmmreg: {1to4}
                 //         ymmreg: {1to8}
                 //         zmmreg: {1to16}

                 // bmem64
                 //         xmmreg: {1to2}
                 //         ymmreg: {1to4}
                 //         zmmreg: {1to8}

                 // in any opcodes not exists a mmregister
                 // e.g. vfpclasspd  k1, [RAX] {1to8}, 0
                 // =>> check flags


                 case oper[i]^.vopext and (OTVE_VECTOR_BCST2 or OTVE_VECTOR_BCST4 or OTVE_VECTOR_BCST8 or OTVE_VECTOR_BCST16) of
                    OTVE_VECTOR_BCST2: if not(IF_BCST2 in p^.flags) then exit;
                    OTVE_VECTOR_BCST4: if not(IF_BCST4 in p^.flags) then exit;
                    OTVE_VECTOR_BCST8: if not(IF_BCST8 in p^.flags) then exit;
                   OTVE_VECTOR_BCST16: if not(IF_BCST16 in p^.flags) then exit;
                               else exit;
                 end;
               end;

             end;

             if (oper[i]^.vopext and OTVE_VECTOR_ER) = OTVE_VECTOR_ER then
              vopext := vopext or OT_VECTORER;

             if (oper[i]^.vopext and OTVE_VECTOR_SAE) = OTVE_VECTOR_SAE then
              vopext := vopext or OT_VECTORSAE;

             if p^.optypes[i] and vopext <> vopext then
              exit;
           end;
        end;

        result:=true;
      end;


    procedure taicpu.ResetPass1;
      begin
        { we need to reset everything here, because the choosen insentry
          can be invalid for a new situation where the previously optimized
          insentry is not correct }
        InsEntry:=nil;
        InsSize:=0;
        LastInsOffset:=-1;
      end;


    procedure taicpu.ResetPass2;
      begin
        { we are here in a second pass, check if the instruction can be optimized }
        if assigned(InsEntry) and
           (IF_PASS2 in InsEntry^.flags) then
         begin
           InsEntry:=nil;
           InsSize:=0;
         end;
        LastInsOffset:=-1;
      end;


    function taicpu.CheckIfValid:boolean; {$ifdef USEINLINE}inline;{$endif USEINLINE}
      begin
        result:=FindInsEntry(nil);
      end;


    function taicpu.FindInsentry(objdata:TObjData):boolean;
      var
        i : longint;
      begin
        result:=false;
      { Things which may only be done once, not when a second pass is done to
        optimize }

        if (Insentry=nil) or (IF_PASS2 in InsEntry^.flags) then
         begin
           current_filepos:=fileinfo;
           { We need intel style operands }
           SetOperandOrder(op_intel);
           { create the .ot fields }

           create_ot(objdata);
           { set the file postion }
         end
        else
         begin
           { we've already an insentry so it's valid }
           result:=true;
           exit;
         end;
        { Lookup opcode in the table }
        InsSize:=-1;
        i:=instabcache^[opcode];
        if i=-1 then
         begin
           Message1(asmw_e_opcode_not_in_table,gas_op2str[opcode]);
           exit;
         end;
        insentry:=@instab[i];
        while (insentry^.opcode=opcode) do
         begin
           if matches(insentry) then
             begin
               result:=true;
               exit;
             end;
           inc(insentry);
         end;
        Message1(asmw_e_invalid_opcode_and_operands,'');
        { No instruction found, set insentry to nil and inssize to -1 }
        insentry:=nil;
        inssize:=-1;
      end;

    function taicpu.CheckUseEVEX: boolean;
    var
      i: integer;
    begin
      result := false;
      for i := 0 to ops - 1 do
      begin
        if (oper[i]^.typ=top_reg) and
           (getregtype(oper[i]^.reg) = R_MMREGISTER) then
         if getsupreg(oper[i]^.reg)>=16 then
          result := true;

        if (oper[i]^.vopext and OTVE_VECTOR_MASK) <> 0 then
         result := true;
      end;
    end;

    procedure taicpu.CheckEVEXTuple(const aInput:toper; aInsEntry: pInsentry; aIsVector128, aIsVector256, aIsVector512, aIsEVEXW1: boolean);
    var
      i: integer;
      tuplesize: integer;
      memsize: integer;
    begin
      if EVEXTupleState = etsUnknown then
      begin
        EVEXTupleState := etsNotTuple;

        if aInsEntry^.Flags * IF_TUPLEMASK <> [] then
        begin
          tuplesize := 0;

          if IF_TFV in aInsEntry^.Flags then
          begin
            for i :=  0 to aInsEntry^.ops - 1 do
             if (aInsEntry^.optypes[i] and OT_BMEM32 = OT_BMEM32) then
             begin
               tuplesize := 4;
               break;
             end
             else if (aInsEntry^.optypes[i] and OT_BMEM64 = OT_BMEM64) then
             begin
               tuplesize := 8;
               break;
             end
             else if (aInsEntry^.optypes[i] and OT_MEMORY = OT_MEMORY) then
             begin
               if aIsVector512 then tuplesize := 64
                else if aIsVector256 then tuplesize := 32
                else tuplesize := 16;

               break;
             end
             else if (aInsEntry^.optypes[i] and OT_REGNORM = OT_REGMEM) then
             begin
               if aIsVector512 then tuplesize := 64
                else if aIsVector256 then tuplesize := 32
                else tuplesize := 16;

               break;
             end;


          end
          else if IF_THV in aInsEntry^.Flags then
          begin
            for i :=  0 to aInsEntry^.ops - 1 do
             if (aInsEntry^.optypes[i] and OT_BMEM32 = OT_BMEM32) then
             begin
               tuplesize := 4;
               break;
             end
             else if (aInsEntry^.optypes[i] and OT_REGNORM = OT_REGMEM) then
             begin
               if aIsVector512 then tuplesize := 32
                else if aIsVector256 then tuplesize := 16
                else tuplesize := 8;

               break;
             end
          end
          else if IF_TFVM in aInsEntry^.Flags then
          begin
            if aIsVector512 then tuplesize := 64
             else if aIsVector256 then tuplesize := 32
             else tuplesize := 16;
          end
          else
          begin
            memsize := 0;

            for i :=  0 to aInsEntry^.ops - 1 do
            begin
              if aInsEntry^.optypes[i] and (OT_REGNORM or OT_MEMORY) = OT_REGMEM then
              begin
                case aInsEntry^.optypes[i] and (OT_BITS32 or OT_BITS64) of
                  OT_BITS32: begin
                               memsize := 32;
                               break;
                             end;
                  OT_BITS64: begin
                               memsize := 64;
                               break;
                             end;
                end;
              end
              else
              case aInsEntry^.optypes[i] and (OT_MEM8 or OT_MEM16 or OT_MEM32 or OT_MEM64) of
                  OT_MEM8: begin
                             memsize := 8;
                             break;
                           end;
                 OT_MEM16: begin
                             memsize := 16;
                             break;
                           end;
                 OT_MEM32: begin
                             memsize := 32;
                             break;
                           end;
                 OT_MEM64: //if aIsEVEXW1 then
                           begin
                             memsize := 64;
                             break;
                           end;
              end;
            end;

            if IF_T1S in aInsEntry^.Flags then
            begin
              case memsize of
                 8: tuplesize := 1;
                16: tuplesize := 2;
                else if aIsEVEXW1 then tuplesize := 8
                      else tuplesize := 4;
              end;
            end
            else if IF_T1S8 in aInsEntry^.Flags then tuplesize := 1
            else if IF_T1S16 in aInsEntry^.Flags then tuplesize := 2
            else if IF_T1F32 in aInsEntry^.Flags then tuplesize := 4
            else if IF_T1F64 in aInsEntry^.Flags then tuplesize := 8
            else if IF_T2 in aInsEntry^.Flags then
            begin
              case aIsEVEXW1 of
                false: tuplesize := 8;
                  else if aIsVector256 or aIsVector512 then tuplesize := 16;
              end;
            end
            else if IF_T4 in aInsEntry^.Flags then
            begin
              case aIsEVEXW1 of
                false: if aIsVector256 or aIsVector512 then tuplesize := 16;
                  else if aIsVector512 then tuplesize := 32;
              end;
            end
            else if IF_T8 in aInsEntry^.Flags then
            begin
              case aIsEVEXW1 of
                false: if aIsVector512 then tuplesize := 32;
                else
                  Internalerror(2019081013);
              end;
            end
            else if IF_THVM in aInsEntry^.Flags then
            begin
              tuplesize := 8; // default 128bit-vectorlength
              if aIsVector256 then tuplesize := 16
               else if aIsVector512 then tuplesize := 32;
            end
            else if IF_TQVM in aInsEntry^.Flags then
            begin
              tuplesize := 4; // default 128bit-vectorlength
              if aIsVector256 then tuplesize := 8
               else if aIsVector512 then tuplesize := 16;
            end
            else if IF_TOVM in aInsEntry^.Flags then
            begin
              tuplesize := 2; // default 128bit-vectorlength
              if aIsVector256 then tuplesize := 4
               else if aIsVector512 then tuplesize := 8;
            end
            else if IF_TMEM128 in aInsEntry^.Flags then tuplesize := 16
            else if IF_TMDDUP in aInsEntry^.Flags then
            begin
              tuplesize := 8; // default 128bit-vectorlength
              if aIsVector256 then tuplesize := 32
               else if aIsVector512 then tuplesize := 64;
            end;
          end;

          if tuplesize > 0 then
          begin
            if aInput.typ = top_ref then
            begin
              if aInput.ref^.base <> NR_NO then
              begin              
                if (aInput.ref^.offset <> 0) and
                   ((aInput.ref^.offset mod tuplesize) = 0) and
                   (abs(aInput.ref^.offset) div tuplesize <= 127) then
                begin
                  aInput.ref^.offset := aInput.ref^.offset div tuplesize;
                  EVEXTupleState := etsIsTuple;
                    end;  
              end;
            end;
          end;
        end;
      end;
    end;



    function taicpu.Pass1(objdata:TObjData):longint;
      begin
        Pass1:=0;
        { Save the old offset and set the new offset }
        InsOffset:=ObjData.CurrObjSec.Size;
        { Error? }
        if (Insentry=nil) and (InsSize=-1) then
          exit;
        { set the file postion }
        current_filepos:=fileinfo;
        { Get InsEntry }
        if FindInsEntry(ObjData) then
         begin

           { Calculate instruction size }
           InsSize:=calcsize(insentry);
           if segprefix<>NR_NO then
            inc(InsSize);
           if NeedAddrPrefix then
            inc(InsSize);
           { Fix opsize if size if forced }
           if insentry^.flags*[IF_SB,IF_SW,IF_SD]<>[] then
             begin
               if insentry^.flags*IF_ARMASK=[] then
                 begin
                   if IF_SB in insentry^.flags then
                     begin
                       if opsize=S_NO then
                         opsize:=S_B;
                     end
                   else if IF_SW in insentry^.flags then
                     begin
                       if opsize=S_NO then
                         opsize:=S_W;
                     end
                   else if IF_SD in insentry^.flags then
                     begin
                       if opsize=S_NO then
                         opsize:=S_L;
                     end;
                 end;
             end;
           LastInsOffset:=InsOffset;
           Pass1:=InsSize;
           exit;
         end;
        LastInsOffset:=-1;
      end;

    const
      segprefixes: array[NR_ES..NR_GS] of Byte=(
      // es  cs   ss   ds   fs   gs
        $26, $2E, $36, $3E, $64, $65
      );

    procedure taicpu.Pass2(objdata:TObjData);
      begin
        { error in pass1 ? }
        if insentry=nil then
         exit;
        current_filepos:=fileinfo;
        { Segment override }
        if (segprefix>=NR_ES) and (segprefix<=NR_GS) then
         begin
           objdata.writebytes(segprefixes[segprefix],1);
           { fix the offset for GenNode }
           inc(InsOffset);
         end
        else if segprefix<>NR_NO then
          InternalError(201001071);
        { Address size prefix? }
        if NeedAddrPrefix then
        begin
          write0x67prefix(objdata);
          { fix the offset for GenNode }
          inc(InsOffset);
        end;
        { Generate the instruction }
        GenCode(objdata);
      end;


    function is_64_bit_ref(const ref:treference):boolean;
      begin
        result:=not is_32_bit_ref(ref);
      end;


    function is_32_bit_ref(const ref:treference):boolean;
      begin
        result:=(ref.refaddr=addr_no) and
                (ref.base<>NR_RIP) and
                (
                 ((ref.index<>NR_NO) and (getsubreg(ref.index)=R_SUBD)) or
                 ((ref.base<>NR_NO) and (getsubreg(ref.base)=R_SUBD))
                );
      end;


    function is_16_bit_ref(const ref:treference):boolean;
      var
        ir,br : Tregister;
        isub,bsub : tsubregister;
      begin
        if (ref.index<>NR_NO) and (getregtype(ref.index)=R_MMREGISTER) then
          exit(false);
        ir:=ref.index;
        br:=ref.base;
        isub:=getsubreg(ir);
        bsub:=getsubreg(br);
        { it's a direct address }
        if (br=NR_NO) and (ir=NR_NO) then
          begin
            {$ifdef i8086}
            result:=true;
            {$else i8086}
            result:=false;
            {$endif}
          end
        else
          { it's an indirection }
          begin
            result := ((ir<>NR_NO) and (isub=R_SUBW)) or
                      ((br<>NR_NO) and (bsub=R_SUBW));
          end;
      end;


    function get_ref_address_size(const ref:treference):byte;
      begin
        if is_64_bit_ref(ref) then
          result:=64
        else if is_32_bit_ref(ref) then
          result:=32
        else if is_16_bit_ref(ref) then
          result:=16
        else
          internalerror(2017101601);
      end;


    function get_default_segment_of_ref(const ref:treference):tregister;
      begin
        { for 16-bit registers, we allow base and index to be swapped, that's
        why we also we check whether ref.index=NR_BP. For 32-bit registers,
        however, index=NR_EBP is encoded differently than base=NR_EBP and has
        a different default segment. }
        if (ref.base=NR_BP) or (ref.index=NR_BP) or
           (ref.base=NR_EBP) or (ref.base=NR_ESP)
        or (ref.base=NR_RBP) or (ref.base=NR_RSP)
           then
          result:=NR_SS
        else
          result:=NR_DS;
      end;


    procedure optimize_ref(var ref:treference; inlineasm: boolean);
      var
        ss_equals_ds: boolean;
        tmpreg: TRegister;
      begin
        { x86_64 in long mode ignores all segment base, limit and access rights
          checks for the DS, ES and SS registers, so we can set ss_equals_ds to
          true (and thus, perform stronger optimizations on the reference),
          regardless of whether this is inline asm or not (so, even if the user
          is doing tricks by loading different values into DS and SS, it still
          doesn't matter while the processor is in long mode) }
        ss_equals_ds:=True;
        { remove redundant segment overrides }
        if (ref.segment<>NR_NO) and
           ((inlineasm and (ref.segment=get_default_segment_of_ref(ref))) or
            ((not inlineasm) and (segment_regs_equal(ref.segment,get_default_segment_of_ref(ref))))) then
          ref.segment:=NR_NO;
        if not is_16_bit_ref(ref) then
          begin
            { Switching index to base position gives shorter assembler instructions.
              Converting index*2 to base+index also gives shorter instructions. }
            if (ref.base=NR_NO) and (ref.index<>NR_NO) and (ref.scalefactor<=2) and
               (ss_equals_ds or (ref.segment<>NR_NO) or (ref.index<>NR_EBP))
               { do not mess with tls references, they have the (,reg,1) format on purpose
                 else the linker cannot resolve/replace them }
               {$ifdef i386} and (ref.refaddr<>addr_tlsgd) {$endif i386} then
              begin
                ref.base:=ref.index;
                if ref.scalefactor=2 then
                  ref.scalefactor:=1
                else
                  begin
                    ref.index:=NR_NO;
                    ref.scalefactor:=0;
                  end;
              end;
            { Switching rBP+reg to reg+rBP sometimes gives shorter instructions (if there's no offset)
              On x86_64 this also works for switching r13+reg to reg+r13. }
            if ((ref.base=NR_EBP) {$ifdef x86_64}or (ref.base=NR_RBP) or (ref.base=NR_R13) or (ref.base=NR_R13D){$endif}) and
               (ref.index<>NR_NO) and
               (ref.index<>NR_EBP) and {$ifdef x86_64}(ref.index<>NR_RBP) and (ref.index<>NR_R13) and (ref.index<>NR_R13D) and{$endif}
               (ref.scalefactor<=1) and (ref.offset=0) and (ref.refaddr=addr_no) and
               (ss_equals_ds or (ref.segment<>NR_NO)) then
              begin
                tmpreg:=ref.base;
                ref.base:=ref.index;
                ref.index:=tmpreg;
              end;
          end;
        { remove redundant segment overrides again }
        if (ref.segment<>NR_NO) and
           ((inlineasm and (ref.segment=get_default_segment_of_ref(ref))) or
            ((not inlineasm) and (segment_regs_equal(ref.segment,get_default_segment_of_ref(ref))))) then
          ref.segment:=NR_NO;
      end;


        function taicpu.NeedAddrPrefix(opidx: byte): boolean;
      begin
        result:=(oper[opidx]^.typ=top_ref) and is_32_bit_ref(oper[opidx]^.ref^);
      end;


    function taicpu.NeedAddrPrefix:boolean;
      var
        i: Integer;
      begin
        for i:=0 to ops-1 do
          if needaddrprefix(i) then
            exit(true);
        result:=false;
      end;


    procedure badreg(r:Tregister);
      begin
        Message1(asmw_e_invalid_register,generic_regname(r));
      end;


    function regval(r:Tregister):byte;
      const
        intsupreg2opcode: array[0..7] of byte=
        // ax cx dx bx si di bp sp   -- in x86reg.dat
        // ax cx dx bx sp bp si di   -- needed order
          (0, 1, 2, 3, 6, 7, 5, 4);
        maxsupreg: array[tregistertype] of tsuperregister=
          (0, 16, 9, 8, 32, 32, 8, 0, 0, 0);
      var
        rs: tsuperregister;
        rt: tregistertype;
      begin
        rs:=getsupreg(r);
        rt:=getregtype(r);

        if (rs>=maxsupreg[rt]) then
         badreg(r);

        result:=rs and 7;
        if (rt=R_INTREGISTER) then
          begin
            if (rs<8) then
              result:=intsupreg2opcode[rs];
            if getsubreg(r)=R_SUBH then
              inc(result,4);
          end;
      end;


    function rexbits(r: tregister): byte;
      begin
        result:=0;
        case getregtype(r) of
          R_INTREGISTER:
            if (getsupreg(r)>=RS_R8) then
          { Either B,X or R bits can be set, depending on register role in instruction.
            Set all three bits here, caller will discard unnecessary ones. }
              result:=result or $47
            else if (getsubreg(r)=R_SUBL) and
              (getsupreg(r) in [RS_RDI,RS_RSI,RS_RBP,RS_RSP]) then
              result:=result or $40
            else if (getsubreg(r)=R_SUBH) then
          { Not an actual REX bit, used to detect incompatible usage of
            AH/BH/CH/DH }
              result:=result or $80;
          R_MMREGISTER:
            //if getsupreg(r)>=RS_XMM8 then
            // AVX512 = 32 register
            //    rexbit = 0 => MMRegister 0..7  or 16..23
            //    rexbit = 1 => MMRegister 8..15 or 24..31
            if (getsupreg(r) and $08) = $08 then
              result:=result or $47;
          else
            ;
        end;
      end;



    function process_ea_ref_64_32(const input:toper;var output:ea;rfield:longint; uselargeoffset: boolean):boolean;
      var
        sym   : tasmsymbol;
        md,s  : byte;
        base,index,scalefactor,
        o     : longint;
        ir,br : Tregister;
        isub,bsub : tsubregister;
      begin
        result:=false;
        ir:=input.ref^.index;
        br:=input.ref^.base;
        isub:=getsubreg(ir);
        bsub:=getsubreg(br);
        s:=input.ref^.scalefactor;
        o:=input.ref^.offset;
        sym:=input.ref^.symbol;

        //if ((ir<>NR_NO) and (getregtype(ir)<>R_INTREGISTER)) or
        //   ((br<>NR_NO) and (br<>NR_RIP) and (getregtype(br)<>R_INTREGISTER)) then
        if ((ir<>NR_NO) and (getregtype(ir)=R_MMREGISTER) and (br<>NR_NO) and (getregtype(br)<>R_INTREGISTER)) or // vector memory (AVX2)
           ((ir<>NR_NO) and (getregtype(ir)<>R_INTREGISTER) and (getregtype(ir)<>R_MMREGISTER)) or
           ((br<>NR_NO) and (br<>NR_RIP) and (getregtype(br)<>R_INTREGISTER)) then
          internalerror(200301081);
        { it's direct address }
        if (br=NR_NO) and (ir=NR_NO) then
         begin
           output.sib_present:=true;
           output.bytes:=4;
           output.modrm:=4 or (rfield shl 3);
           output.sib:=$25;
         end
        else if (br=NR_RIP) and (ir=NR_NO) then
          begin
            { rip based }
            output.sib_present:=false;
            output.bytes:=4;
            output.modrm:=5 or (rfield shl 3);
          end
        else
        { it's an indirection }
         begin
           if ((br=NR_RIP) and (ir<>NR_NO)) or
             (ir=NR_RIP) then
             message(asmw_e_illegal_use_of_rip);
           if ir=NR_STACK_POINTER_REG then
             Message(asmw_e_illegal_use_of_sp);
           { 16 bit? }

           if ((ir<>NR_NO) and (isub in [R_SUBMMX,R_SUBMMY,R_SUBMMZ]) and
               (br<>NR_NO) and (bsub=R_SUBQ)
              ) then
           begin
             // vector memory (AVX2) =>> ignore
           end
           else if ((ir<>NR_NO) and (isub<>R_SUBQ) and (isub<>R_SUBD)) or
                   ((br<>NR_NO) and (bsub<>R_SUBQ) and (bsub<>R_SUBD)) then
           begin
             message(asmw_e_16bit_32bit_not_supported);
           end;

           { wrong, for various reasons }
           if (ir=NR_ESP) or ((s<>1) and (s<>2) and (s<>4) and (s<>8) and (ir<>NR_NO)) then
            exit;

           output.rex:=output.rex or (rexbits(br) and $F1) or (rexbits(ir) and $F2);
           result:=true;


           { base }
           case br of
             NR_R8D,
             NR_EAX,
             NR_R8,
             NR_RAX : base:=0;
             NR_R9D,
             NR_ECX,
             NR_R9,
             NR_RCX : base:=1;
             NR_R10D,
             NR_EDX,
             NR_R10,
             NR_RDX : base:=2;
             NR_R11D,
             NR_EBX,
             NR_R11,
             NR_RBX : base:=3;
             NR_R12D,
             NR_ESP,
             NR_R12,
             NR_RSP : base:=4;
             NR_R13D,
             NR_EBP,
             NR_R13,
             NR_NO,
             NR_RBP : base:=5;
             NR_R14D,
             NR_ESI,
             NR_R14,
             NR_RSI : base:=6;
             NR_R15D,
             NR_EDI,
             NR_R15,
             NR_RDI : base:=7;
           else
             exit;
           end;
           { index }
           case ir of
             NR_R8D,
             NR_EAX,
             NR_R8,
             NR_RAX,
             NR_XMM0,
             NR_XMM8,
             NR_XMM16,
             NR_XMM24,
             NR_YMM0,
             NR_YMM8,
             NR_YMM16,
             NR_YMM24,
             NR_ZMM0,
             NR_ZMM8,
             NR_ZMM16,
             NR_ZMM24: index:=0;
             NR_R9D,
             NR_ECX,
             NR_R9,
             NR_RCX,
             NR_XMM1,
             NR_XMM9,
             NR_XMM17,
             NR_XMM25,
             NR_YMM1,
             NR_YMM9,
             NR_YMM17,
             NR_YMM25,
             NR_ZMM1,
             NR_ZMM9,
             NR_ZMM17,
             NR_ZMM25: index:=1;
             NR_R10D,
             NR_EDX,
             NR_R10,
             NR_RDX,
             NR_XMM2,
             NR_XMM10,
             NR_XMM18,
             NR_XMM26,
             NR_YMM2,
             NR_YMM10,
             NR_YMM18,
             NR_YMM26,
             NR_ZMM2,
             NR_ZMM10,
             NR_ZMM18,
             NR_ZMM26: index:=2;
             NR_R11D,
             NR_EBX,
             NR_R11,
             NR_RBX,
             NR_XMM3,
             NR_XMM11,
             NR_XMM19,
             NR_XMM27,
             NR_YMM3,
             NR_YMM11,
             NR_YMM19,
             NR_YMM27,
             NR_ZMM3,
             NR_ZMM11,
             NR_ZMM19,
             NR_ZMM27: index:=3;
             NR_R12D,
             NR_ESP,
             NR_R12,
             NR_NO,
             NR_XMM4,
             NR_XMM12,
             NR_XMM20,
             NR_XMM28,
             NR_YMM4,
             NR_YMM12,
             NR_YMM20,
             NR_YMM28,
             NR_ZMM4,
             NR_ZMM12,
             NR_ZMM20,
             NR_ZMM28: index:=4;
             NR_R13D,
             NR_EBP,
             NR_R13,
             NR_RBP,
             NR_XMM5,
             NR_XMM13,
             NR_XMM21,
             NR_XMM29,
             NR_YMM5,
             NR_YMM13,
             NR_YMM21,
             NR_YMM29,
             NR_ZMM5,
             NR_ZMM13,
             NR_ZMM21,
             NR_ZMM29: index:=5;
             NR_R14D,
             NR_ESI,
             NR_R14,
             NR_RSI,
             NR_XMM6,
             NR_XMM14,
             NR_XMM22,
             NR_XMM30,
             NR_YMM6,
             NR_YMM14,
             NR_YMM22,
             NR_YMM30,
             NR_ZMM6,
             NR_ZMM14,
             NR_ZMM22,
             NR_ZMM30: index:=6;
             NR_R15D,
             NR_EDI,
             NR_R15,
             NR_RDI,
             NR_XMM7,
             NR_XMM15,
             NR_XMM23,
             NR_XMM31,
             NR_YMM7,
             NR_YMM15,
             NR_YMM23,
             NR_YMM31,
             NR_ZMM7,
             NR_ZMM15,
             NR_ZMM23,
             NR_ZMM31: index:=7;
           else
             exit;
           end;
           case s of
            0,
            1 : scalefactor:=0;
            2 : scalefactor:=1;
            4 : scalefactor:=2;
            8 : scalefactor:=3;
           else
            exit;
           end;
           { If rbp or r13 is used we must always include an offset }
           if (br=NR_NO) or
              ((br<>NR_RBP) and (br<>NR_R13) and (br<>NR_EBP) and (br<>NR_R13D) and (o=0) and (sym=nil)) then
            md:=0
           else
            if ((o>=-128) and (o<=127) and (sym=nil) and (not(uselargeoffset) or (o = 0))) then
             md:=1
            else
             md:=2;
           if (br=NR_NO) or (md=2) then
            output.bytes:=4
           else
            output.bytes:=md;
           { SIB needed ? }
           if (ir=NR_NO) and (br<>NR_RSP) and (br<>NR_R12) and (br<>NR_ESP) and (br<>NR_R12D)  then
            begin
              output.sib_present:=false;
              output.modrm:=(md shl 6) or (rfield shl 3) or base;
            end
           else
            begin
              output.sib_present:=true;
              output.modrm:=(md shl 6) or (rfield shl 3) or 4;
              output.sib:=(scalefactor shl 6) or (index shl 3) or base;
            end;
         end;
        output.size:=1+ord(output.sib_present)+output.bytes;
        result:=true;
      end;

    function process_ea(const input:toper;out output:ea;rfield:longint; uselargeoffset: boolean):boolean;
      var
        rv  : byte;
      begin
        result:=false;
        fillchar(output,sizeof(output),0);
        {Register ?}
        if (input.typ=top_reg) then
          begin
            rv:=regval(input.reg);
            output.modrm:=$c0 or (rfield shl 3) or rv;
            output.size:=1;
            output.rex:=output.rex or (rexbits(input.reg) and $F1);
            result:=true;
            exit;
          end;
        {No register, so memory reference.}
        if input.typ<>top_ref then
          internalerror(200409263);
        result:=process_ea_ref_64_32(input,output,rfield, uselargeoffset);
      end;

    function taicpu.calcsize(p:PInsEntry):shortint;
      var
        codes : pchar;
        c     : byte;
        len     : shortint;
        ea_data : ea;
        exists_evex: boolean;
        exists_vex: boolean;
        exists_vex_extension: boolean;
        exists_prefix_66: boolean;
        exists_prefix_F2: boolean;
        exists_prefix_F3: boolean;
        exists_l256: boolean;
        exists_l512: boolean;
        exists_EVEXW1: boolean;
        omit_rexw : boolean;
      begin

        len:=0;

        codes:=@p^.code[0];
        exists_vex := false;
        exists_vex_extension := false;
        exists_prefix_66 := false;
        exists_prefix_F2 := false;
        exists_prefix_F3 := false;
        exists_evex      := false;
        exists_l256      := false;
        exists_l512      := false;
        exists_EVEXW1    := false;
        rex:=0;
        omit_rexw:=false;
        repeat
          c:=ord(codes^);
          inc(codes);
          case c of
            &0 :
              break;
            &1,&2,&3 :
              begin
                inc(codes,c);
                inc(len,c);
              end;
            &10,&11,&12 :
              begin
                rex:=rex or (rexbits(oper[c-&10]^.reg) and $F1);
                inc(codes);
                inc(len);
              end;
            &13,&23 :
              begin
                inc(codes);
                inc(len);
              end;
            &4,&5,&6,&7 :
              begin
                if opsize={$ifdef i8086}S_L{$else}S_W{$endif} then
                  inc(len,2)
                else
                  inc(len);
              end;
            &14,&15,&16,
            &20,&21,&22,
            &24,&25,&26,&27,
            &50,&51,&52 :
              inc(len);
            &30,&31,&32,
            &37,
            &60,&61,&62 :
              inc(len,2);
            &34,&35,&36:
              begin
                if opsize=S_Q then
                  inc(len,8)
                else
                  inc(len,4);
              end;
            &44,&45,&46:
              inc(len,sizeof(pint));
            &54,&55,&56:
              inc(len,8);
            &40,&41,&42,
            &70,&71,&72,
            &254,&255,&256 :
              inc(len,4);
            &64,&65,&66:
              inc(len,4);
            &74,&75,&76,&77: ; // ignore vex-coded operand-idx
            &320,&321,&322 :
              begin
                case (oper[c-&320]^.ot and OT_SIZE_MASK) of
                  OT_BITS16 :
                    inc(len);
                  OT_BITS64:
                    begin
                      rex:=rex or $48;
                    end;
                end;
              end;
            &310 :
              { every insentry with code 0310 must be marked with NOX86_64 }
              InternalError(2011051301);
            &311 :
              inc(len)
              ;
            &324 :
              inc(len)
              ;
            &326 :
              begin
                rex:=rex or $48;
              end;
            &312,
            &323,
            &327,
            &331,&332: ;
            &325:
              ;

            &333:
              begin
                inc(len);
                exists_prefix_F2 := true;
              end;
            &334:
              begin
                inc(len);
                exists_prefix_F3 := true;
              end;
            &361:
              begin
                inc(len);
                exists_prefix_66 := true;
              end;
            &335:
              omit_rexw:=true
              ;
            &336,
            &337: {nothing};
            &100..&227 :
              begin
                 if (c<&177) then
                  begin
                    if (oper[c and 7]^.typ=top_reg) then
                      begin
                        rex:=rex or (rexbits(oper[c and 7]^.reg) and $F4);
                      end;
                  end;
                if (oper[(c shr 3) and 7]^.typ = top_ref) and
                   (oper[(c shr 3) and 7]^.ref^.offset <> 0) then
                begin
                  if (exists_vex and exists_evex and CheckUseEVEX) or
                     (not(exists_vex) and exists_evex) then
                  begin
                    CheckEVEXTuple(oper[(c shr 3) and 7]^, p, not(exists_l256 or exists_l512), exists_l256, exists_l512, exists_EVEXW1);
                    //const aInput:toper; aInsEntry: pInsentry; aIsVector128, aIsVector256, aIsVector512, aIsEVEXW1: boolean);
                  end;
                end;
                if process_ea(oper[(c shr 3) and 7]^, ea_data, 0, EVEXTupleState = etsNotTuple) then
                 inc(len,ea_data.size)
                  else Message(asmw_e_invalid_effective_address);
                rex:=rex or ea_data.rex;
              end;
            &350:
              begin
                exists_evex := true;
              end;
            &351: exists_l512 := true; // EVEX length bit 512
            &352: exists_EVEXW1 := true; // EVEX W1
            &362: // VEX prefix for AVX (length = 2 or 3 bytes, dependens on REX.XBW or opcode-prefix ($0F38 or $0F3A))
                  // =>> DEFAULT = 2 Bytes
              begin
                //if not(exists_vex) then
                //begin
                //  inc(len, 2);
                //end;

                exists_vex := true;
              end;
            &363: // REX.W = 1
                  // =>> VEX prefix length = 3
              begin
                if not(exists_vex_extension) then
                begin
                  //inc(len);
                  exists_vex_extension := true;
                end;
              end;
            &364: exists_l256 := true; // VEX length bit 256

            &366, // operand 2 (ymmreg) encoded immediate byte (bit 4-7)
            &367: inc(len); // operand 3 (ymmreg) encoded immediate byte (bit 4-7)
            &370: // VEX-Extension prefix $0F
                  // ignore for calculating length
                  ;
            &371, // VEX-Extension prefix $0F38
            &372: // VEX-Extension prefix $0F3A
              begin
                if not(exists_vex_extension) then
                begin
                  //inc(len);
                  exists_vex_extension := true;
                end;
              end;
            &300,&301,&302:
              begin
                if (oper[c and 3]^.ot and OT_SIZE_MASK)=OT_BITS32 then
                  inc(len);
              end;
            else
             InternalError(200603141);
          end;
        until false;
        if ((rex and $80)<>0) and ((rex and $4F)<>0) then
          Message(asmw_e_bad_reg_with_rex);
        rex:=rex and $4F;      { reset extra bits in upper nibble }
        if omit_rexw then
          begin
            if rex=$48 then    { remove rex entirely? }
              rex:=0
            else
              rex:=rex and $F7;
          end;
        if not(exists_vex or exists_evex) then
        begin
          if rex<>0 then
            Inc(len);
        end;
        if exists_evex and
           exists_vex then
        begin
          if CheckUseEVEX then
          begin
            inc(len, 4);
          end
          else
          begin
            inc(len, 2);
            if exists_vex_extension then inc(len);
              if not(exists_vex_extension) then
              if rex and $0B <> 0 then inc(len);  // REX.WXB <> 0 =>> needed VEX-Extension
          end;
          if exists_prefix_66 then dec(len);
          if exists_prefix_F2 then dec(len);
          if exists_prefix_F3 then dec(len);
        end
        else if exists_evex then
        begin
          inc(len, 4);
          if exists_prefix_66 then dec(len);
          if exists_prefix_F2 then dec(len);
          if exists_prefix_F3 then dec(len);
        end
        else
        begin
          if exists_vex then
          begin
            inc(len,2);
            if exists_prefix_66 then dec(len);
            if exists_prefix_F2 then dec(len);
            if exists_prefix_F3 then dec(len);
            if exists_vex_extension then inc(len);
            if not(exists_vex_extension) then
              if rex and $0B <> 0 then inc(len);  // REX.WXB <> 0 =>> needed VEX-Extension
          end;
        end;
        calcsize:=len;
      end;


    procedure taicpu.write0x66prefix(objdata:TObjData);
      const
        b66: Byte=$66;
      begin
        objdata.writebytes(b66,1);
      end;


    procedure taicpu.write0x67prefix(objdata:TObjData);
      const
        b67: Byte=$67;
      begin
        objdata.writebytes(b67,1);
      end;


        procedure taicpu.gencode(objdata: TObjData);
      {
       * the actual codes (C syntax, i.e. octal):
       * \0            - terminates the code. (Unless it's a literal of course.)
       * \1, \2, \3    - that many literal bytes follow in the code stream
       * \4, \6        - the POP/PUSH (respectively) codes for CS, DS, ES, SS
       *                 (POP is never used for CS) depending on operand 0
       * \5, \7        - the second byte of POP/PUSH codes for FS, GS, depending
       *                 on operand 0
       * \10, \11, \12 - a literal byte follows in the code stream, to be added
       *                 to the register value of operand 0, 1 or 2
       * \13           - a literal byte follows in the code stream, to be added
       *                 to the condition code value of the instruction.
       * \14, \15, \16 - a signed byte immediate operand, from operand 0, 1 or 2
       * \20, \21, \22 - a byte immediate operand, from operand 0, 1 or 2
       * \23           - a literal byte follows in the code stream, to be added
       *                 to the inverted condition code value of the instruction
       *                 (inverted version of \13).
       * \24, \25, \26, \27 - an unsigned byte immediate operand, from operand 0, 1, 2 or 3
       * \30, \31, \32 - a word immediate operand, from operand 0, 1 or 2
       * \34, \35, \36 - select between \3[012] and \4[012] depending on 16/32 bit
       *                 assembly mode or the address-size override on the operand
       * \37           - a word constant, from the _segment_ part of operand 0
       * \40, \41, \42 - a long immediate operand, from operand 0, 1 or 2
       * \44, \45, \46 - select between \3[012], \4[012] or \5[456] depending
                         on the address size of instruction
       * \50, \51, \52 - a byte relative operand, from operand 0, 1 or 2
       * \54, \55, \56 - a qword immediate, from operand 0, 1 or 2
       * \60, \61, \62 - a word relative operand, from operand 0, 1 or 2
       * \64, \65, \66 - select between \6[012] and \7[012] depending on 16/32 bit
       *                 assembly mode or the address-size override on the operand
       * \70, \71, \72 - a long relative operand, from operand 0, 1 or 2
       * \74, \75, \76 - a vex-coded vector operand, from operand 0, 1 or 2
       * \1ab          - a ModRM, calculated on EA in operand a, with the spare
       *                 field the register value of operand b.
       * \2ab          - a ModRM, calculated on EA in operand a, with the spare
       *                 field equal to digit b.
       * \254,\255,\256 - a signed 32-bit immediate to be extended to 64 bits
       * \300,\301,\302 - might be an 0x67, depending on the address size of
       *                 the memory reference in operand x.
       * \310          - indicates fixed 16-bit address size, i.e. optional 0x67.
       * \311          - indicates fixed 32-bit address size, i.e. optional 0x67.
       * \312          - (disassembler only) invalid with non-default address size.
       * \320,\321,\322 - might be an 0x66 or 0x48 byte, depending on the operand
       *                 size of operand x.
       * \324          - indicates fixed 16-bit operand size, i.e. optional 0x66.
       * \325          - indicates fixed 32-bit operand size, i.e. optional 0x66.
       * \326          - indicates fixed 64-bit operand size, i.e. optional 0x48.
       * \327          - indicates that this instruction is only valid when the
       *                 operand size is the default (instruction to disassembler,
       *                 generates no code in the assembler)
       * \331          - instruction not valid with REP prefix.  Hint for
       *                 disassembler only; for SSE instructions.
       * \332               - disassemble a rep (0xF3 byte) prefix as repe not rep.
       * \333          - 0xF3 prefix for SSE instructions
       * \334          - 0xF2 prefix for SSE instructions
       * \335          - Indicates 64-bit operand size with REX.W not necessary / 64-bit scalar vector operand size
       * \336          - Indicates 32-bit scalar vector operand size
       * \337          - Indicates 64-bit scalar vector operand size

       * \350          - EVEX prefix for AVX instructions
       * \351          - EVEX Vector length 512
       * \352          - EVEX W1

       * \361          - 0x66 prefix for SSE instructions

       * \362          - VEX prefix for AVX instructions
       * \363          - VEX W1
       * \364          - VEX Vector length 256

       * \366          - operand 2 (ymmreg,zmmreg) encoded in bit 4-7 of the immediate byte
       * \367          - operand 3 (ymmreg,zmmreg) encoded in bit 4-7 of the immediate byte

       * \370          - VEX 0F-FLAG
       * \371          - VEX 0F38-FLAG
       * \372          - VEX 0F3A-FLAG
      }

      var
        currval : aint;
        currsym : tobjsymbol;
        currrelreloc,
        currabsreloc,
        currabsreloc32 : TObjRelocationType;
        rexwritten : boolean;

        procedure getvalsym(opidx:longint);
          begin
            case oper[opidx]^.typ of
              top_ref :
                begin
                  currval:=oper[opidx]^.ref^.offset;
                  currsym:=ObjData.symbolref(oper[opidx]^.ref^.symbol);
                  if oper[opidx]^.ref^.refaddr=addr_pic then
                    begin
                      currrelreloc:=RELOC_PLT32;
                      currabsreloc:=RELOC_GOTPCREL;
                      currabsreloc32:=RELOC_GOTPCREL;
                    end
                  else if oper[opidx]^.ref^.refaddr=addr_pic_no_got then
                    begin
                      currrelreloc:=RELOC_RELATIVE;
                      currabsreloc:=RELOC_RELATIVE;
                      currabsreloc32:=RELOC_RELATIVE;
                    end
                  else if oper[opidx]^.ref^.refaddr=addr_tpoff then
                    begin
                      currrelreloc:=RELOC_TPOFF;
                      currabsreloc:=RELOC_TPOFF;
                      currabsreloc32:=RELOC_TPOFF;
                    end
                  else if oper[opidx]^.ref^.refaddr=addr_tlsgd then
                    begin
                      currrelreloc:=RELOC_TLSGD;
                      currabsreloc:=RELOC_TLSGD;
                      currabsreloc32:=RELOC_TLSGD;
                    end
                  else
                    begin
                      currrelreloc:=RELOC_RELATIVE;
                      currabsreloc:=RELOC_ABSOLUTE;
                      currabsreloc32:=RELOC_ABSOLUTE32;
                    end;
                end;
              top_const :
                begin
                  currval:=aint(oper[opidx]^.val);
                  currsym:=nil;
                  currabsreloc:=RELOC_ABSOLUTE;
                  currabsreloc32:=RELOC_ABSOLUTE32;
                end;
              else
                Message(asmw_e_immediate_or_reference_expected);
            end;
          end;

       procedure maybewriterex;
       begin
          if (rex<>0) and not(rexwritten) then
            begin
              rexwritten:=true;
              objdata.writebytes(rex,1);
            end;
        end;

       procedure objdata_writereloc(Data:TRelocDataInt;len:aword;p:TObjSymbol;Reloctype:TObjRelocationType);
         begin
           objdata.writereloc(data,len,p,Reloctype);
           { Computed offset is not yet correct for GOTPC relocation }
           { RELOC_GOTPCREL, RELOC_REX_GOTPCRELX, RELOC_GOTPCRELX need special handling }
           if assigned(p) and (RelocType in [RELOC_GOTPCREL, RELOC_REX_GOTPCRELX, RELOC_GOTPCRELX]) and
              { These relocations seem to be used only for ELF
                which always has relocs_use_addend set to true 
                so that it is the orgsize of the last relocation which needs to be fixed PM  }
              (insend<>objdata.CurrObjSec.size) then
             dec(TObjRelocation(objdata.CurrObjSec.ObjRelocations.Last).orgsize,insend-objdata.CurrObjSec.size);
         end;

      const
        CondVal:array[TAsmCond] of byte=($0,
         $7, $3, $2, $6, $2, $4, $F, $D, $C, $E, $6, $2,
         $3, $7, $3, $5, $E, $C, $D, $F, $1, $B, $9, $5,
         $0, $A, $A, $B, $8, $4);
      var
        i: integer;
        c : byte;
        pb : pbyte;
        codes : pchar;
        bytes : array[0..3] of byte;
        rfield,
        data,s,opidx : longint;
        ea_data : ea;
        relsym : TObjSymbol;

        needed_VEX_Extension: boolean;
        needed_VEX: boolean;
        needed_EVEX: boolean;
        opmode: integer;
        VEXvvvv: byte;
        VEXmmmmm: byte;
{
        VEXw    : byte;
        VEXpp   : byte;
        VEXll   : byte;
}
        EVEXvvvv: byte;
        EVEXpp: byte;
        EVEXr: byte;
        EVEXx: byte;
        EVEXv: byte;
        EVEXll: byte;
        EVEXw1: byte;
        EVEXz   : byte;
        EVEXaaa : byte;
        EVEXb   : byte;
        EVEXmm  : byte;

      begin
        { safety check }
        if objdata.currobjsec.size<>longword(insoffset) then
          internalerror(200130121);

        { those variables are initialized inside local procedures, the dfa cannot handle this yet }
        currsym:=nil;
        currval:=0;

        { load data to write }
        codes:=insentry^.code;
        rexwritten:=false;
        { Force word push/pop for registers }
        if (opsize={$ifdef i8086}S_L{$else}S_W{$endif}) and ((codes[0]=#4) or (codes[0]=#6) or
            ((codes[0]=#1) and ((codes[2]=#5) or (codes[2]=#7)))) then
          write0x66prefix(objdata);

        // needed VEX Prefix (for AVX etc.)

        needed_VEX    := false;
        needed_EVEX   := false;
        needed_VEX_Extension := false;
        opmode   := -1;
        VEXvvvv  := 0;
        VEXmmmmm := 0;
{
        VEXll    := 0;
        VEXw     := 0;
        VEXpp    := 0;
}
        EVEXpp   := 0;
        EVEXvvvv := 0;
        EVEXr    := 0;
        EVEXx    := 0;
        EVEXv    := 0;
        EVEXll   := 0;
        EVEXw1   := 0;
        EVEXz    := 0;
        EVEXaaa  := 0;
        EVEXb    := 0;
        EVEXmm   := 0;

        repeat
          c:=ord(codes^);
          inc(codes);

          case c of
             &0: break;
             &1,
             &2,
             &3: inc(codes,c);
            &10,
            &11,
            &12: inc(codes, 1);
            &74: opmode := 0;
            &75: opmode := 1;
            &76: opmode := 2;
     &100..&227: begin
                   // AVX 512 - EVEX
                   // check operands

                   if (c shr 6) = 1 then
                   begin
                     opidx := c and 7;
                     if ops > opidx then
                     begin
                       if (oper[opidx]^.typ=top_reg) then
                        if getsupreg(oper[opidx]^.reg) and $10 = $0 then EVEXr := 1;
                     end
                   end
                   else EVEXr := 1; // modrm:reg not used =>> 1

                   opidx := (c shr 3) and 7;
                   if ops > opidx then
                    case oper[opidx]^.typ of
                      top_reg: if getsupreg(oper[opidx]^.reg) and $10 = $0 then EVEXx := 1;
                      top_ref: begin
                                 if getsupreg(oper[opidx]^.ref^.index) and $08 = $0 then EVEXx := 1;
                                 if getsubreg(oper[opidx]^.ref^.index) in [R_SUBMMX,R_SUBMMY,R_SUBMMZ] then
                                 begin
                                   // VSIB memory addresing
                                   if getsupreg(oper[opidx]^.ref^.index) and $10 = $0 then EVEXv := 1; // VECTOR-Index
                                 end;
                               end;
                      else
                        Internalerror(2019081014);
                    end;


                 end;
           &333: begin
                   VEXvvvv              := VEXvvvv  OR $02; // set SIMD-prefix $F3
                   //VEXpp                := $02;             // set SIMD-prefix $F3
                   EVEXpp               := $02;             // set SIMD-prefix $F3
                 end;
           &334: begin
                   VEXvvvv              := VEXvvvv  OR $03; // set SIMD-prefix $F2
                   //VEXpp                := $03;             // set SIMD-prefix $F2
                   EVEXpp               := $03;             // set SIMD-prefix $F2
                 end;
           &350: needed_EVEX            := true;            // AVX512 instruction or AVX128/256/512-instruction (depended on operands [x,y,z]mm16..)
           &351: EVEXll                 := $02;             // vectorlength = 512 bits AND no scalar
           &352: EVEXw1                 := $01;
           &361: begin
                   VEXvvvv              := VEXvvvv  OR $01; // set SIMD-prefix $66
                   //VEXpp                := $01;             // set SIMD-prefix $66
                   EVEXpp               := $01;             // set SIMD-prefix $66
                 end;
           &362: needed_VEX             := true;
           &363: begin
                   needed_VEX_Extension := true;
                   VEXvvvv              := VEXvvvv  OR (1 shl 7); // set REX.W
                   //VEXw                 := 1;
                 end;
           &364: begin
                   VEXvvvv              := VEXvvvv  OR $04; // vectorlength = 256 bits AND no scalar
                   //VEXll                := $01;
                   EVEXll               := $01;
                 end;
           &366,
           &367: begin
                   opidx:=c-&364;  { 0366->operand 2, 0367->operand 3 }
                   if (ops > opidx) and
                      (oper[opidx]^.typ=top_reg) and
                      ((oper[opidx]^.ot and OT_REG_EXTRA_MASK = otf_reg_xmm) or
                       (oper[opidx]^.ot and OT_REG_EXTRA_MASK = otf_reg_ymm) or
                       (oper[opidx]^.ot and OT_REG_EXTRA_MASK = otf_reg_zmm)) then
                    if (getsupreg(oper[opidx]^.reg) and $10 = $0) then EVEXx := 1;
                 end;
           &370: begin
                   VEXmmmmm             := VEXmmmmm OR $01; // set leading opcode byte $0F
                   EVEXmm               := $01;
                 end;
           &371: begin
                   needed_VEX_Extension := true;
                   VEXmmmmm             := VEXmmmmm OR $02; // set leading opcode byte $0F38
                   EVEXmm               := $02;
                 end;
           &372: begin
                   needed_VEX_Extension := true;
                   VEXmmmmm             := VEXmmmmm OR $03; // set leading opcode byte $0F3A
                   EVEXmm               := $03;
                 end;

          end;
        until false;

        if needed_VEX or needed_EVEX then
        begin
          if (opmode > ops) or
             (opmode < -1) then
          begin
            Internalerror(777100);
          end
          else if opmode = -1 then
          begin
            VEXvvvv  := VEXvvvv or ($0F shl 3); // set VEXvvvv bits (bits 6-3) to 1
            EVEXvvvv := $0F;
          end
          else if oper[opmode]^.typ = top_reg then
          begin
            VEXvvvv  := VEXvvvv or ((not(regval(oper[opmode]^.reg)) and $07) shl 3);
            EVEXvvvv := not(regval(oper[opmode]^.reg)) and $07;

            {$ifdef x86_64}
              if rexbits(oper[opmode]^.reg) = 0 then VEXvvvv := VEXvvvv or (1 shl 6);

              if rexbits(oper[opmode]^.reg) = 0 then EVEXvvvv := EVEXvvvv or (1 shl 3);
              if getsupreg(oper[opmode]^.reg) and $10 = 0 then EVEXv := 1;
            {$else}
              VEXvvvv := VEXvvvv or (1 shl 6);
              EVEXvvvv := EVEXvvvv or (1 shl 3);
            {$endif x86_64}
          end
          else Internalerror(777101);

          if not(needed_VEX_Extension) then
          begin
              if rex and $0B <> 0 then needed_VEX_Extension := true;
          end;

          //TG
          if needed_EVEX and needed_VEX then
          begin
            needed_EVEX := false;


            if CheckUseEVEX then
            begin
              // EVEX-Flags r,v,x indicate extended-MMregister
              // Flag = 0 =>> [x,y,z]mm16..[x,y,z]mm31
              // Flag = 1 =>> [x,y,z]mm00..[x,y,z]mm15

              needed_EVEX := true;

              needed_VEX := false;
              needed_VEX_Extension := false;
            end;
          end;

          if needed_EVEX then
          begin
            EVEXaaa:= 0;
            EVEXz  := 0;

            for i := 0 to ops - 1 do
             if (oper[i]^.vopext and OTVE_VECTOR_MASK) <> 0 then
             begin
               if oper[i]^.vopext and OTVE_VECTOR_WRITEMASK = OTVE_VECTOR_WRITEMASK then
               begin
                 EVEXaaa := oper[i]^.vopext and $07;
                 if oper[i]^.vopext and OTVE_VECTOR_ZERO = OTVE_VECTOR_ZERO then EVEXz := 1;
               end;

               if oper[i]^.vopext and OTVE_VECTOR_BCST = OTVE_VECTOR_BCST then
               begin
                 EVEXb := 1;
               end;

               // flag EVEXb is multiple use (broadcast, sae and er)
               if oper[i]^.vopext and OTVE_VECTOR_SAE = OTVE_VECTOR_SAE then
               begin
                 EVEXb := 1;
               end;

               if oper[i]^.vopext and OTVE_VECTOR_ER = OTVE_VECTOR_ER then
               begin
                 EVEXb := 1;

                 case oper[i]^.vopext and OTVE_VECTOR_ER_MASK of
                   OTVE_VECTOR_RNSAE: EVEXll := 0;
                   OTVE_VECTOR_RDSAE: EVEXll := 1;
                   OTVE_VECTOR_RUSAE: EVEXll := 2;
                   OTVE_VECTOR_RZSAE: EVEXll := 3;
                                 else EVEXll := 0;
                 end;
               end;
             end;


            bytes[0] := $62;

            bytes[1] := ((EVEXmm   and $03) shl 0)  or
                      {$ifdef x86_64}
                        ((not(rex) and $05) shl 5)  or
                      {$else}
                        (($05) shl 5)               or
                      {$endif x86_64}
                        ((EVEXr    and $01) shl 4)  or
                        ((EVEXx    and $01) shl 6);

            bytes[2] := ((EVEXpp   and $03) shl 0)  or
                        ((1        and $01) shl 2)  or  // fixed in AVX512
                        ((EVEXvvvv and $0F) shl 3)  or
                        ((EVEXw1   and $01) shl 7);

            bytes[3] := ((EVEXaaa  and $07) shl 0)  or
                        ((EVEXv    and $01) shl 3)  or
                        ((EVEXb    and $01) shl 4)  or
                        ((EVEXll   and $03) shl 5)  or
                        ((EVEXz    and $01) shl 7);

            objdata.writebytes(bytes,4);
          end
          else if needed_VEX_Extension then
          begin
            // VEX-Prefix-Length = 3 Bytes
            {$ifdef x86_64}
              VEXmmmmm := VEXmmmmm or ((not(rex) and $07) shl 5);  // set REX.rxb
              VEXvvvv  := VEXvvvv or ((rex and $08) shl 7);        // set REX.w
            {$else}
              VEXmmmmm := VEXmmmmm or (7 shl 5);  //
            {$endif x86_64}

            bytes[0]:=$C4;
            bytes[1]:=VEXmmmmm;
            bytes[2]:=VEXvvvv;
            objdata.writebytes(bytes,3);
          end
          else
          begin
            // VEX-Prefix-Length = 2 Bytes
            {$ifdef x86_64}
              if rex and $04 = 0 then
            {$endif x86_64}
            begin
              VEXvvvv := VEXvvvv or (1 shl 7);
            end;

            bytes[0]:=$C5;
            bytes[1]:=VEXvvvv;
            objdata.writebytes(bytes,2);
          end;
        end
        else
        begin
          needed_VEX_Extension := false;
          opmode := -1;
        end;

        if not(needed_EVEX) then
        begin
          for opidx := 0 to ops - 1 do
          begin
            if ops > opidx then
             if (oper[opidx]^.typ=top_reg) and
                (getregtype(oper[opidx]^.reg) = R_MMREGISTER) then
              if getsupreg(oper[opidx]^.reg) and $10 = $10 then
              begin
                Message1(asmw_e_invalid_opcode_and_operands,'');
                break;
              end;
              //badreg(oper[opidx]^.reg);
          end;
        end;

        { load data to write }
        codes:=insentry^.code;
        repeat
          c:=ord(codes^);
          inc(codes);
          case c of
            &0 :
              break;
            &1,&2,&3 :
              begin
                if not(needed_VEX or needed_EVEX) then  // TG
                  maybewriterex;
                objdata.writebytes(codes^,c);
                inc(codes,c);
              end;
            &4,&6 :
              begin
                case oper[0]^.reg of
                  NR_CS:
                    bytes[0]:=$e;
                  NR_NO,
                  NR_DS:
                    bytes[0]:=$1e;
                  NR_ES:
                    bytes[0]:=$6;
                  NR_SS:
                    bytes[0]:=$16;
                  else
                    internalerror(777004);
                end;
                if c=&4 then
                  inc(bytes[0]);
                objdata.writebytes(bytes,1);
              end;
            &5,&7 :
              begin
                case oper[0]^.reg of
                  NR_FS:
                    bytes[0]:=$a0;
                  NR_GS:
                    bytes[0]:=$a8;
                  else
                    internalerror(777005);
                end;
                if c=&5 then
                  inc(bytes[0]);
                objdata.writebytes(bytes,1);
              end;
            &10,&11,&12 :
              begin
                if not(needed_VEX or needed_EVEX) then  // TG
                  maybewriterex;
                bytes[0]:=ord(codes^)+regval(oper[c-&10]^.reg);
                inc(codes);
                objdata.writebytes(bytes,1);
              end;
            &13 :
              begin
                bytes[0]:=ord(codes^)+condval[condition];
                inc(codes);
                objdata.writebytes(bytes,1);
              end;
            &14,&15,&16 :
              begin
                getvalsym(c-&14);
                if (currval<-128) or (currval>127) then
                 Message2(asmw_e_value_exceeds_bounds,'signed byte',tostr(currval));
                if assigned(currsym) then
                  objdata_writereloc(currval,1,currsym,currabsreloc)
                else
                  objdata.writebytes(currval,1);
              end;
            &20,&21,&22 :
              begin
                getvalsym(c-&20);
                if (currval<-256) or (currval>255) then
                 Message2(asmw_e_value_exceeds_bounds,'byte',tostr(currval));
                if assigned(currsym) then
                 objdata_writereloc(currval,1,currsym,currabsreloc)
                else
                 objdata.writebytes(currval,1);
              end;
            &23 :
              begin
                bytes[0]:=ord(codes^)+condval[inverse_cond(condition)];
                inc(codes);
                objdata.writebytes(bytes,1);
              end;
            &24,&25,&26,&27 :
              begin
                getvalsym(c-&24);
                if IF_IMM3 in insentry^.flags then
                  begin
                    if (currval<0) or (currval>7) then
                      Message2(asmw_e_value_exceeds_bounds,'unsigned triad',tostr(currval));
                  end
                else if IF_IMM4 in insentry^.flags then
                  begin
                    if (currval<0) or (currval>15) then
                      Message2(asmw_e_value_exceeds_bounds,'unsigned nibble',tostr(currval));
                  end
                else
                  if (currval<0) or (currval>255) then
                    Message2(asmw_e_value_exceeds_bounds,'unsigned byte',tostr(currval));
                if assigned(currsym) then
                 objdata_writereloc(currval,1,currsym,currabsreloc)
                else
                 objdata.writebytes(currval,1);
              end;
            &30,&31,&32 :     // 030..032
              begin
                getvalsym(c-&30);
                { currval is an aint so this cannot happen on i8086 and causes only a warning }
                if (currval<-65536) or (currval>65535) then
                 Message2(asmw_e_value_exceeds_bounds,'word',tostr(currval));
                if assigned(currsym)
                then
                 objdata_writereloc(currval,2,currsym,currabsreloc)
                else
                 objdata.writebytes(currval,2);
              end;
            &34,&35,&36 :     // 034..036
              { !!! These are intended (and used in opcode table) to select depending
                    on address size, *not* operand size. Works by coincidence only. }
              begin
                getvalsym(c-&34);
                if opsize=S_Q then
                  begin
                    if assigned(currsym) then
                     objdata_writereloc(currval,8,currsym,currabsreloc)
                    else
                     objdata.writebytes(currval,8);
                  end
                else
                  begin
                    if assigned(currsym) then
                      objdata_writereloc(currval,4,currsym,currabsreloc32)
                    else
                      objdata.writebytes(currval,4);
                  end
              end;
            &40,&41,&42 :    // 040..042
              begin
                getvalsym(c-&40);
                if assigned(currsym)
                then
                 objdata_writereloc(currval,4,currsym,currabsreloc32)
                else
                 objdata.writebytes(currval,4);
              end;
            &44,&45,&46 :// 044..046 - select between word/dword/qword depending on
              begin      // address size (we support only default address sizes).
                getvalsym(c-&44);
                if assigned(currsym) then
                  objdata_writereloc(currval,8,currsym,currabsreloc)
                else
                  objdata.writebytes(currval,8);
              end;
            &50,&51,&52 :   // 050..052 - byte relative operand
              begin
                getvalsym(c-&50);
                data:=currval-insend;
{$push}
{$r-,q-} { disable also overflow as address returns a qword for x86_64 }
                if assigned(currsym) then
                 inc(data,currsym.address);
{$pop}
                if (data>127) or (data<-128) then
                 Message1(asmw_e_short_jmp_out_of_range,tostr(data));
                objdata.writebytes(data,1);
              end;
            &54,&55,&56:   // 054..056 - qword immediate operand
              begin
                getvalsym(c-&54);
                if assigned(currsym) then
                  objdata_writereloc(currval,8,currsym,currabsreloc)
                else
                  objdata.writebytes(currval,8);
              end;
            &60,&61,&62 :
              begin
                getvalsym(c-&60);
                InternalError(2020100821);
              end;
            &64,&65,&66 :  // 064..066 - select between 16/32 address mode, but we support only 32 (only 16 on i8086)
              begin
                getvalsym(c-&64);
                if assigned(currsym) then
                 objdata_writereloc(currval,4,currsym,currrelreloc)
                else
                 objdata_writereloc(currval-insend,4,nil,currabsreloc32)
              end;
            &70,&71,&72 :  // 070..072 - long relative operand
              begin
                getvalsym(c-&70);
                if assigned(currsym) then
                 objdata_writereloc(currval,4,currsym,currrelreloc)
                else
                 objdata_writereloc(currval-insend,4,nil,currabsreloc32)
              end;
            &74,&75,&76 : ; // 074..076 - vex-coded vector operand
                            // ignore
            &254,&255,&256 :  // 0254..0256 - dword implicitly sign-extended to 64-bit (x86_64 only)
              begin
                getvalsym(c-&254);
                { for i386 as aint type is longint the
                  following test is useless }
                if (currval<low(longint)) or (currval>high(longint)) then
                  Message2(asmw_e_value_exceeds_bounds,'signed dword',tostr(currval));
                if assigned(currsym) then
                  objdata_writereloc(currval,4,currsym,currabsreloc32)
                else
                  objdata.writebytes(currval,4);
              end;
            &300,&301,&302:
              begin
                if (oper[c and 3]^.ot and OT_SIZE_MASK)=OT_BITS32 then
                  write0x67prefix(objdata);
              end;
            &310 :   { fixed 16-bit addr }
              { every insentry having code 0310 must be marked with NOX86_64 }
              InternalError(2011051302);
            &311 :   { fixed 32-bit addr }
              write0x67prefix(objdata)
              ;
            &320,&321,&322 :
              begin
                case oper[c-&320]^.ot and OT_SIZE_MASK of
                  OT_BITS16 :
                    write0x66prefix(objdata);
                end;
              end;
            &323 : {no action needed};
            &325:
              {no action needed};
            &324,
            &361:
              begin
                if not(needed_VEX or needed_EVEX) then
                  write0x66prefix(objdata);
              end;
            &326 :
              begin
              end;
            &333 :
              begin
                if not(needed_VEX or needed_EVEX) then
                begin
                  bytes[0]:=$f3;
                  objdata.writebytes(bytes,1);
                end;
              end;
            &334 :
              begin
                if not(needed_VEX or needed_EVEX) then
                begin
                  bytes[0]:=$f2;
                  objdata.writebytes(bytes,1);
                end;
              end;
            &335:
              ;
            &336: ; // indicates 32-bit scalar vector operand {no action needed}
            &337: ; // indicates 64-bit scalar vector operand {no action needed}
            &312,
            &327,
            &331,&332 :
              begin
                { these are dissambler hints or 32 bit prefixes which
                  are not needed }
              end;
            &362..&364: ; // VEX flags =>> nothing todo
            &366, &367:
              begin
                opidx:=c-&364;  { 0366->operand 2, 0367->operand 3 }
                if (needed_VEX or needed_EVEX) and
                  (ops=4) and
                  (oper[opidx]^.typ=top_reg) and
                  (
                   ((oper[opidx]^.ot and OT_REG_EXTRA_MASK)=otf_reg_xmm) or
                   ((oper[opidx]^.ot and OT_REG_EXTRA_MASK)=otf_reg_ymm) or
                   ((oper[opidx]^.ot and OT_REG_EXTRA_MASK)=otf_reg_zmm)
                  ) then
                  begin
                    bytes[0] := ((getsupreg(oper[opidx]^.reg) and 15) shl 4);
                    objdata.writebytes(bytes,1);
                  end
                else
                  Internalerror(2014032001);
              end;
            &350..&352: ; // EVEX flags =>> nothing todo
            &370..&372: ; // VEX flags =>> nothing todo
            &37:
              begin
                InternalError(2020100822);
              end;
            else
              begin
                { rex should be written at this point }
                if not(needed_VEX or needed_EVEX) then  // TG
                  if (rex<>0) and not(rexwritten) then
                    internalerror(200603191);
                if (c>=&100) and (c<=&227) then  // 0100..0227
                 begin
                   if (c<&177) then            // 0177
                    begin
                      if (oper[c and 7]^.typ=top_reg) then
                        rfield:=regval(oper[c and 7]^.reg)
                      else
                        rfield:=regval(oper[c and 7]^.ref^.base);
                    end
                   else
                    rfield:=c and 7;
                   opidx:=(c shr 3) and 7;

                   if not process_ea(oper[opidx]^,ea_data,rfield, EVEXTupleState = etsNotTuple) then
                    Message(asmw_e_invalid_effective_address);


                   pb:=@bytes[0];
                   pb^:=ea_data.modrm;
                   inc(pb);
                   if ea_data.sib_present then
                    begin
                      pb^:=ea_data.sib;
                      inc(pb);
                    end;

                   s:=pb-@bytes[0];
                   objdata.writebytes(bytes,s);

                   case ea_data.bytes of
                     0 : ;
                     1 :
                       begin
                         if (oper[opidx]^.ot and OT_MEMORY)=OT_MEMORY then
                           begin
                             currsym:=objdata.symbolref(oper[opidx]^.ref^.symbol);
                             if oper[opidx]^.ref^.refaddr=addr_pic then
                               currabsreloc:=RELOC_GOTPCREL
                             else
                               currabsreloc:=RELOC_ABSOLUTE;
                             objdata_writereloc(oper[opidx]^.ref^.offset,1,currsym,currabsreloc);
                           end
                         else
                          begin
                            bytes[0]:=oper[opidx]^.ref^.offset;
                            objdata.writebytes(bytes,1);
                          end;
                         inc(s);
                       end;
                     2,4 :
                       begin
                         currsym:=objdata.symbolref(oper[opidx]^.ref^.symbol);
                         currval:=oper[opidx]^.ref^.offset;
                         if oper[opidx]^.ref^.refaddr=addr_pic then
                           currabsreloc:=RELOC_GOTPCREL
                         else if oper[opidx]^.ref^.refaddr=addr_tlsgd then
                           currabsreloc:=RELOC_TLSGD
                         else if oper[opidx]^.ref^.refaddr=addr_tpoff then
                           currabsreloc:=RELOC_TPOFF
                         else
                           if oper[opidx]^.ref^.base=NR_RIP then
                             begin
                               currabsreloc:=RELOC_RELATIVE;
                               { Adjust reloc value by number of bytes following the displacement,
                                 but not if displacement is specified by literal constant }
                               if Assigned(currsym) then
                                 Dec(currval,InsEnd-objdata.CurrObjSec.Size-ea_data.bytes);
                             end
                           else
                             currabsreloc:=RELOC_ABSOLUTE32;

                           if (currabsreloc in [RELOC_ABSOLUTE32{$ifdef i8086},RELOC_ABSOLUTE{$endif}]) and
                            (Assigned(oper[opidx]^.ref^.relsymbol)) then
                           begin
                             relsym:=objdata.symbolref(oper[opidx]^.ref^.relsymbol);
                             if relsym.objsection=objdata.CurrObjSec then
                               begin
                                 currval:=objdata.CurrObjSec.size+ea_data.bytes-relsym.offset+currval;
                                   currabsreloc:=RELOC_RELATIVE;
                               end;
                           end;
                         objdata_writereloc(currval,ea_data.bytes,currsym,currabsreloc);
                         inc(s,ea_data.bytes);
                       end;
                   end;
                 end
                else
                 InternalError(777007);
              end;
          end;
        until false;
      end;


    function taicpu.is_same_reg_move(regtype: Tregistertype):boolean;
      begin
        result:=(((opcode=A_MOV) or (opcode=A_XCHG)) and
                 (regtype = R_INTREGISTER) and
                 (ops=2) and
                 (oper[0]^.typ=top_reg) and
                 (oper[1]^.typ=top_reg) and
                 (oper[0]^.reg=oper[1]^.reg)
                ) or
                ({ checking the opcodes is a long "or" chain, so check first the registers which is more selective }
                 ((regtype = R_MMREGISTER) and
                  (ops=2) and
                  (oper[0]^.typ=top_reg) and
                  (oper[1]^.typ=top_reg) and
                  (oper[0]^.reg=oper[1]^.reg)) and
                  (
                   (opcode=A_MOVSS) or (opcode=A_MOVSD) or
                   (opcode=A_MOVQ) or (opcode=A_MOVD) or
                   (opcode=A_MOVAPS) or (opcode=A_MOVAPD) or
                   (opcode=A_MOVUPS) or (opcode=A_MOVUPD) or
                   (opcode=A_MOVDQA) or (opcode=A_MOVDQU) or
                   (opcode=A_VMOVSS) or (opcode=A_VMOVSD) or
                   (opcode=A_VMOVQ) or (opcode=A_VMOVD) or
                   (opcode=A_VMOVAPS) or (opcode=A_VMOVAPD) or
                   (opcode=A_VMOVUPS) or (opcode=A_VMOVUPD) or
                   (opcode=A_VMOVDQA) or (opcode=A_VMOVDQU)
                  )
                );
      end;


    procedure build_spilling_operation_type_table;
      var
        opcode : tasmop;
      begin
        new(operation_type_table);
        fillchar(operation_type_table^,sizeof(toperation_type_table),byte(operand_read));
        for opcode:=low(tasmop) to high(tasmop) do
          with InsProp[opcode] do
          begin
              if Ch_Rop1 in Ch then
                    operation_type_table^[opcode,0]:=operand_read;
              if Ch_Wop1 in Ch then
                    operation_type_table^[opcode,0]:=operand_write;
              if [Ch_RWop1,Ch_Mop1]*Ch<>[] then
                    operation_type_table^[opcode,0]:=operand_readwrite;
              if Ch_Rop2 in Ch then
                    operation_type_table^[opcode,1]:=operand_read;
              if Ch_Wop2 in Ch then
                    operation_type_table^[opcode,1]:=operand_write;
              if [Ch_RWop2,Ch_Mop2]*Ch<>[] then
                    operation_type_table^[opcode,1]:=operand_readwrite;
              if Ch_Rop3 in Ch then
                    operation_type_table^[opcode,2]:=operand_read;
              if Ch_Wop3 in Ch then
                    operation_type_table^[opcode,2]:=operand_write;
              if [Ch_RWop3,Ch_Mop3]*Ch<>[] then
                    operation_type_table^[opcode,2]:=operand_readwrite;
              if Ch_Rop4 in Ch then
                operation_type_table^[opcode,3]:=operand_read;
              if Ch_Wop4 in Ch then
                operation_type_table^[opcode,3]:=operand_write;
              if [Ch_RWop4,Ch_Mop4]*Ch<>[] then
                operation_type_table^[opcode,3]:=operand_readwrite;
                end;
              end;


    function taicpu.spilling_get_operation_type(opnr: longint): topertype;
      begin
        { the information in the instruction table is made for the string copy
          operation MOVSD so hack here (FK)

          VMOVSS and VMOVSD has two and three operand flavours, this cannot modelled by x86ins.dat
          so fix it here (FK)
        }
        if ((opcode=A_MOVSD) or (opcode=A_VMOVSS) or (opcode=A_VMOVSD)) and (ops=2) then
          begin
            case opnr of
              0:
                result:=operand_read;
              1:
                result:=operand_write;
              else
                internalerror(200506055);
            end
          end
        { IMUL has 1, 2 and 3-operand forms }
        else if opcode=A_IMUL then
          begin
            case ops of
              1:
                if opnr=0 then
                  result:=operand_read
                else
                  internalerror(2014011802);
              2:
                begin
                  case opnr of
                    0:
                      result:=operand_read;
                    1:
                      result:=operand_readwrite;
                    else
                      internalerror(2014011803);
                  end;
                end;
              3:
                begin
                  case opnr of
                    0,1:
                      result:=operand_read;
                    2:
                      result:=operand_write;
                    else
                      internalerror(2014011804);
                  end;
                end;
              else
                internalerror(2014011805);
            end;
          end
        else
          result:=operation_type_table^[opcode,opnr];
      end;


    function spilling_create_load(const ref:treference;r:tregister):Taicpu;
      var
        tmpref: treference;
      begin
        tmpref:=ref;
        case getregtype(r) of
          R_INTREGISTER :
            begin
              if getsubreg(r)=R_SUBH then
                inc(tmpref.offset);
              { we don't need special code here for 32 bit loads on x86_64, since
                those will automatically zero-extend the upper 32 bits. }
              result:=taicpu.op_ref_reg(A_MOV,reg2opsize(r),tmpref,r);
            end;
          R_MMREGISTER :
              case getsubreg(r) of
                R_SUBMMD:
                  result:=taicpu.op_ref_reg(A_MOVSD,S_NO,tmpref,r);
                R_SUBMMS:
                  result:=taicpu.op_ref_reg(A_MOVSS,S_NO,tmpref,r);
                R_SUBQ,
                R_SUBMMWHOLE:
                  result:=taicpu.op_ref_reg(A_MOVQ,S_NO,tmpref,r);
                R_SUBMMX:
                  result:=taicpu.op_ref_reg(A_MOVDQA,S_NO,tmpref,r);
                else
                  internalerror(2005060405);
              end;
          else
            internalerror(2004010411);
        end;
      end;


    function spilling_create_store(r:tregister; const ref:treference):Taicpu;
      var
        size: topsize;
        tmpref: treference;
      begin
        tmpref:=ref;
        case getregtype(r) of
          R_INTREGISTER :
            begin
              if getsubreg(r)=R_SUBH then
                inc(tmpref.offset);
              size:=reg2opsize(r);
              { even if it's a 32 bit reg, we still have to spill 64 bits
                because we often perform 64 bit operations on them }
              if (size=S_L) then
                begin
                  size:=S_Q;
                  r:=newreg(getregtype(r),getsupreg(r),R_SUBWHOLE);
                end;
              result:=taicpu.op_reg_ref(A_MOV,size,r,tmpref);
            end;
          R_MMREGISTER :
              case getsubreg(r) of
                R_SUBMMD:
                  result:=taicpu.op_reg_ref(A_MOVSD,S_NO,r,tmpref);
                R_SUBMMS:
                  result:=taicpu.op_reg_ref(A_MOVSS,S_NO,r,tmpref);
                R_SUBQ,
                R_SUBMMWHOLE:
                  result:=taicpu.op_reg_ref(A_MOVQ,S_NO,r,tmpref);
                else
                  internalerror(2005060404);
              end;
          else
            internalerror(2004010412);
        end;
      end;


{*****************************************************************************
                              Instruction table
*****************************************************************************}

    procedure BuildInsTabCache;
      var
        i : longint;
      begin
        new(instabcache);
        FillChar(instabcache^,sizeof(tinstabcache),$ff);
        i:=0;
        while (i<InsTabEntries) do
         begin
           if InsTabCache^[InsTab[i].OPcode]=-1 then
            InsTabCache^[InsTab[i].OPcode]:=i;
           inc(i);
         end;
      end;


    procedure BuildInsTabMemRefSizeInfoCache;
    var
      AsmOp: TasmOp;
      i,j: longint;
      iCntOpcodeValError: longint;
      insentry  : PInsEntry;

      MRefInfo: TMemRefSizeInfo;
      SConstInfo: TConstSizeInfo;
      actRegSize: int64;
      actMemSize: int64;
      actConstSize: int64;
      actRegCount: integer;
      actMemCount: integer;
      actConstCount: integer;
      actRegTypes  : int64;
      actRegMemTypes: int64;
      NewRegSize: int64;

      actVMemCount  : integer;
      actVMemTypes  : int64;

      RegMMXSizeMask: int64;
      RegXMMSizeMask: int64;
      RegYMMSizeMask: int64;
      RegZMMSizeMask: int64;

      RegMMXConstSizeMask: int64;
      RegXMMConstSizeMask: int64;
      RegYMMConstSizeMask: int64;
      RegZMMConstSizeMask: int64;

      RegBCSTSizeMask: int64;
      RegBCSTXMMSizeMask: int64;
      RegBCSTYMMSizeMask: int64;
      RegBCSTZMMSizeMask: int64;
      ExistsMemRef      : boolean;

      bitcount          : integer;
      ExistsCode336     : boolean;
      ExistsCode337     : boolean;
      ExistsSSEAVXReg   : boolean;

      hs1,hs2 : String;

      function bitcnt(aValue: int64): integer;
      var
        i: integer;
      begin
        result := 0;

        for i := 0 to 63 do
        begin
          if (aValue mod 2) = 1 then
          begin
            inc(result);
          end;

          aValue := aValue shr 1;
        end;
      end;

    begin
      new(InsTabMemRefSizeInfoCache);
      FillChar(InsTabMemRefSizeInfoCache^,sizeof(TInsTabMemRefSizeInfoCache),0);

      iCntOpcodeValError := 0;
      for AsmOp := low(TAsmOp) to high(TAsmOp) do
      begin
        i := InsTabCache^[AsmOp];

        if i >= 0 then
        begin
          InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize           := msiUnknown;
          InsTabMemRefSizeInfoCache^[AsmOp].MemRefSizeBCST       := msbUnknown;
          InsTabMemRefSizeInfoCache^[AsmOp].BCSTXMMMultiplicator := 0;
          InsTabMemRefSizeInfoCache^[AsmOp].ConstSize            := csiUnknown;
          InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX         := false;
          InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes            := [];

          insentry:=@instab[i];
          RegMMXSizeMask := 0;
          RegXMMSizeMask := 0;
          RegYMMSizeMask := 0;
          RegZMMSizeMask := 0;

          RegMMXConstSizeMask := 0;
          RegXMMConstSizeMask := 0;
          RegYMMConstSizeMask := 0;
          RegZMMConstSizeMask := 0;

          RegBCSTSizeMask:= 0;
          RegBCSTXMMSizeMask := 0;
          RegBCSTYMMSizeMask := 0;
          RegBCSTZMMSizeMask := 0;
          ExistsMemRef       := false;

          while (insentry^.opcode=AsmOp) do
          begin
            MRefInfo         := msiUnknown;

            actRegSize       := 0;
            actRegCount      := 0;
            actRegTypes      := 0;
            NewRegSize       := 0;

            actMemSize       := 0;
            actMemCount      := 0;
            actRegMemTypes   := 0;

            actVMemCount     := 0;
            actVMemTypes     := 0;

            actConstSize     := 0;
            actConstCount    := 0;

            ExistsCode336   := false; // indicate fixed operand size 32 bit
            ExistsCode337   := false; // indicate fixed operand size 64 bit
            ExistsSSEAVXReg := false;

            // parse insentry^.code for &336 and &337
            // &336 (octal) = 222 (decimal) == fixed operand size 32 bit
            // &337 (octal) = 223 (decimal) == fixed operand size 64 bit
            for i := low(insentry^.code) to high(insentry^.code) do
            begin
              case insentry^.code[i] of
                #222: ExistsCode336 := true;
                #223: ExistsCode337 := true;
                #0,#1,#2,#3: break;
              end;
            end;

            for i := 0 to insentry^.ops -1 do
            begin
              if (insentry^.optypes[i] and OT_REGISTER) = OT_REGISTER then
               case insentry^.optypes[i] and (OT_XMMREG or OT_YMMREG or OT_ZMMREG or OT_KREG or OT_REG_EXTRA_MASK) of
                  OT_XMMREG,
                  OT_YMMREG,
                  OT_ZMMREG: ExistsSSEAVXReg := true;
                        else;
               end;
            end;


            for j := 0 to insentry^.ops -1 do
            begin
              if ((insentry^.optypes[j] and OT_XMEM32) = OT_XMEM32) OR
                 ((insentry^.optypes[j] and OT_XMEM64) = OT_XMEM64) OR
                 ((insentry^.optypes[j] and OT_YMEM32) = OT_YMEM32) OR
                 ((insentry^.optypes[j] and OT_YMEM64) = OT_YMEM64) OR
                 ((insentry^.optypes[j] and OT_ZMEM32) = OT_ZMEM32) OR
                 ((insentry^.optypes[j] and OT_ZMEM64) = OT_ZMEM64) then
              begin
                inc(actVMemCount);

                case insentry^.optypes[j] and (OT_XMEM32 OR OT_XMEM64 OR OT_YMEM32 OR OT_YMEM64 OR OT_ZMEM32 OR OT_ZMEM64) of
                  OT_XMEM32: actVMemTypes := actVMemTypes or OT_XMEM32;
                  OT_XMEM64: actVMemTypes := actVMemTypes or OT_XMEM64;
                  OT_YMEM32: actVMemTypes := actVMemTypes or OT_YMEM32;
                  OT_YMEM64: actVMemTypes := actVMemTypes or OT_YMEM64;
                  OT_ZMEM32: actVMemTypes := actVMemTypes or OT_ZMEM32;
                  OT_ZMEM64: actVMemTypes := actVMemTypes or OT_ZMEM64;
                        else InternalError(777206);
                end;
              end
              else if (insentry^.optypes[j] and OT_REGISTER) = OT_REGISTER then
              begin
                inc(actRegCount);

                  NewRegSize := (insentry^.optypes[j] and OT_SIZE_MASK);
                  if NewRegSize = 0 then
                    begin
                      case insentry^.optypes[j] and (OT_MMXREG or OT_XMMREG or OT_YMMREG or OT_ZMMREG or OT_KREG or OT_REG_EXTRA_MASK) of
                        OT_MMXREG: begin
                                     NewRegSize := OT_BITS64;
                                   end;
                        OT_XMMREG: begin
                                     NewRegSize := OT_BITS128;
                                     InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX := true;
                                   end;
                        OT_YMMREG: begin
                                     NewRegSize := OT_BITS256;
                                     InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX := true;
                                   end;
                        OT_ZMMREG: begin
                                     NewRegSize := OT_BITS512;
                                     InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX := true;
                                   end;
                          OT_KREG: begin
                                     InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX := true;
                                   end;

                              else NewRegSize := not(0);
                      end;
                  end;

                actRegSize  := actRegSize or NewRegSize;
                actRegTypes := actRegTypes or (insentry^.optypes[j] and (OT_MMXREG or OT_XMMREG or OT_YMMREG or OT_ZMMREG or OT_KREG or OT_REG_EXTRA_MASK));
                end
              else if ((insentry^.optypes[j] and OT_MEMORY) <> 0) then
                begin
                  inc(actMemCount);


                  if ExistsSSEAVXReg and ExistsCode336 then
                    actMemSize := actMemSize or OT_BITS32
                  else if ExistsSSEAVXReg and ExistsCode337 then
                    actMemSize := actMemSize or OT_BITS64
                  else
                    actMemSize:=actMemSize or (insentry^.optypes[j] and (OT_SIZE_MASK OR OT_VECTORBCST));

                  if (insentry^.optypes[j] and OT_REGMEM) = OT_REGMEM then
                    begin
                      actRegMemTypes  := actRegMemTypes or insentry^.optypes[j];
                    end;
                end
              else if ((insentry^.optypes[j] and OT_IMMEDIATE) = OT_IMMEDIATE) then
              begin
                inc(actConstCount);

                actConstSize    := actConstSize or (insentry^.optypes[j] and OT_SIZE_MASK);
              end
            end;

            if actConstCount > 0 then
            begin
              case actConstSize of
                0: SConstInfo := csiNoSize;
                OT_BITS8: SConstInfo := csiMem8;
                OT_BITS16: SConstInfo := csiMem16;
                OT_BITS32: SConstInfo := csiMem32;
                OT_BITS64: SConstInfo := csiMem64;
                else SConstInfo := csiMultiple;
              end;

              if InsTabMemRefSizeInfoCache^[AsmOp].ConstSize = csiUnknown then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].ConstSize := SConstInfo;
              end
              else if InsTabMemRefSizeInfoCache^[AsmOp].ConstSize <> SConstInfo then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].ConstSize := csiMultiple;
              end;
            end;

            if actVMemCount > 0 then
            begin
              if actVMemCount = 1 then
              begin
                if actVMemTypes > 0 then
                begin
                  case actVMemTypes of
                    OT_XMEM32: MRefInfo := msiXMem32;
                    OT_XMEM64: MRefInfo := msiXMem64;
                    OT_YMEM32: MRefInfo := msiYMem32;
                    OT_YMEM64: MRefInfo := msiYMem64;
                    OT_ZMEM32: MRefInfo := msiZMem32;
                    OT_ZMEM64: MRefInfo := msiZMem64;
                          else InternalError(777208);
                  end;

                  case actRegTypes of
                    OT_XMMREG: case MRefInfo of
                                 msiXMem32,
                                 msiXMem64: RegXMMSizeMask := RegXMMSizeMask or OT_BITS128;
                                 msiYMem32,
                                 msiYMem64: RegXMMSizeMask := RegXMMSizeMask or OT_BITS256;
                                 msiZMem32,
                                 msiZMem64: RegXMMSizeMask := RegXMMSizeMask or OT_BITS512;
                                       else InternalError(777210);
                               end;
                    OT_YMMREG: case MRefInfo of
                                 msiXMem32,
                                 msiXMem64: RegYMMSizeMask := RegYMMSizeMask or OT_BITS128;
                                 msiYMem32,
                                 msiYMem64: RegYMMSizeMask := RegYMMSizeMask or OT_BITS256;
                                 msiZMem32,
                                 msiZMem64: RegYMMSizeMask := RegYMMSizeMask or OT_BITS512;
                                       else InternalError(2020100823);
                               end;
                    OT_ZMMREG: case MRefInfo of
                                 msiXMem32,
                                 msiXMem64: RegZMMSizeMask := RegZMMSizeMask or OT_BITS128;
                                 msiYMem32,
                                 msiYMem64: RegZMMSizeMask := RegZMMSizeMask or OT_BITS256;
                                 msiZMem32,
                                 msiZMem64: RegZMMSizeMask := RegZMMSizeMask or OT_BITS512;
                                       else InternalError(2020100824);
                               end;

                          //else InternalError(777209);
                  end;


                  if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize = msiUnknown then
                  begin
                    InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := MRefInfo;
                  end
                  else if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize <> MRefInfo then
                  begin
                    if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize in [msiXMem32, msiXMem64, msiYMem32, msiYMem64, msiZMem32, msiZMem64] then
                    begin
                      InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiVMemMultiple;
                    end
                    else if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize <> msiVMemMultiple then InternalError(777212);
                  end;

                end;
              end
              else InternalError(777207);
            end
            else
              begin
                if (actMemCount=2) and ((AsmOp=A_MOVS) or (AsmOp=A_CMPS)) then actMemCount:=1;

                ExistsMemRef := ExistsMemRef or (actMemCount > 0);

                case actMemCount of
                  0: ; // nothing todo
                  1: begin
                       MRefInfo := msiUnknown;

                       if not(ExistsCode336 or ExistsCode337) then
                         case actRegMemTypes and (OT_MMXRM or OT_XMMRM or OT_YMMRM or OT_ZMMRM or OT_REG_EXTRA_MASK) of
                           OT_MMXRM: actMemSize := actMemSize or OT_BITS64;
                           OT_XMMRM: actMemSize := actMemSize or OT_BITS128;
                           OT_YMMRM: actMemSize := actMemSize or OT_BITS256;
                           OT_ZMMRM: actMemSize := actMemSize or OT_BITS512;
                         end;

                       case actMemSize of
                                  0: MRefInfo := msiNoSize;
                           OT_BITS8: MRefInfo := msiMem8;
                          OT_BITS16: MRefInfo := msiMem16;
                          OT_BITS32: MRefInfo := msiMem32;
                         OT_BITSB32: MRefInfo := msiBMem32;
                          OT_BITS64: MRefInfo := msiMem64;
                         OT_BITSB64: MRefInfo := msiBMem64;
                         OT_BITS128: MRefInfo := msiMem128;
                         OT_BITS256: MRefInfo := msiMem256;
                         OT_BITS512: MRefInfo := msiMem512;
                         OT_BITS80,
                         OT_FAR,
                         OT_NEAR,
                         OT_SHORT: ; // ignore
                         else
                           begin
                             bitcount := bitcnt(actMemSize);

                             if bitcount > 1 then MRefInfo := msiMultiple
                             else InternalError(777203);
                           end;
                       end;

                       if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize = msiUnknown then
                         begin
                           InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := MRefInfo;
                         end
                       else
                       begin
                         // ignore broadcast-memory
                         if not(MRefInfo in [msiBMem32, msiBMem64]) then
                         begin
                           if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize <> MRefInfo then
                           begin
                             with InsTabMemRefSizeInfoCache^[AsmOp] do
                             begin
                               if ((MemRefSize in [msiMem8, msiMULTIPLEMinSize8]) OR (MRefInfo = msiMem8))   then MemRefSize := msiMultipleMinSize8
                               else if ((MemRefSize in [ msiMem16,  msiMULTIPLEMinSize16])  OR (MRefInfo =  msiMem16)) then MemRefSize := msiMultipleMinSize16
                               else if ((MemRefSize in [ msiMem32,  msiMULTIPLEMinSize32])  OR (MRefInfo =  msiMem32)) then MemRefSize := msiMultipleMinSize32
                               else if ((MemRefSize in [ msiMem64,  msiMULTIPLEMinSize64])  OR (MRefInfo =  msiMem64)) then MemRefSize := msiMultipleMinSize64
                               else if ((MemRefSize in [msiMem128, msiMULTIPLEMinSize128])  OR (MRefInfo = msiMem128)) then MemRefSize := msiMultipleMinSize128
                               else if ((MemRefSize in [msiMem256, msiMULTIPLEMinSize256])  OR (MRefInfo = msiMem256)) then MemRefSize := msiMultipleMinSize256
                               else if ((MemRefSize in [msiMem512, msiMULTIPLEMinSize512])  OR (MRefInfo = msiMem512)) then MemRefSize := msiMultipleMinSize512
                               else MemRefSize := msiMultiple;
                             end;
                           end;
                         end;
                       end;

                       //if not(MRefInfo in [msiBMem32, msiBMem64]) and (actRegCount > 0) then
                       if actRegCount > 0 then
                       begin
                         if MRefInfo in [msiBMem32, msiBMem64] then
                         begin
                           if IF_BCST2  in insentry^.flags then InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes := InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes + [bt1to2];
                           if IF_BCST4  in insentry^.flags then InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes := InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes + [bt1to4];
                           if IF_BCST8  in insentry^.flags then InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes := InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes + [bt1to8];
                           if IF_BCST16 in insentry^.flags then InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes := InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes + [bt1to16];

                           //InsTabMemRefSizeInfoCache^[AsmOp].BCSTTypes

                           // BROADCAST - OPERAND
                           RegBCSTSizeMask := RegBCSTSizeMask or actMemSize;

                           case actRegTypes and (OT_XMMREG or OT_YMMREG or OT_ZMMREG or OT_REG_EXTRA_MASK) of
                             OT_XMMREG: RegBCSTXMMSizeMask := RegBCSTXMMSizeMask or actMemSize;
                             OT_YMMREG: RegBCSTYMMSizeMask := RegBCSTYMMSizeMask or actMemSize;
                             OT_ZMMREG: RegBCSTZMMSizeMask := RegBCSTZMMSizeMask or actMemSize;
                                   else begin

                                          RegBCSTXMMSizeMask := not(0);
                                          RegBCSTYMMSizeMask := not(0);
                                          RegBCSTZMMSizeMask := not(0);
                                        end;
                           end;
                         end
                         else
                         case actRegTypes and (OT_MMXREG or OT_XMMREG or OT_YMMREG or OT_ZMMREG or OT_REG_EXTRA_MASK) of
                           OT_MMXREG: if actConstCount > 0 then RegMMXConstSizeMask := RegMMXConstSizeMask or actMemSize
                                       else RegMMXSizeMask := RegMMXSizeMask or actMemSize;
                           OT_XMMREG: if actConstCount > 0 then RegXMMConstSizeMask := RegXMMConstSizeMask or actMemSize
                                       else RegXMMSizeMask := RegXMMSizeMask or actMemSize;
                           OT_YMMREG: if actConstCount > 0 then RegYMMConstSizeMask := RegYMMConstSizeMask or actMemSize
                                       else RegYMMSizeMask := RegYMMSizeMask or actMemSize;
                           OT_ZMMREG: if actConstCount > 0 then RegZMMConstSizeMask := RegZMMConstSizeMask or actMemSize
                                       else RegZMMSizeMask := RegZMMSizeMask or actMemSize;
                                 else begin
                                        RegMMXSizeMask := not(0);
                                        RegXMMSizeMask := not(0);
                                        RegYMMSizeMask := not(0);
                                        RegZMMSizeMask := not(0);

                                        RegMMXConstSizeMask := not(0);
                                        RegXMMConstSizeMask := not(0);
                                        RegYMMConstSizeMask := not(0);
                                        RegZMMConstSizeMask := not(0);
                                      end;
                         end;
                       end
                       else


                     end
                  else InternalError(777202);
                end;
              end;

            inc(insentry);
          end;

          if InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX then
          begin
            case RegBCSTSizeMask of
                      0: ; // ignore;
              OT_BITSB32: begin
                            InsTabMemRefSizeInfoCache^[AsmOp].MemRefSizeBCST       := msbBCST32;
                            InsTabMemRefSizeInfoCache^[AsmOp].BCSTXMMMultiplicator := 4;
                          end;
              OT_BITSB64: begin
                            InsTabMemRefSizeInfoCache^[AsmOp].MemRefSizeBCST       := msbBCST64;
                            InsTabMemRefSizeInfoCache^[AsmOp].BCSTXMMMultiplicator := 2;
                          end;
                    else begin
                           InsTabMemRefSizeInfoCache^[AsmOp].MemRefSizeBCST := msbMultiple;
                         end;
            end;
          end;


          if (InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize in MemRefMultiples) and
             (InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX)then
          begin
            if InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize = msiVMemMultiple then
            begin
              if ((RegXMMSizeMask = OT_BITS128) or (RegXMMSizeMask = 0))     and
                 ((RegYMMSizeMask = OT_BITS256) or (RegYMMSizeMask = 0))     and
                 ((RegZMMSizeMask = OT_BITS512) or (RegZMMSizeMask = 0))     and
                 ((RegXMMSizeMask or RegYMMSizeMask or RegZMMSizeMask) <> 0) then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiVMemRegSize;
              end;
            end
            else if (RegMMXSizeMask or RegMMXConstSizeMask) <> 0 then
            begin
              if ((RegMMXSizeMask or RegMMXConstSizeMask) = OT_BITS64)  and
                 ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS128) and
                 ((RegYMMSizeMask or RegYMMConstSizeMask) = 0)          and
                 ((RegZMMSizeMask or RegZMMConstSizeMask) = 0) then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegSize;
              end;
            end
            else if (((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS128) or ((RegXMMSizeMask or RegXMMConstSizeMask) = 0)) and
                    (((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS256) or ((RegYMMSizeMask or RegYMMConstSizeMask) = 0)) and
                    (((RegZMMSizeMask or RegZMMConstSizeMask) = OT_BITS512) or ((RegZMMSizeMask or RegZMMConstSizeMask) = 0)) and
                    (((RegXMMSizeMask or RegXMMConstSizeMask or
                       RegYMMSizeMask or RegYMMConstSizeMask or
                       RegZMMSizeMask or RegZMMConstSizeMask)) <> 0) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegSize;
            end
            else if (RegXMMSizeMask or RegXMMConstSizeMask = OT_BITS16) and
                    (RegYMMSizeMask or RegYMMConstSizeMask = OT_BITS32) and
                    (RegZMMSizeMask or RegZMMConstSizeMask = 0) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx16y32;
            end
            else if (RegXMMSizeMask or RegXMMConstSizeMask = OT_BITS16) and
                    (RegYMMSizeMask or RegYMMConstSizeMask = OT_BITS32) and
                    (RegZMMSizeMask or RegZMMConstSizeMask = OT_BITS64) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx16y32z64;
            end
            else if ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS32) and
                    ((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS64) then
            begin
              if ((RegZMMSizeMask or RegZMMConstSizeMask) = 0) then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx32y64;
              end
              else if ((RegZMMSizeMask or RegZMMConstSizeMask) = OT_BITS128) then
              begin
                InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx32y64z128;
              end;
            end
            else if ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS64)  and
                    ((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS128) and
                    ((RegZMMSizeMask or RegZMMConstSizeMask) = 0) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx64y128;
            end
            else if ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS64)  and
                    ((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS128) and
                    ((RegZMMSizeMask or RegZMMConstSizeMask) = OT_BITS256) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx64y128z256;
            end
            else if ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS64)  and
                    ((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS256) and
                    ((RegZMMSizeMask or RegZMMConstSizeMask) = 0) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx64y256;
            end
            else if ((RegXMMSizeMask or RegXMMConstSizeMask) = OT_BITS64)  and
                    ((RegYMMSizeMask or RegYMMConstSizeMask) = OT_BITS256) and
                    ((RegZMMSizeMask or RegZMMConstSizeMask) = OT_BITS512) then
            begin
              InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegx64y256z512;
            end
            else if ((RegXMMConstSizeMask = 0) or (RegXMMConstSizeMask = OT_BITS128))     and
                    ((RegYMMConstSizeMask = 0) or (RegYMMConstSizeMask = OT_BITS256))     and
                    ((RegZMMConstSizeMask = 0) or (RegZMMConstSizeMask = OT_BITS512))     and
                    ((RegXMMConstSizeMask or RegYMMConstSizeMask or RegZMMConstSizeMask) <> 0) and
                    (
                     ((RegXMMSizeMask or RegYMMSizeMask or RegZMMSizeMask) = OT_BITS128) or
                     ((RegXMMSizeMask or RegYMMSizeMask or RegZMMSizeMask) = OT_BITS256) or
                     ((RegXMMSizeMask or RegYMMSizeMask or RegZMMSizeMask) = OT_BITS512)
                    ) then
            begin
              case RegXMMSizeMask or RegYMMSizeMask or RegZMMSizeMask of
                OT_BITS128: InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegConst128;
                OT_BITS256: InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegConst256;
                OT_BITS512: InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMemRegConst512;
                       else InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiMultiple;
              end;
            end
            else
            begin
              if not(
                     (AsmOp = A_CVTSI2SS) or
                     (AsmOp = A_CVTSI2SD) or
                     (AsmOp = A_CVTPD2DQ) or
                     (AsmOp = A_VCVTPD2DQ) or
                     (AsmOp = A_VCVTPD2PS) or
                     (AsmOp = A_VCVTSI2SD) or
                     (AsmOp = A_VCVTSI2SS) or
                     (AsmOp = A_VCVTTPD2DQ) or
                     (AsmOp = A_VCVTPD2UDQ) or
                     (AsmOp = A_VCVTQQ2PS) or
                     (AsmOp = A_VCVTTPD2UDQ) or
                     (AsmOp = A_VCVTUQQ2PS) or
                     (AsmOp = A_VCVTUSI2SD) or
                     (AsmOp = A_VCVTUSI2SS) or


                     // TODO check
                     (AsmOp = A_VCMPSS)


                    ) then

              InternalError(777205);
            end;

          end
          else if (InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX) and
                  (InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize = msiUnknown) and
                  (not(ExistsMemRef)) then
          begin
            InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize := msiNoMemRef;
          end;

          InsTabMemRefSizeInfoCache^[AsmOp].RegXMMSizeMask:=RegXMMSizeMask;
          InsTabMemRefSizeInfoCache^[AsmOp].RegYMMSizeMask:=RegYMMSizeMask;
          InsTabMemRefSizeInfoCache^[AsmOp].RegZMMSizeMask:=RegZMMSizeMask;

          if (InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX) and
             (gas_needsuffix[AsmOp] <> AttSufNONE) and
             (not(InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize in MemRefMultiples)) then
          begin
            // combination (attsuffix <> "AttSufNONE") and (MemRefSize is not in MemRefMultiples) is not supported =>> check opcode-definition in x86ins.dat

            if (AsmOp <> A_CVTSI2SD) and
               (AsmOp <> A_CVTSI2SS) then
            begin            
              inc(iCntOpcodeValError);
              Str(gas_needsuffix[AsmOp],hs1);
              Str(InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize,hs2);
              Message3(asmr_e_not_supported_combination_attsuffix_memrefsize_type,
                       '',hs1,hs2);
            end;               
          end;
        end;
      end;

      if iCntOpcodeValError > 0 then
       InternalError(2021011201);

      for AsmOp := low(TAsmOp) to high(TAsmOp) do
      begin


        // only supported intructiones with SSE- or AVX-operands
        if not(InsTabMemRefSizeInfoCache^[AsmOp].ExistsSSEAVX) then
        begin
          InsTabMemRefSizeInfoCache^[AsmOp].MemRefSize  := msiUnknown;
          InsTabMemRefSizeInfoCache^[AsmOp].ConstSize   := csiUnknown;
        end;
      end;
    end;

begin
  cai_align:=tai_align;
  cai_cpu:=taicpu;
build_spilling_operation_type_table;
BuildInsTabCache;
BuildInsTabMemRefSizeInfoCache;
end.
